/*
 * File: $Id: DicomImageSeriesData.java,v 1.2 2006/05/05 07:16:58 hans Exp $
 * History
 * Date: 6. Februar 2003, 09:34    Authors: Hans Slotboom
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.dicom;
 

import org.spectrim.imagegraphicobjects.Plane3D;
import org.spectrim.imagegraphicobjects.Point3D;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;


/**
 * Class that handles the image data of a series of images
 *
 * @author  slotboom
 */
public class DicomImageSeriesData 
{
    
    
    /** 
     * Creates a new instance of DicomImageDataContainer -
     * Default constructor.
     */
    public DicomImageSeriesData(  )
    {
        // no operations required yet!
    }
    
    /**
     * Adds a DicomImageData object to the TreeMap of DicomImageData 
     * instances. 
     * @param data
     */ 
    public void addDicomImageData( DicomImageData data )
    {
         m_imageSeries.put( data.getHeader( ).getSOPInstanceUID( ), data );
    }
    
    /** 
     * Creates a new instance of DicomImageDataContainer.
     * @param fileReferences
     */
    public DicomImageSeriesData( LinkedList<String> fileReferences )
    {
        m_fileReferences = fileReferences;
        
        int counter = 0;
        Iterator it = fileReferences.iterator( );
        while ( it.hasNext( ) )
        {
            String fileName = (String)it.next( );
            DicomImageData data = new DicomImageData( fileName );
            m_imageSeries.put( fileName, data );
            m_imageNumberToFilenameMap.put( new Integer( counter ), fileName );
            m_imageSeriesDescription = data.getHeader( ).getSeriesDescription( );
            counter++;
        }

    }

    /**
     * Returns the TreeSet containing file references of the Dicom data.
     * @return
     */
    public LinkedList<String> getFileReferences( )
    {
        return m_fileReferences;
    }
    
    /**
     * Returns the pixel data of one image of the Series having 
     * filename 'fileName'.
     * @param fileName
     * @return 
     */
    public int[] getImagePixelData( String fileName )
    {
        DicomImageData data = m_imageSeries.get( fileName );
        return data.getPixelData( );
    }

    /**
     * Returns the pixel data of the complete image series.
     * @return 
     */
    public int[] getSeriesPixelData( )
    {
        if ( m_imageSeriesIsOrdered == false )
        {
            return getAllImagePixelData( );
        }
        else
        {
            return getAllOrderedImagePixelData( );
        }
    }
    
    
    /**
     *  Returns the image pixel data of all images in the image series.
     */
    private int[] getAllImagePixelData( )
    {
        // get a reference to all the keys of the Treemap
        Iterator iter = m_imageSeries.keySet( ).iterator( );
        
        // determine the size of the images in the series
        DicomImageData data = m_imageSeries.get( (String)iter.next( ) );
        FileInfo finfo = data.getHeader( );
        int imageSize = finfo.getImageWidth( ) * finfo.getImageHeight( );
        
        // allocate memory for the pixel values
        int[] allImageData = new int[ m_imageSeries.size( ) * imageSize ] ;
        
        // get a reference to all the keys of the Treemap
        Iterator it = m_imageSeries.keySet( ).iterator( );
        
        m_numberOfImages = 0;
        // assign the int[]
        while ( it.hasNext( ) )
        {
            String fileNameKey = (String)it.next( );
            data = m_imageSeries.get( fileNameKey );
            int[] image_data = data.getPixelData( );
            System.arraycopy( image_data, 0, allImageData, m_numberOfImages * imageSize, imageSize ); 
            m_numberOfImages++;
        }
        
        return allImageData;
    }
    
    /**
     *  Returns the image pixel data of all ordered images in the image series.
     */
    private int[] getAllOrderedImagePixelData( )
    {
        // get a reference to all the keys of the Treemap
        Iterator iter = m_orderedImageSeries.keySet( ).iterator( );
        
        // determine the size of the images in the series
        DicomImageData data = m_orderedImageSeries.get( (String)iter.next( ) );
        FileInfo finfo = data.getHeader( );
        //int imageSize = finfo.getImageWidth( ) * finfo.getImageHeight( );
        
        
        int imageSize = finfo.getImageWidth( ) * finfo.getImageHeight( );
  
        m_imageSeriesPatientName = finfo.getPatientName( );
        m_imageSeriesDescription = finfo.getSeriesDescription( );
        
        // allocate memory for the pixel values
        int[] allOrderedImageData = new int[ m_orderedImageSeries.size( ) * imageSize ] ;
        
        // get a reference to all the keys of the Treemap
        Iterator it = m_orderedImageSeries.keySet( ).iterator( );
        
        int counter = 0;
        
        // assign the int[]
        while ( it.hasNext( ) )
        {
            String fileNameKey = (String)it.next( );
            data = m_orderedImageSeries.get( fileNameKey );
            int[] image_data = data.getPixelData( );
            System.arraycopy( image_data, 0, allOrderedImageData, counter * imageSize, imageSize ); 
            counter++;
        }
        
        return allOrderedImageData;
    }
    
    /**
     * Returns the number of images in the dataset.
     */
    public int getNumberOfImages( )
    {
        if ( m_imageSeriesIsOrdered == false )
        {
            return m_imageSeries.size( );
        }
        else
        {
            return m_orderedImageSeries.size( );
        }
    }

    /**
     * 
     * Returns the DicomImageData object belonging to fileNumber.
     * 
     * @param fileNumber
     * @return
     */
    public DicomImageData getDicomImageDataObject( int fileNumber )
    {
        if ( m_imageSeriesIsOrdered == false )
        {
            DicomImageData imData = m_imageSeries.get( m_imageNumberToFilenameMap.get( new Integer( fileNumber ) ) );
            return imData;
        }
        else
        {
            DicomImageData imData = m_orderedImageSeries.get( m_orderedImageNumberToImageOffsetMap.get( new Integer( fileNumber ) ) );
            return imData;
        }
    }
    /**
     * Returns the Dicom data series description.
     * @return 
     */
    public String getDataSeriesDescription( )
    {
        return m_imageSeriesDescription;
    }
    
    /**
     * Returns the Dicom image header object.
     * @param fileNumber
     * @return 
     */
    public FileInfo getFileInfo( int fileNumber )
    {
        if ( m_imageSeriesIsOrdered == false )
        {
            DicomImageData imData = m_imageSeries.get( m_imageNumberToFilenameMap.get( new Integer( fileNumber ) ) );
            return imData.getHeader( );
        }
        else
        {
            DicomImageData imData = m_orderedImageSeries.get( m_orderedImageNumberToImageOffsetMap.get( new Integer( fileNumber ) ) );
            return imData.getHeader( );
        }
    }
    
    /**
     * Returns the Dicom image header object.
     * @param fileNumber
     * @return 
     */
    public LinkedHashMap getFileSiemensInfo( int fileNumber )
    {
        if ( m_imageSeriesIsOrdered == false )
        {
            DicomImageData imData = m_imageSeries.get( m_imageNumberToFilenameMap.get( new Integer( fileNumber ) ) );
            return imData.getSiemensSpecificDataFromFile( );
        }
        else
        {
            DicomImageData imData = m_orderedImageSeries.get( m_orderedImageNumberToImageOffsetMap.get( new Integer( fileNumber ) ) );
            return imData.getSiemensSpecificDataFromFile( );
        }
    }
    
    /**
     * Orders the current m_imageSeries map by the slicePosition attribute.
     */ 
    public void orderBySlicePosition( )
    {

        // Sets the member that indicates that the image series is  ordered
        m_imageSeriesIsOrdered = true;
        m_isOrderedBySlicePosition = true;
        m_orderedImageSeries.clear( );
        // Get a reference to all the keys of the sub map containing the series
        // info only
        int cntr=0;
        Iterator iter = m_imageSeries.keySet( ).iterator( );
        while ( iter.hasNext( ) )
        {
            String fileName = (String)iter.next( );
            DicomImageData dicom_data = m_imageSeries.get( fileName );
            FileInfo finfo = dicom_data.getHeader( );
            if (finfo.isImageData() == true)
            {
                if ( m_volumePerPixel == 0.0 )
                {
                    if ( finfo.getSliceThickness( ).compareTo( "N.A." ) != 0 )
                    {
                        m_pixelDepth = new Double( finfo.getSliceThickness( ) ).doubleValue( );
                    }
                    else
                    {
                        m_pixelDepth = 0.0;
                    }
                    if ( finfo.getPixelHeight( ).compareTo( "N.A." ) != 0 )
                    {
                        m_pixelHeight = new Double( finfo.getPixelHeight( ) ).doubleValue( );
                    }
                    else
                    {
                        m_pixelHeight = 0.0;
                    }
                    if ( finfo.getPixelWidth( ).compareTo( "N.A." ) != 0 )
                    {
                        m_pixelWidth = new Double( finfo.getPixelWidth( ) ).doubleValue( );
                    }
                    else
                    {
                        m_pixelWidth = 0.0;
                    }
                    m_volumePerPixel = m_pixelDepth * m_pixelHeight * m_pixelWidth;
                    m_imageSeriesPatientName = finfo.getPatientName( );
                }


                String ima_pos;
            
                if ( ( finfo.getSliceLocation( ) != null ) &&
                     ( finfo.getSliceLocation( ).compareTo( "N.A." ) != 0 ) &&
                     ( finfo.getSliceLocation( ).compareTo( "" ) != 0 ) )
                {
                    ima_pos = new String( cntr + "_" + new Double( finfo.getSliceLocation( ).trim( ) ).doubleValue( ) );
                    //ima_pos = new String( cntr + "_" + finfo.getSliceLocation( ) );
                }
                else
                {
                    ima_pos = new String( cntr + "_" + new Double( finfo.getInstanceNumber( ).trim( ) ).doubleValue( ) );
                }

                m_orderedImageSeries.put( ima_pos, dicom_data );
                cntr++;
            }
        

            Iterator it = m_orderedImageSeries.keySet( ).iterator( );
            int cntr2 = 0;

            while ( it.hasNext( ) )
            {
                String pos = (String)it.next( );
                String pos1 = pos.substring( pos.indexOf('_') + 1, pos.length( ) );
                // pos = pos.substring( 0, pos.indexOf( '_' ) );
                if ( ( pos1.compareTo( "N.A." ) != 0 ) && ( pos1.compareTo( "" ) != 0 ) )
                {
                    Double posDouble = new Double( pos1 );
                    m_orderedImageNumberToImageOffsetMap.put( new Integer( cntr2 ), cntr2 + "_" + posDouble.doubleValue( ) );
                }
                else
                {
                    pos = pos.substring( 0, pos.indexOf( '_' ) );
                    Double posDouble = new Double( pos );
                    m_orderedImageNumberToImageOffsetMap.put( new Integer( cntr2 ), cntr2 + "_" + posDouble.doubleValue( ) );
                }
            
                cntr2++;
            }
        }
    }
    
    
    /**
     * Orders the current m_imageSeries map by the acquisition number attribute.
     */ 
    public void orderBySliceAcquisitionNumber( )
    {
        // sets the member that indicates that the image series is  ordered
        m_imageSeriesIsOrdered = true;
        
        m_orderedImageSeries.clear( );
        // get a reference to all the keys of the sub map containing the series
        // info only
        Iterator iter = m_imageSeries.keySet( ).iterator( );
        while ( iter.hasNext( ) )
        {
            String fileName = (String)iter.next( );
            DicomImageData dicom_data = m_imageSeries.get( fileName );
            FileInfo finfo = dicom_data.getHeader( );
            if ( m_volumePerPixel == 0.0 )
            {
                m_pixelDepth = new Double( finfo.getSliceThickness( ) ).doubleValue( );
                m_pixelHeight = new Double( finfo.getPixelHeight( ) ).doubleValue( );
                m_pixelWidth = new Double( finfo.getPixelWidth( ) ).doubleValue( );

                m_volumePerPixel = m_pixelDepth * m_pixelHeight * m_pixelWidth;

                m_imageSeriesPatientName = finfo.getPatientName( );
            }
            String acq_num = new String( "" + finfo.getAcquisitionNumber( ) );
            m_orderedImageSeries.put( acq_num, dicom_data );
        }
        Iterator it = m_orderedImageSeries.keySet( ).iterator( );
        int cntr= 0;
        while ( it.hasNext( ) )
        {
            String pos = (String)it.next( );
            pos = pos.substring( pos.indexOf('_') + 1, pos.length( ) );
            Double posDouble = new Double( pos );
            m_orderedImageNumberToImageOffsetMap.put( new Integer( cntr ), cntr + "_" + posDouble.doubleValue( ) );
            cntr++;
        }
    }
    
    /**
     * Returns the name of the patient whose data are in this series.
     * @return 
     */
    public String getImageSeriesPatientName( )
    {
        return m_imageSeriesPatientName;
    }
    
    /**
     * Returns the name of the patient whose data are in this series.
     * @return 
     */
    public String getImageSeriesDescription( )
    {
        return m_imageSeriesDescription;
    }
    
    /**
     * Returns the name of the patient whose data are in this series.
     * @return 
     */
    public double getVolumePerPixel( )
    {
        return m_volumePerPixel;
    }
    
    /**
     * Returns the name of the patient whose data are in this series.
     * @param position
     * @return 
     */
    public int getClosestSliceNumber( double position )
    {
        int cntr = 0;
        double minimumDifference = Double.MAX_VALUE;
        double diff;
        int sliceNumber = -1;
        
        Iterator it = m_orderedImageNumberToImageOffsetMap.keySet( ).iterator( );
        while ( it.hasNext( ) )
        {
            
            String str = m_orderedImageNumberToImageOffsetMap.get( it.next( ) );
            str = str.substring( str.indexOf('_'), str.length( ) );
            double slicePos = new Double( str ).doubleValue( );
            
            diff = Math.abs( slicePos - position );
            if ( diff < minimumDifference )
            {
                minimumDifference = diff;
                sliceNumber = cntr;
            }
            cntr++;
        }
        
        return sliceNumber;
    }
    
    /**
     * Returns the position of a slice number if the data is ordered after
     * slice position.
     * @param orderedSliceNumber
     * @return 
     */
    public double getSlicePositionOfOrderedImages( int orderedSliceNumber )
    {
        String str = m_orderedImageNumberToImageOffsetMap.get( orderedSliceNumber );
        str = str.substring( str.indexOf('_'), str.length( ) );
        double dbl = new Double( str ).doubleValue( );
        
        return dbl;
    }
    
    /**
     * Returns the DicomImageData element.
     * @param key
     * @return 
     */
    public DicomImageData getDicomImageData( String key )
    {
        return m_imageSeries.get( key );
    }
    
    /**
     * Returns the iterator of the image series.
     * @return 
     */
    public Iterator getImageSeriesIterator( )
    {
        return m_imageSeries.keySet( ).iterator( );
    }


    /**
     * Defines the geometry formulas for the images.
     */
    public void defineGeometricImagePlanes( )
    {

        m_geometricImagePlanes.clear( );

        if ( m_isOrderedBySlicePosition == false )
        {
            orderBySlicePosition( );
        }

        Iterator it = m_orderedImageSeries.keySet( ).iterator( );

        while ( it.hasNext( ) )
        {
            String key = (String)it.next( );
            DicomImageData data = m_orderedImageSeries.get( key );

            double r11 = data.getHeader( ).getImageOrientationPatient_1_1( );
            double r12 = data.getHeader( ).getImageOrientationPatient_1_2( );
            double r13 = data.getHeader( ).getImageOrientationPatient_1_3( );

            double r21 = data.getHeader( ).getImageOrientationPatient_2_1( );
            double r22 = data.getHeader( ).getImageOrientationPatient_2_2( );
            double r23 = data.getHeader( ).getImageOrientationPatient_2_3( );

            double pp1 = data.getHeader( ).getImagePositionPatientX( );
            double pp2 = data.getHeader( ).getImagePositionPatientY( );
            double pp3 = data.getHeader( ).getImagePositionPatientZ( );

            int numberOfInitialPointsX = data.getHeader( ).getInitialImageWidth( );
            int numberOfInitialPointsY = data.getHeader( ).getInitialImageHeight( );

            int numberCurrentOfPointsX = data.getHeader( ).getImageWidth( );
            int numberCurrentOfPointsY = data.getHeader( ).getImageHeight( );
            float sliceThickness = new Float( data.getHeader( ).getSliceThickness( ) ).floatValue( );

            double FOV_X = data.getHeader( ).getInitialImageWidth( ) * new Double( data.getHeader( ).getPixelWidth( ) ).doubleValue( );
            double FOV_Y = data.getHeader( ).getInitialImageHeight( ) * new Double( data.getHeader( ).getPixelHeight( ) ).doubleValue( );

            Point3D rv1 = new Point3D( r11, r12, r13 );
            Point3D rv2 = new Point3D( r21, r22, r23 );
            Point3D pop = new Point3D( pp1, pp2, pp3 );

            Plane3D imagePlane = new Plane3D( pop, rv1, rv2,
                                              numberOfInitialPointsX, numberOfInitialPointsY,
                                              numberCurrentOfPointsX, numberCurrentOfPointsY, FOV_X, FOV_Y,
                                              sliceThickness );

            m_geometricImagePlanes.put( key, imagePlane );
        }
    }

    /**
     * Returns the geometric image plane objects HashMap.
     * @return
     */
    public LinkedHashMap< String, Plane3D > getGeometricImagePlanesObjects( )
    {
        return m_geometricImagePlanes;
    }

    /**
     * Returns the slice number of the slice which is closest to Point3D p.
     * @param p
     * @return 
     */
    public int getClosestSlice( Point3D p )
    {

        if ( m_isOrderedBySlicePosition == false )
        {
            orderBySlicePosition( );
        }

        Iterator it = m_orderedImageSeries.keySet( ).iterator( );

        int minimum = -1;
        int cntr = 0;

        double distance = Double.MAX_VALUE;

        while ( it.hasNext( ) )
        {
            DicomImageData data = m_orderedImageSeries.get( (Double)it.next( ) );

            double r11 = data.getHeader( ).getImageOrientationPatient_1_1( );
            double r12 = data.getHeader( ).getImageOrientationPatient_1_2( );
            double r13 = data.getHeader( ).getImageOrientationPatient_1_3( );

            double r21 = data.getHeader( ).getImageOrientationPatient_2_1( );
            double r22 = data.getHeader( ).getImageOrientationPatient_2_2( );
            double r23 = data.getHeader( ).getImageOrientationPatient_2_3( );

            double pp1 = data.getHeader( ).getImagePositionPatientX( );
            double pp2 = data.getHeader( ).getImagePositionPatientY( );
            double pp3 = data.getHeader( ).getImagePositionPatientZ( );

            Point3D rv1 = new Point3D( r11, r12, r13 );
            Point3D rv2 = new Point3D( r21, r22, r23 );
            Point3D pop = new Point3D( pp1, pp2, pp3 );

            Plane3D imagePlane = new Plane3D( pop, rv1, rv2 );

            double dist = imagePlane.getDistanceToPlane( p );

            if ( dist < distance )
            {
                System.out.println( "distance = " + dist );
                minimum = cntr;
                distance = dist;
            }

            cntr++;
        }

        return minimum;
    }

    /**
     * Computes the Point3D attribute of "sliceNumber"
     * @param sliceNumber
     * @return
     */
    public Point3D getCoordinateOfTopLeftPoint( int sliceNumber )
    {
        DicomImageData imaDat = m_orderedImageSeries.get( sliceNumber );
        FileInfo finfo = imaDat.getHeader( );
        finfo.setGeometricSliceAttributes( );
        Point3D topLeftPoint = finfo.getTopLeftPoint( );
        return topLeftPoint;
    }

    // Attributes
    
    private LinkedList<String> m_fileReferences;
    
    // m_imageSeries contains the non ordered image series treemap
    protected LinkedHashMap<String,DicomImageData> m_imageSeries = new LinkedHashMap<String,DicomImageData>( );
    
    // maps the imageNumber to a filename reference
    protected LinkedHashMap<Integer, String> m_imageNumberToFilenameMap = new LinkedHashMap<Integer, String>( );
     
    // m_imageSeries contains the ordered image series treemap
    protected LinkedHashMap<String,DicomImageData> m_orderedImageSeries = new LinkedHashMap<String,DicomImageData>( );

    protected LinkedHashMap<String, Plane3D> m_geometricImagePlanes = new LinkedHashMap< String,Plane3D >( );

    // maps the imageNumber of the ordered image to a filename reference
    protected LinkedHashMap<Integer,String> m_orderedImageNumberToImageOffsetMap = new LinkedHashMap<Integer,String>( );
    
    boolean m_imageSeriesIsOrdered      = false;
    boolean m_isOrderedBySlicePosition  = false;
   
    private String m_imageSeriesPatientName = "";
    private String m_imageSeriesDescription = "";
    
    private double m_pixelDepth;
    private double m_pixelHeight;
    private double m_pixelWidth;
    
    private double m_volumePerPixel;
    
    // Indicates the number of images in the Dicom Image Series
    int m_numberOfImages;
    boolean m_orderingBySlicePositionDone = false;

    // Attributes necessary to store the texture map of the image
    private int[] m_texture;
    private int   m_selectedTextureType;
}
