/*
 * File: $Id: Spectroscopic3DVoxel.java,v 1.1.1.1 2006/04/19 09:38:47 hans Exp $
 * History
 * Date: 19.10.2004  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2003 - 2004 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */
package org.spectrim.imagegraphicobjects;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Color;
import java.util.ArrayList;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.DecompositionSolver;
import org.apache.commons.math3.linear.LUDecomposition;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;

/**
 *
 * A spectroscopic voxel looks as follows is a cube
 *
 *
 * x6 /------------------------/ x7 / | / | / | / | / | / | / | / | x x2
 * ------------------------- x3 | | | | | | | | | Origin | | /------ z | | O | |
 * / | x5 / ---------------------- / x8 y | / | / | / | / | / | / | / | /
 * ------------------------- x4 x1
 *
 *
 * For a transverse Image Gx, Gy, Gz For a Sagital Image Gz, Gy, Gx For a
 * Coronal Image Gx, Gz, Gy
 *
 * If you have a Sagital description, voxel points ( x, y, z ) are given by:
 *
 * ( sSpecPara.sVoI.dThickness, sSpecPara.sVoI.dPhaseFOV,
 * sSpecPara.sVoI.dReadoutFOV )
 *
 * If you have a Transversal description, voxel points ( x, y, z ) of the voxel
 * are given by:
 *
 * ( sSpecPara.sVoI.dReadoutFOV, sSpecPara.sVoI.dPhaseFOV,
 * sSpecPara.sVoI.dThickness )
 *
 * If you have a Coronal description voxel points ( x, y, z ) of the voxel are
 * given by:
 *
 * ( sSpecPara.sVoI.dPhaseFOV, sSpecPara.sVoI.dThickness,
 * sSpecPara.sVoI.dReadoutFOV )
 *
 * @author hans
 */
public class Spectroscopic3DVoxel 
{

    /**
     * Creates a new instance of Spectroscopic3DVoxel; 
     * A spectroscopic voxel has 8 edge points.
     */
    public Spectroscopic3DVoxel( ) 
    {
    }

    /**
     * Creates a new instance of Spectroscopic3DVoxel A spectroscopic voxel has 
     * 8 edge points.
     * @param leftFrontBottom
     * @param leftFrontTop
     * @param rightFrontTop
     * @param rightFrontBottom
     * @param leftBackBottom
     * @param leftBackTop
     * @param rightBackTop
     * @param rightBackBottom
     */
    public Spectroscopic3DVoxel( Point3D leftFrontBottom, Point3D leftFrontTop, 
                                 Point3D rightFrontTop, Point3D rightFrontBottom,
                                 Point3D leftBackBottom, Point3D leftBackTop, 
                                 Point3D rightBackTop, Point3D rightBackBottom ) 
    {
        m_c1 = leftFrontBottom;
        m_c2 = leftFrontTop;
        m_c3 = rightFrontTop;
        m_c4 = rightFrontBottom;

        m_c5 = leftBackBottom;
        m_c6 = leftBackTop;
        m_c7 = rightBackTop;
        m_c8 = rightBackBottom;

        // Calculate the 6 border planes of the voxel

        // Plane 1: front plane through x1, x2, x4
        Point3D frontPlane_r1 = new Point3D(m_c1.m_x1 - m_c2.m_x1, m_c1.m_x2 - m_c2.m_x2, m_c1.m_x3 - m_c2.m_x3);
        Point3D frontPlane_r2 = new Point3D(m_c1.m_x1 - m_c4.m_x1, m_c1.m_x2 - m_c4.m_x2, m_c1.m_x3 - m_c4.m_x3);

        m_frontVoxelPlane = new Plane3D(m_c1, frontPlane_r1, frontPlane_r2);

        // Plane 2: back plane through x5, x6, x8
        Point3D backPlane_r1 = new Point3D(m_c5.m_x1 - m_c6.m_x1, m_c5.m_x2 - m_c6.m_x2, m_c5.m_x3 - m_c6.m_x3);
        Point3D backPlane_r2 = new Point3D(m_c5.m_x1 - m_c8.m_x1, m_c5.m_x2 - m_c8.m_x2, m_c5.m_x3 - m_c8.m_x3);

        m_backVoxelPlane = new Plane3D(m_c5, backPlane_r1, backPlane_r2);

        // Plane 3: top plane through x2, x3, x6

        Point3D topPlane_r1 = new Point3D(m_c2.m_x1 - m_c3.m_x1, m_c2.m_x2 - m_c3.m_x2, m_c2.m_x3 - m_c3.m_x3);
        Point3D topPlane_r2 = new Point3D(m_c2.m_x1 - m_c6.m_x1, m_c2.m_x2 - m_c6.m_x2, m_c2.m_x3 - m_c6.m_x3);

        m_topVoxelPlane = new Plane3D(m_c2, topPlane_r1, topPlane_r2);

        // Plane 4: bottom plane through x1, x4, x5

        Point3D bottomPlane_r1 = new Point3D(m_c1.m_x1 - m_c4.m_x1, m_c1.m_x2 - m_c4.m_x2, m_c1.m_x3 - m_c4.m_x3);
        Point3D bottomPlane_r2 = new Point3D(m_c1.m_x1 - m_c5.m_x1, m_c1.m_x2 - m_c5.m_x2, m_c1.m_x3 - m_c5.m_x3);

        m_bottomVoxelPlane = new Plane3D(m_c1, bottomPlane_r1, bottomPlane_r2);

        // Plane 5: left plane through x1, x2, x5

        Point3D leftPlane_r1 = new Point3D(m_c1.m_x1 - m_c2.m_x1, m_c1.m_x2 - m_c2.m_x2, m_c1.m_x3 - m_c2.m_x3);
        Point3D leftPlane_r2 = new Point3D(m_c1.m_x1 - m_c5.m_x1, m_c1.m_x2 - m_c5.m_x2, m_c1.m_x3 - m_c5.m_x3);

        m_leftVoxelPlane = new Plane3D(m_c1, leftPlane_r1, leftPlane_r2);

        // Plane 6: right plane through x3, x4, x8
        Point3D rightPlane_r1 = new Point3D(m_c4.m_x1 - m_c3.m_x1, m_c4.m_x2 - m_c3.m_x2, m_c4.m_x3 - m_c3.m_x3);
        Point3D rightPlane_r2 = new Point3D(m_c4.m_x1 - m_c8.m_x1, m_c4.m_x2 - m_c8.m_x2, m_c4.m_x3 - m_c8.m_x3);

        m_rightVoxelPlane = new Plane3D(m_c3, rightPlane_r1, rightPlane_r2);

    }

    /**
     * Performs a translation of the voxel about translation.
     * @param translation
     */
    public void translateVoxel( Point3D translation ) 
    {
        m_c1 = m_c1.translate( translation );
        m_c2 = m_c1.translate( translation );
        m_c3 = m_c1.translate( translation );
        m_c4 = m_c1.translate( translation );
        m_c5 = m_c1.translate( translation );
        m_c6 = m_c1.translate( translation );
        m_c7 = m_c1.translate( translation );
        m_c8 = m_c1.translate( translation );
    }

    /**
     * Paints it self on a Graphics context g on a plane that has p as topmost
     * point r1 as direction vector1 with FOV = FOVX r2 aS direction vector2
     * with FOV = FOVY.
     * @param g
     * @param p
     * @param r1
     * @param r2
     * @param FOVX
     * @param NpixX
     * @param FOVY
     * @param NpixY
     * @param comment
     * @param title_comment
     */
    public void paintComponent( Graphics g, Point3D p, Point3D r1, Point3D r2,
                                double FOVX, double FOVY, int NpixX, int NpixY,
                                String comment, String title_comment) 
    {
        Point3D dummy = new Point3D();

        Point3D pc1 = dummy.projectPointOnPlane( m_c1, p, r1, r2 );
        Point3D pc2 = dummy.projectPointOnPlane( m_c2, p, r1, r2 );
        Point3D pc3 = dummy.projectPointOnPlane( m_c3, p, r1, r2 );
        Point3D pc4 = dummy.projectPointOnPlane( m_c4, p, r1, r2 );
        Point3D pc5 = dummy.projectPointOnPlane( m_c5, p, r1, r2 );
        Point3D pc6 = dummy.projectPointOnPlane( m_c6, p, r1, r2 );
        Point3D pc7 = dummy.projectPointOnPlane( m_c7, p, r1, r2 );
        Point3D pc8 = dummy.projectPointOnPlane( m_c8, p, r1, r2 );

        Point p1 = calculate2DPoint( pc1, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p2 = calculate2DPoint( pc2, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p3 = calculate2DPoint( pc3, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p4 = calculate2DPoint( pc4, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p5 = calculate2DPoint( pc5, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p6 = calculate2DPoint( pc6, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p7 = calculate2DPoint( pc7, p, r1, r2, FOVX, FOVY, NpixX, NpixY );
        Point p8 = calculate2DPoint( pc8, p, r1, r2, FOVX, FOVY, NpixX, NpixY );

        g.setColor( Color.RED );
        g.drawLine( p1.x, p1.y, p2.x, p2.y );
        g.drawLine( p1.x + 1, p1.y + 1, p2.x + 1, p2.y + 1 );

        g.setColor( Color.BLUE );
        g.drawLine( p2.x, p2.y, p3.x, p3.y );
        g.drawLine( p2.x + 1, p2.y + 1, p3.x + 1, p3.y + 1 );

        g.setColor( Color.RED );
        g.drawLine( p3.x, p3.y, p4.x, p4.y );
        g.drawLine( p3.x + 1, p3.y + 1, p4.x + 1, p4.y + 1 );

        g.setColor( Color.BLUE );
        g.drawLine( p4.x, p4.y, p1.x, p1.y );
        g.drawLine( p4.x + 1, p4.y + 1, p1.x + 1, p1.y + 1 );

        g.setColor( Color.YELLOW );
        g.drawLine( p1.x, p1.y, p5.x, p5.y );
        g.drawLine( p1.x + 1, p1.y + 1, p5.x + 1, p5.y + 1 );

        g.drawLine( p2.x, p2.y, p6.x, p6.y );
        g.drawLine( p2.x + 1, p2.y + 1, p6.x + 1, p6.y + 1 );

        g.drawLine( p3.x, p3.y, p7.x, p7.y);
        g.drawLine( p3.x + 1, p3.y + 1, p7.x + 1, p7.y + 1 );

        g.drawLine( p4.x, p4.y, p8.x, p8.y);
        g.drawLine( p4.x + 1, p4.y + 1, p8.x + 1, p8.y + 1 );


        g.setColor( Color.RED );
        g.drawLine( p5.x, p5.y, p6.x, p6.y );
        g.drawLine( p5.x + 1, p5.y + 1, p6.x + 1, p6.y + 1 );

        g.setColor( Color.BLUE );
        g.drawLine( p6.x, p6.y, p7.x, p7.y );
        g.drawLine( p6.x + 1, p6.y + 1, p7.x + 1, p7.y + 1 );

        g.setColor( Color.RED );
        g.drawLine( p7.x, p7.y, p8.x, p8.y );
        g.drawLine( p7.x + 1, p7.y + 1, p8.x + 1, p8.y + 1 );

        g.setColor( Color.BLUE );
        g.drawLine( p8.x, p8.y, p5.x, p5.y );
        g.drawLine( p8.x + 1, p8.y + 1, p5.x + 1, p5.y + 1 );

        if ( comment != null ) 
        {
            if ( comment.length( ) > 0 ) 
            {
                g.setColor( Color.green );
                g.drawString( comment, 5, NpixY - 5 );
            }
        }

        if ( title_comment != null )  
        {
            if ( title_comment.length( ) > 0 ) 
            {
                g.setColor( Color.green );
                g.drawString( title_comment, 5, 15 );
            }
        }

    }

    /**
     * Computes the 2D point from 3D point in a plane that has
     * p as topmost point r1 as direction vector1 with FOV = FOVX r2 aS
     * direction vector2 with FOV = FOVY NpixX number of points in the r1
     * direction NpixY number of points in the r2 direction.
     *
     */
    
    /**
     *
     * Computes the 2D point from 3D point in a plane that has
     * p as topmost point r1 as direction vector1 with FOV = FOVX r2 aS
     * direction vector2 with FOV = FOVY NpixX number of points in the r1
     * direction NpixY number of points in the r2 direction.
     * 
     * @param a
     * @param p
     * @param r1
     * @param r2
     * @param FOVX
     * @param FOVY
     * @param NpixX
     * @param NpixY
     * @return 
     */
    private Point calculate2DPoint( Point3D a, Point3D p, 
                                    Point3D r1, Point3D r2,
                                    double FOVX, double FOVY, 
                                    int NpixX, int NpixY) 
    {

        double solLambda = 0;
        double solMhu = 0;

        double deltaX = a.m_x1 - p.m_x1;
        double deltaY = a.m_x2 - p.m_x2;
        double deltaZ = a.m_x3 - p.m_x3;

        // try to find solution from x, y equation
        double numeratorXY_mhu = r1.m_x2 * deltaX - r1.m_x1 * deltaY;
        double dominatorXY_mhu = r1.m_x2 * r2.m_x1 - r1.m_x1 * r2.m_x2;

        double numeratorXY_lambda = deltaX * r2.m_x2 - deltaY * r2.m_x1;
        double dominatorXY_lambda = r1.m_x1 * r2.m_x2 - r1.m_x2 * r2.m_x1;

        // try to find solution from x, z equation
        double numeratorXZ_mhu = r1.m_x3 * deltaX - r1.m_x1 * deltaZ;
        double dominatorXZ_mhu = r1.m_x3 * r2.m_x1 - r1.m_x1 * r2.m_x3;

        double numeratorXZ_lambda = deltaX * r2.m_x3 - deltaZ * r2.m_x1;
        double dominatorXZ_lambda = r1.m_x1 * r2.m_x3 - r1.m_x3 * r2.m_x1;

        // try to find solution from y, z equation
        double numeratorYZ_mhu = r1.m_x3 * deltaY - r1.m_x2 * deltaZ;
        double dominatorYZ_mhu = r1.m_x3 * r2.m_x2 - r1.m_x2 * r2.m_x3;

        double numeratorYZ_lambda = deltaY * r2.m_x3 - deltaZ * r2.m_x2;
        double dominatorYZ_lambda = r1.m_x2 * r2.m_x3 - r1.m_x3 * r2.m_x2;

        if ( ( Math.abs(dominatorXY_mhu) > 1.0e-9 ) && 
             ( Math.abs(dominatorXY_lambda) > 1.0e-9 ) ) 
        {
            solLambda = (numeratorXY_lambda / dominatorXY_lambda);
            solMhu = (numeratorXY_mhu / dominatorXY_mhu);
        }

        if ( ( Math.abs(dominatorXZ_mhu) > 1.0e-9 ) && 
             ( Math.abs(dominatorXZ_lambda) > 1.0e-9 ) ) 
        {
            solLambda = (numeratorXZ_lambda / dominatorXZ_lambda);
            solMhu = (numeratorXZ_mhu / dominatorXZ_mhu);
        }

        if ( ( Math.abs(dominatorYZ_mhu) > 1.0e-9 ) && 
             ( Math.abs(dominatorYZ_lambda) > 1.0e-9 ) ) 
        {
            solLambda = (numeratorYZ_lambda / dominatorYZ_lambda);
            solMhu = (numeratorYZ_mhu / dominatorYZ_mhu);
        }

        Point P = new Point( );
        P.x = (int)((float) (solLambda * NpixX) / FOVX);
        P.y = (int)((float) (solMhu * NpixY) / FOVY);

        return P;
    }

    /**
     * This method returns whether a spatial 3D point is inside or outside the
     * spectroscopic voxel.
     * @param p
     * @return 
     */
    public boolean isPointWithin3DVoxel( Point3D p ) 
    {

        boolean withinVoxel = true;

        double frontplane_k = m_frontVoxelPlane.getNormalPlaneEquationConstant( );
        double backplane_k = m_backVoxelPlane.getNormalPlaneEquationConstant( );
        double back_point_p_k = m_backVoxelPlane.computePoint( p );


        double topplane_k = m_topVoxelPlane.getNormalPlaneEquationConstant( );
        double bottomplane_k = m_bottomVoxelPlane.getNormalPlaneEquationConstant( );
        double bottom_point_p_k = m_bottomVoxelPlane.computePoint(p);


        double leftplane_k = m_leftVoxelPlane.getNormalPlaneEquationConstant( );
        double rightplane_k = m_rightVoxelPlane.getNormalPlaneEquationConstant( );
        double right_point_p_k = m_rightVoxelPlane.computePoint( p );

        if ( !( ( ( frontplane_k <= back_point_p_k ) && 
                  ( back_point_p_k <= backplane_k ) ) || 
                  ( ( frontplane_k >= back_point_p_k ) && 
                    (back_point_p_k >= backplane_k ) ) ) ) 
        {
            withinVoxel = false;
        }

        if ( !( ( ( topplane_k <= bottom_point_p_k) && 
                  ( bottom_point_p_k <= bottomplane_k ) ) || 
                  ( ( topplane_k >= bottom_point_p_k ) && 
                  ( bottom_point_p_k >= bottomplane_k ) ) ) ) 
        {
            withinVoxel = false;
        }

        if ( !( ( ( leftplane_k <= right_point_p_k ) && 
                  ( right_point_p_k <= rightplane_k ) ) || 
                  ( ( leftplane_k >= right_point_p_k ) && 
                  ( right_point_p_k >= rightplane_k ) ) ) ) 
        {
            withinVoxel = false;
        }

        return withinVoxel;
    }

    /**
     * Returns whether the spectroscopic voxel is
     * @param specvox
     * @return 
     */
    public boolean isInside( Spectroscopic3DVoxel specvox ) 
    {
        boolean ans = true;
        Point3D v1 = m_c6.subtractPoint( m_c5 );
        Point3D v2 = m_c1.subtractPoint( m_c5 );
        Point3D v3 = m_c8.subtractPoint( m_c5 );
        ArrayList<RealVector> constants = new ArrayList<>( );

        Point3D res = new Point3D( );
        
        double[][] matrixData = 
        {
            {v1.m_x1, v2.m_x1, v3.m_x1},
            {v1.m_x2, v2.m_x2, v3.m_x2},
            {v1.m_x3, v2.m_x3, v3.m_x3}
        };
        
        RealMatrix n = new Array2DRowRealMatrix(matrixData);

        constants.add(new ArrayRealVector( new double[]{specvox.m_c1.m_x1, specvox.m_c1.m_x3, specvox.m_c1.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c2.m_x1, specvox.m_c2.m_x3, specvox.m_c2.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c3.m_x1, specvox.m_c3.m_x3, specvox.m_c3.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c4.m_x1, specvox.m_c4.m_x3, specvox.m_c4.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c5.m_x1, specvox.m_c5.m_x3, specvox.m_c5.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c6.m_x1, specvox.m_c6.m_x3, specvox.m_c6.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c7.m_x1, specvox.m_c7.m_x3, specvox.m_c7.m_x3}, false ) );
        constants.add(new ArrayRealVector( new double[]{specvox.m_c8.m_x1, specvox.m_c8.m_x3, specvox.m_c8.m_x3}, false ) );
        DecompositionSolver solver = new LUDecomposition(n).getSolver( );
        for ( int i = 0; i < constants.size(); i++ ) 
        {
            try 
            {
                RealVector solution = solver.solve( constants.get( i ) );
                double[] sol;
                sol = solution.toArray( );
                if ( sol[1] < 0d || sol[1] > 1d || 
                     sol[2] < 0d || sol[2] > 1d || 
                     sol[3] < 0d || sol[3] > 1d ) 
                {
                    ans = false;
                    i = constants.size( );
                }
            } catch ( Exception ex ) 
            {
                // Numeric exception occurred.
            }
        }
        return ans;
    }
    
    // private attributes
    public Point3D m_c1;
    public Point3D m_c2;
    public Point3D m_c3;
    public Point3D m_c4;
    public Point3D m_c5;
    public Point3D m_c6;
    public Point3D m_c7;
    public Point3D m_c8;
    
    // 6 planes encapsulating the voxel
    private Plane3D m_frontVoxelPlane;
    private Plane3D m_backVoxelPlane;
    private Plane3D m_topVoxelPlane;
    private Plane3D m_bottomVoxelPlane;
    private Plane3D m_leftVoxelPlane;
    private Plane3D m_rightVoxelPlane;
}
