/*
 * File: $Id: AdaptiveFilter.java,v 1.1.1.1 2006/04/19 09:38:47 hans Exp $
 * History
 * Date: 07.12.2004 Hans Slotboom  Changes:
 *
 * Copyright (C) 2003-2010 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

/*
 * AdaptiveFiltter.java
 *
 * Created on December 10, 2004, 2:31 PM
 */

package org.spectrim.imageprocessing;

/**
 *
 * Class that allows for adaptive filtering of medical images.
 *
 * @author  hans
 */
public class AdaptiveFilter 
{
    
    /** 
     * Creates a new instance of AdaptiveFiltter.
     */
    public AdaptiveFilter( ) 
    {
    }
    
    
    /**
     * Applies an adaptive mean filter to the integer input image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @param variance
     * @return 
     */
    public int[] applyAdaptiveMeanFilter( int[] x, 
                                          int imageWidth, int imageHeight, int numberOfImages,
                                          int xFilterSize, int yFilterSize, float variance )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float filterSize = xFilterSize * yFilterSize;

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {
            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                
                    float localMean = 0.0f;
                    float localVariance = 0.0f;
                
                    // calculate the mean over the ROI
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            localMean += x[  ima * m_imageWidth * m_imageHeight + ii + imageWidth * jj ];
                        }
                    }
                
                    // Calculate the variance over the ROI
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            localVariance += ( x[  ima * m_imageWidth * m_imageHeight + ii + imageWidth * jj ] - localMean ) * ( x[  ima * m_imageWidth * m_imageHeight + ii + imageWidth * jj ] - localMean );
                        }
                    }

                    procImage[ ima * m_imageWidth * m_imageHeight + coord ] = Math.round( ( 1.0f - ( variance / localVariance ) ) * x[ima * m_imageWidth * m_imageHeight + coord] + ( variance / localVariance ) * localMean );
                }
            }
        }
        
        return procImage;
    }

    
     /**
     * Applies an adaptive median filter to the input integer image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param maxFilterSize
     * @return 
     */
    public int[] applyAdaptiveMedianFilter( int[] x, 
                                            int imageWidth, int imageHeight, int numberOfImages,
                                            int maxFilterSize )
    {
        
        m_maxFilterSize = maxFilterSize;
        
        m_imageWidth = imageWidth;
        m_imageHeight = imageHeight;
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        for ( int ima = 0; ima < numberOfImages; ima++ )
        {
            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    float localMean = 0.0f;
                    float localVariance = 0.0f;
                    int size = 1;
                    procImage[ ima * m_imageWidth * m_imageHeight + coord ] = AMF( x, i, j, ima, size );
                }
            }
        }
        
        return procImage;
    }
    
    /**
     * Applies the adaptive median filtering algorithm.
     */
    private int AMF( int[] x, int i, int j, 
                     int imaNumber, int size )
    {
         
        int[] values = determineSortedValueArrayAroundxy( x, i, j, imaNumber, size );
        
        if ( ( values[0] < values[ values.length / 2 ] ) && ( values[ values.length / 2 ]  < values[ values.length - 1 ] ) )
        {
            return med_or_itself( x, i, j, imaNumber, size );
        }
        else
        {
            size++;
            if ( size < m_maxFilterSize )
            {
                AMF( x, i, j, imaNumber, size );
            }
            else
            {
                return x[ imaNumber * m_imageWidth* m_imageHeight + i * m_imageWidth + j ];
            }
        }
        return 0;
    }
    
    /**
     * Returns the pixel value itself or the median if spike.
     */
    private int med_or_itself( int[] x, int i, int j, 
                               int imaNumber, int size )
    {
        int[] values = determineSortedValueArrayAroundxy( x, i, j, imaNumber,size );
        
        if ( ( values[0] < x[ imaNumber * m_imageWidth * m_imageHeight + i * m_imageWidth + j] ) && ( x[ imaNumber * m_imageWidth * m_imageHeight + i * m_imageWidth + j ] < values[ values.length - 1 ] ) )
        {
            return x[ imaNumber * m_imageWidth* m_imageHeight + i * m_imageWidth + j ];
        }
        else
        {
            return values[ values.length / 2 ];
        }
    }
    
    /**
     * 
     * Determine the pixel value array around ( x, y ) = ( i, j )
     * 
     * @param x
     * @param i
     * @param j
     * @param imaNumber
     * @param size
     * @return 
     */
    public int[]  determineValueArrayAroundxy( int[] x, int i, int j, int imaNumber, int size )
    {
        int[] values = new int[ ( 2 * size + 1 ) * ( 2 * size + 1 ) ];
        
        // current coordinate
        int coord = i * m_imageWidth + j;
        int cntr = 0;
        // Calculate the variance over the ROI

        for ( int k = j - size ; k <= j + size; k++ )
        {
            for ( int l = i - size ; l <= i + size; l++ )
            {
                int ii = ( k + m_imageWidth ) % m_imageWidth;
                int jj = ( l + m_imageHeight ) % m_imageHeight;
                        
                values[cntr] = x[ imaNumber * m_imageWidth* m_imageHeight + ii + m_imageWidth * jj ];
                cntr++;
            }
        }
        
        return values;
    }
    
    /**
     * Returns the pixel value itself or the median if spike.
     * @param x
     * @param i
     * @param j
     * @param imaNumber
     * @param size
     * @return 
     */
    public int[]  determineSortedValueArrayAroundxy( int[] x, int i, int j, 
                                                     int imaNumber, int size )
    {
        int[] sortedValues = determineValueArrayAroundxy( x, i, j, imaNumber, size );
        sortedValues = QuickSort.quickSort( sortedValues );
        
        return sortedValues;
    }
    
    
    // private attributes
    private int m_maxFilterSize = 0;
    private final QuickSort m_qSort = new QuickSort( );
    private int m_imageWidth, m_imageHeight;
    
}
