/*
 * File: $Id: ImageRestorationFilterFrame.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 08.12.2004 Hans Slotboom  
 * Changes: 7-4-2010 a. allow image stacks;
 *                   b. returns filtered data
 *
 * Copyright (C) 2004-2010 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.imageprocessing;

import org.spectrim.dicom.DicomImageSeriesData;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Point;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JTabbedPane;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;


import java.util.Iterator;
import java.util.LinkedList;


import org.spectrim.dicom.DicomSliceCanvasPolyDraw;
import org.spectrim.dicom.FileInfo;

import org.spectrim.lib.DecimalTextField;
import org.spectrim.lib.MessageBox;

/**
 * Frame that enables the filtering of images stack.
 * @author  hans
 */
public class ImageRestorationFilterFrame extends JFrame implements ActionListener 
{

    /** 
     * Constructor of a new ImageRestorationFilterFrame. 
     */
    public ImageRestorationFilterFrame( JFrame parent, DicomImageSeriesData imageSeriesData )
    {
        
        m_parent = parent;

        m_fileInfo = imageSeriesData.getFileInfo( 0 );
        
        m_imagePanel = new JPanel( );
        m_imagePanel.setLayout( new BoxLayout( m_imagePanel, BoxLayout.X_AXIS ) );
        
        m_imagePanel.setMinimumSize( new Dimension( 512, 512 ) );
        m_imagePanel.setMaximumSize( new Dimension( 512, 512 ) );
        m_imagePanel.setPreferredSize( new Dimension( 512, 512 ) );
        
        m_dicomCanvas = new DicomSliceCanvasPolyDraw( this );

        if ( m_fileInfo.getFileName( ) != null )
        {
            m_dicomCanvas.setTitle( m_fileInfo.getFileName( ) );
        }
        else
        {
            setTitle( m_fileInfo.getPatientName( ) + "  " + m_fileInfo.getPatientBirthDate( ) );
        }
        
        m_dicomCanvas.addActionListener( this );

        m_dicomCanvas.setDicomImages( imageSeriesData );
        
        //m_dicomCanvas.setNumberOfImages( numberOfImages );
        //m_dicomCanvas.setDicomFileInfo( m_fileInfo );
        
        m_numberOfImages = imageSeriesData.getNumberOfImages( );

        int[] rawData = imageSeriesData.getSeriesPixelData( );
        
        m_originalPixelData = new int[rawData.length];
        m_filteredPixelData = new int[rawData.length];

        System.arraycopy( rawData, 0, m_originalPixelData, 0, rawData.length );
        System.arraycopy( rawData, 0, m_filteredPixelData, 0, rawData.length );
        
        m_dicomCanvas.setPreferredSize( new Dimension( 512, 512 ) );
        m_dicomCanvas.setMinimumSize( new Dimension( 512, 512 ) );
        m_dicomCanvas.setMaximumSize( new Dimension( 512, 512 ) );

        m_imagePanel.add( Box.createHorizontalGlue( ) );
        m_imagePanel.add( m_dicomCanvas );
        m_imagePanel.add( Box.createHorizontalGlue( ) );


        if ( m_fileInfo.getImageWidth( ) != 0 )
        {
            m_appliedZoom = 512f / m_fileInfo.getImageWidth( );
        }
        
        if ( m_numberOfImages > 1 )
        {
            m_dicomCanvas.enableImagePaging( true );
        }
        else
        {
            m_dicomCanvas.enableImagePaging( false );
        }

        m_dicomCanvas.setVisible( true );
        
        createFilterTypePanel( );
        createApplyCancelUndoPanel( );
         
        m_collectionPanel = new JPanel( );
        m_collectionPanel.setLayout( new BorderLayout( ) );
        
        m_collectionPanel.add( m_imagePanel, BorderLayout.NORTH );
        m_collectionPanel.add( m_filterFamilyType, BorderLayout.CENTER );
        m_collectionPanel.add( m_selectionPanel, BorderLayout.SOUTH );
        
        getContentPane( ).add( m_collectionPanel );
        setVisible( true );
        
        m_dicomCanvas.displayDicomImage( 0 );
        m_dicomCanvas.setZoom( m_appliedZoom );
        
        pack( );
        
        // Put the Dialog nicely in the middle of its parent
        Point centerParent = new Point( m_parent.getX( ) + ( m_parent.getWidth( ) / 2 ) - this.getWidth( ) / 2,
                                        m_parent.getY( ) + ( m_parent.getHeight( ) / 2 ) - this.getHeight( ) / 2 );
        
        setLocation( centerParent );

        m_dicomCanvas.update( true );

    }

    
    /** 
     * Sets the processedData.
     * @param processedData
     */
    public void setImageSeriesPixelData( int[] processedData )
    {
        m_originalPixelData = new int[processedData.length];
        m_filteredPixelData = new int[processedData.length];

        System.arraycopy( processedData, 0, m_originalPixelData, 0, processedData.length );
        System.arraycopy( processedData, 0, m_filteredPixelData, 0, processedData.length );

         m_dicomCanvas.setImageSeriesPixelData( processedData );
         m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
         m_dicomCanvas.update( true );
    }
    
    /**
     * Creates a filter type panel.
     */
    private void createFilterTypePanel( )
    {
        m_filterFamilyType = new JTabbedPane( JTabbedPane.BOTTOM );
        
        createMeanFilterPanel( );
        
        m_filterFamilyType.addTab( "Mean Filters", m_meanFilterCollectionPanel );
        
        createOrderStatisticsPanel( );
        m_filterFamilyType.addTab( "Order Statistics Filters", m_orderStatisticsCollectionPanel );
        
        createAdaptiveFilterPanel( );
        m_filterFamilyType.addTab( "Adaptive Filters", m_adaptiveFilterCollectionPanel );
        
    }
    
    /**
     * Creates a subpanel of the GUI.
     */
    private void createApplyCancelUndoPanel( )
    {
        m_selectionPanel = new JPanel( );
        
        m_selectionPanel.setLayout( new BoxLayout( m_selectionPanel, BoxLayout.X_AXIS ) );
        m_selectionPanel.setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );
        
        m_applyButton = new JButton( "Apply" );
        m_applyAndQuitButton = new JButton( "Apply and Quit" );
        m_undoAllButton = new JButton( "Undo All" );
        m_cancelButton = new JButton( "Cancel" );
        
        m_applyButton.addActionListener( this );
        m_applyAndQuitButton.addActionListener( this );
        m_cancelButton.addActionListener( this );
        m_undoAllButton.addActionListener( this );
        
        m_selectionPanel.add( Box.createGlue( ) );
        m_selectionPanel.add( m_applyButton );
        m_selectionPanel.add( Box.createGlue( ) );
        m_selectionPanel.add( m_applyAndQuitButton );
        m_selectionPanel.add( Box.createGlue( ) );
        m_selectionPanel.add( m_undoAllButton );
        m_selectionPanel.add( Box.createGlue( ) );
        m_selectionPanel.add( m_cancelButton );
        m_selectionPanel.add( Box.createGlue( ) );
    }
    
    /**
     * Creates the Order statistics filter panel.
     */
    private void createOrderStatisticsPanel( )
    {
        m_orderStatisticsCollectionPanel = new JPanel( );
        m_orderStatisticsCollectionPanel.setLayout( new BoxLayout( m_orderStatisticsCollectionPanel, BoxLayout.Y_AXIS ) );
        
        
        m_minFilter = new JRadioButton( "Minimum" );
        m_minFilter.addActionListener( this );
        m_maxFilter = new JRadioButton( "Maximum" );
        m_maxFilter.addActionListener( this );
        m_medianFilter = new JRadioButton( "Median" );
        m_medianFilter.addActionListener( this );
        m_medianFilter.setSelected( true );
        m_midPointFilter = new JRadioButton( "Midpoint" );
        m_midPointFilter.addActionListener( this );
        m_alphatrimmedMeanFilter= new JRadioButton( "Alpha-trimmed Mean" );
        m_alphatrimmedMeanFilter.addActionListener( this );
        
        m_orderStatisticsRButtonPanel = new JPanel( );
        
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsRButtonPanel.add( m_minFilter );
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsRButtonPanel.add( m_maxFilter );
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsRButtonPanel.add( m_medianFilter );
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsRButtonPanel.add( m_midPointFilter );
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsRButtonPanel.add( m_alphatrimmedMeanFilter );
        m_orderStatisticsRButtonPanel.add( Box.createHorizontalGlue( ) );
        
        m_filterTypeButtonGroup = new ButtonGroup( );
        
        m_filterTypeButtonGroup.add( m_minFilter );
        m_filterTypeButtonGroup.add( m_maxFilter );
        m_filterTypeButtonGroup.add( m_medianFilter );
        m_filterTypeButtonGroup.add( m_midPointFilter );
        m_filterTypeButtonGroup.add( m_alphatrimmedMeanFilter );
                
        ///////////////////////////////////////////////////////////////////////
        m_orderStatisticsFilterParameterPanel = new JPanel( );
        m_orderStatisticsFilterParameterPanel.setLayout( new BoxLayout( m_orderStatisticsFilterParameterPanel, BoxLayout.X_AXIS ) );
        
        m_xOrderStatisticsFilterSizeLabel = new JLabel( " Horizontal Filter Size " );
        m_xOrderStatisticsFilterSizeCombobox = new JComboBox( m_meanFilterSizes );
        m_xOrderStatisticsFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_xOrderStatisticsFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_xOrderStatisticsFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_xOrderStatisticsFilterSizeCombobox.addActionListener( this );
        m_xOrderStatisticsFilterSizeCombobox.setSelectedIndex( 1 );

        m_yOrderStatisticsFilterSizeLabel = new JLabel( " Vertical Filter Size " );
        m_yOrderStatisticsFilterSizeCombobox = new JComboBox( m_meanFilterSizes );
        m_yOrderStatisticsFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_yOrderStatisticsFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_yOrderStatisticsFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_yOrderStatisticsFilterSizeCombobox.addActionListener( this );
        m_yOrderStatisticsFilterSizeCombobox.setSelectedIndex( 1 );
        
        m_orderStatisticsFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsFilterParameterPanel.add( m_xOrderStatisticsFilterSizeLabel );
        m_orderStatisticsFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsFilterParameterPanel.add( m_xOrderStatisticsFilterSizeCombobox );
        m_orderStatisticsFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsFilterParameterPanel.add( m_yOrderStatisticsFilterSizeLabel );
        m_orderStatisticsFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_orderStatisticsFilterParameterPanel.add( m_yOrderStatisticsFilterSizeCombobox );
        m_orderStatisticsFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        
        
        m_meanFilterAlphaParameterPanel = new JPanel( );
        m_meanFilterAlphaParameterPanel.setLayout( new BoxLayout( m_meanFilterAlphaParameterPanel, BoxLayout.X_AXIS ) );
        
        m_AlphaValueLabel = new JLabel( "Alpha-value " );
        
        m_orderStatisticsFilterAlphaValue = new DecimalTextField( "0", 1.0f, 1f, 2f, 4f );
        m_orderStatisticsFilterAlphaValue.setPreferredSize( new Dimension( 100, 25 ) );
        m_orderStatisticsFilterAlphaValue.setMaximumSize( new Dimension( 100, 25 ) );
        m_orderStatisticsFilterAlphaValue.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_orderStatisticsFilterAlphaParameterPanel = new JPanel();
        m_orderStatisticsFilterAlphaParameterPanel.setLayout( new BoxLayout( m_orderStatisticsFilterAlphaParameterPanel, BoxLayout.X_AXIS ) );
        
        m_orderStatisticsFilterAlphaParameterPanel.add( m_AlphaValueLabel );
        m_orderStatisticsFilterAlphaParameterPanel.add( m_orderStatisticsFilterAlphaValue );
        m_orderStatisticsFilterAlphaParameterPanel.add( Box.createHorizontalGlue( ) );
     
         // Add sub components to collection component
        m_orderStatisticsCollectionPanel.add( m_orderStatisticsFilterParameterPanel );
        m_orderStatisticsCollectionPanel.add( m_orderStatisticsFilterAlphaParameterPanel );
        m_orderStatisticsCollectionPanel.add( m_orderStatisticsRButtonPanel );
        
        
    }
    
    
    
    /**
     * Creates the GUI to select mean filter type / order and filter parameters.
     */
    private void createMeanFilterPanel( )
    {
        m_meanFilterCollectionPanel = new JPanel( );
        
        m_meanFilterCollectionPanel.setLayout( new BoxLayout(m_meanFilterCollectionPanel, BoxLayout.Y_AXIS ) );
        
        m_arithmeticMeanRadioButton = new JRadioButton( "Arithmetic Mean Filter" );
        m_arithmeticMeanRadioButton.addActionListener( this );
        m_arithmeticMeanRadioButton.setSelected( true );
        
        m_geometricMeanFilter = new JRadioButton( "Geometric Mean Filter" );
        m_geometricMeanFilter.addActionListener( this );
        
        m_HarmonicMeanFilter = new JRadioButton( "Harmonic Mean Filter" );
        m_HarmonicMeanFilter.addActionListener( this );
        
        m_contraHarmonicMeanFilter = new JRadioButton( "Contraharmonic Mean Filter" );
        m_contraHarmonicMeanFilter.addActionListener( this );

        m_meanFilterRButtonPanel = new JPanel( );
        m_meanFilterRButtonPanel.setLayout( new BoxLayout( m_meanFilterRButtonPanel, BoxLayout.X_AXIS ) );
        
        m_meanFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterRButtonPanel.add( m_arithmeticMeanRadioButton );
        m_meanFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterRButtonPanel.add( m_geometricMeanFilter );
        m_meanFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterRButtonPanel.add( m_HarmonicMeanFilter );
        m_meanFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterRButtonPanel.add( m_contraHarmonicMeanFilter );
        m_meanFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        
        m_filterTypeButtonGroup = new ButtonGroup( );
        
        m_filterTypeButtonGroup.add( m_arithmeticMeanRadioButton );
        m_filterTypeButtonGroup.add( m_geometricMeanFilter );
        m_filterTypeButtonGroup.add( m_HarmonicMeanFilter );
        m_filterTypeButtonGroup.add( m_contraHarmonicMeanFilter );
        
        m_meanFilterParameterPanel = new JPanel( );
        m_meanFilterParameterPanel.setLayout( new BoxLayout( m_meanFilterParameterPanel, BoxLayout.X_AXIS ) );
    
        m_xMeanFilterSizeLabel = new JLabel( " Horizontal Filter Size " );
        m_xMeanFilterSizeCombobox = new JComboBox( m_meanFilterSizes );
        m_xMeanFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_xMeanFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_xMeanFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_xMeanFilterSizeCombobox.addActionListener( this );
        m_xMeanFilterSizeCombobox.setSelectedIndex( 1 );

        m_yMeanFilterSizeLabel = new JLabel( " Vertical Filter Size " );
        m_yMeanFilterSizeCombobox = new JComboBox( m_meanFilterSizes );
        m_yMeanFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_yMeanFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_yMeanFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_yMeanFilterSizeCombobox.addActionListener( this );
        m_yMeanFilterSizeCombobox.setSelectedIndex( 1 );
        
        m_meanFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterParameterPanel.add( m_xMeanFilterSizeLabel );
        m_meanFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterParameterPanel.add( m_xMeanFilterSizeCombobox );
        m_meanFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterParameterPanel.add( m_yMeanFilterSizeLabel );
        m_meanFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_meanFilterParameterPanel.add( m_yMeanFilterSizeCombobox );
        m_meanFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        
        
        m_meanFilterQParameterPanel = new JPanel( );
        m_meanFilterQParameterPanel.setLayout( new BoxLayout( m_meanFilterQParameterPanel, BoxLayout.X_AXIS ) );
        
        m_QValueLabel = new JLabel( "Q-value " );
        
        m_meanFilterQValue = new DecimalTextField( "0.000", 0.5f, 0.001f, 0.01f, 0.1f );
        m_meanFilterQValue.setPreferredSize(new Dimension( 100, 25 ) );
        m_meanFilterQValue.setMaximumSize(new Dimension( 100, 25 ) );
        m_meanFilterQValue.setMinimumSize(new Dimension( 100, 25 ) );
        
        m_meanFilterQParameterPanel.add( m_QValueLabel );
        m_meanFilterQParameterPanel.add( m_meanFilterQValue );
        m_meanFilterQParameterPanel.add( Box.createHorizontalGlue( ) );
        
        // Add sub components to collection component
        m_meanFilterCollectionPanel.add( m_meanFilterParameterPanel );
        m_meanFilterCollectionPanel.add( m_meanFilterQParameterPanel );
        m_meanFilterCollectionPanel.add( m_meanFilterRButtonPanel );

    }

    
    
    /**
     * Creates the GUI to select mean filter type / order and filter parameters.
     */
    private void createAdaptiveFilterPanel( )
    {
        m_adaptiveFilterCollectionPanel = new JPanel( );
        
        m_adaptiveFilterCollectionPanel.setLayout( new BoxLayout( m_adaptiveFilterCollectionPanel, BoxLayout.Y_AXIS ) );
        
        m_adaptiveMeanRadioButton = new JRadioButton( "Adaptive Mean Filter" );
        m_adaptiveMeanRadioButton.addActionListener( this );
        m_adaptiveMeanRadioButton.setSelected( true );
        
        m_adaptiveMedianRadioButton = new JRadioButton( "Adaptive Median Filter" );
        m_adaptiveMedianRadioButton.addActionListener( this );
        

        m_adaptiveFilterRButtonPanel = new JPanel( );
        m_adaptiveFilterRButtonPanel.setLayout( new BoxLayout( m_adaptiveFilterRButtonPanel, BoxLayout.X_AXIS ) );
        
        m_adaptiveFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterRButtonPanel.add( m_adaptiveMeanRadioButton );
        m_adaptiveFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterRButtonPanel.add( m_adaptiveMedianRadioButton );
        m_adaptiveFilterRButtonPanel.add( Box.createHorizontalGlue( ) );
        
        m_adaptiveFilterTypeButtonGroup = new ButtonGroup( );
        
        m_adaptiveFilterTypeButtonGroup.add(  m_adaptiveMeanRadioButton );
        m_adaptiveFilterTypeButtonGroup.add( m_adaptiveMedianRadioButton );
       
        
        m_adaptiveFilterParameterPanel = new JPanel( );
        m_adaptiveFilterParameterPanel.setLayout( new BoxLayout( m_adaptiveFilterParameterPanel, BoxLayout.X_AXIS ) );
    
        m_xAdaptiveFilterSizeLabel = new JLabel( " Horizontal Filter Size " );
        m_xAdaptiveFilterSizeCombobox = new JComboBox( m_adaptiveFilterSizes );
        m_xAdaptiveFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_xAdaptiveFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_xAdaptiveFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_xAdaptiveFilterSizeCombobox.addActionListener( this );
        m_xAdaptiveFilterSizeCombobox.setSelectedIndex( 1 );

        m_yAdaptiveFilterSizeLabel = new JLabel( " Vertical Filter Size " );
        m_yAdaptiveFilterSizeCombobox = new JComboBox( m_meanFilterSizes );
        m_yAdaptiveFilterSizeCombobox.setPreferredSize( new Dimension( 100, 25 ) );
        m_yAdaptiveFilterSizeCombobox.setMaximumSize( new Dimension( 100, 25 ) );
        m_yAdaptiveFilterSizeCombobox.setMinimumSize( new Dimension( 100, 25 ) );
        
        m_yAdaptiveFilterSizeCombobox.addActionListener( this );
        m_yAdaptiveFilterSizeCombobox.setSelectedIndex( 1 );
        
        m_adaptiveFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterParameterPanel.add( m_xAdaptiveFilterSizeLabel );
        m_adaptiveFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterParameterPanel.add( m_xAdaptiveFilterSizeCombobox );
        m_adaptiveFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterParameterPanel.add( m_yAdaptiveFilterSizeLabel );
        m_adaptiveFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        m_adaptiveFilterParameterPanel.add( m_yAdaptiveFilterSizeCombobox );
        m_adaptiveFilterParameterPanel.add( Box.createHorizontalGlue( ) );
        
        
        m_adaptiveFilterVarianceParameterPanel = new JPanel( );
        m_adaptiveFilterVarianceParameterPanel.setLayout( new BoxLayout( m_adaptiveFilterVarianceParameterPanel, BoxLayout.X_AXIS ) );
        
        m_VarianceValueLabel = new JLabel( "Noise variance : " );
        
        m_VarianceValue = new DecimalTextField( "0.000", 0.5f, 0.001f, 0.01f, 0.1f );
        m_VarianceValue.setPreferredSize(new Dimension( 100, 25 ) );
        m_VarianceValue.setMaximumSize(new Dimension( 100, 25 ) );
        m_VarianceValue.setMinimumSize(new Dimension( 100, 25 ) );
        
        m_adaptiveFilterVarianceParameterPanel.add( m_VarianceValueLabel );
        m_adaptiveFilterVarianceParameterPanel.add( m_VarianceValue );
        m_adaptiveFilterVarianceParameterPanel.add( Box.createHorizontalGlue( ) );
        
        // Add sub components to collection component
        m_adaptiveFilterCollectionPanel.add( m_adaptiveFilterParameterPanel );
        m_adaptiveFilterCollectionPanel.add( m_adaptiveFilterVarianceParameterPanel );
        m_adaptiveFilterCollectionPanel.add( m_adaptiveFilterRButtonPanel );

    }

    
    
    
    /** 
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents( )
    {

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        pack();
    }
    
    
    /**
     * Is called whenever a registered component raises an ActionEvent.
     */
    public void actionPerformed( ActionEvent e )
    {
                
        if ( e.getSource( ) == this.m_cancelButton )
        {
            dispose( );
        }
        
        if ( e.getSource( ) == this.m_undoAllButton )
        {
            m_dicomCanvas.setImageSeriesPixelData( m_originalPixelData );
            m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
            m_dicomCanvas.update( true );
        }
        
        if ( e.getSource( ) == this.m_applyAndQuitButton )
        {
            
            notifyActionListeners( APPLY_FILTER );
            
            dispose( );
        }

        if ( e.getSource( ) == m_meanFilters )
        {
            if ( e.getID( ) == MeanFilters.FILTERING_APPLIED )
            {
                m_filteredPixelData = m_meanFilters.getFilteredImageData( );
                m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
                m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
                m_dicomCanvas.update( true );
            }
        }


        if ( e.getSource( ) == m_osfFilters )
        {
            if ( e.getID( ) == OrderStatisticsFilter.FILTERING_APPLIED )
            {
                m_filteredPixelData = m_osfFilters.getFilteredImageData( );

                m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
                m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
                m_dicomCanvas.update( true );
            }
        }

        if ( e.getSource( ) == m_applyButton )
        {
            if ( this.m_filterFamilyType.getSelectedIndex( ) == 0 )
            {
                if ( m_arithmeticMeanRadioButton.isSelected( ) == true )
                {
                    m_meanFilters = new MeanFilters( this );

                    m_meanFilters.applyArithmeticMeanFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ),
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );
               
                }
            
                if ( m_geometricMeanFilter.isSelected( ) == true )
                {
                    m_meanFilters = new MeanFilters( this );
                    m_meanFilters.applyGeometricMeanFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ),
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );
                
                }
            
                if ( m_HarmonicMeanFilter.isSelected( ) == true )
                {
                    m_meanFilters = new MeanFilters( this );
                    m_meanFilters.applyHarmonicMeanFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ),
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );

                }
            
                if ( m_contraHarmonicMeanFilter.isSelected( ) == true )
                {
                    m_meanFilters = new MeanFilters( this );
                    m_meanFilters.applyContraHarmonicMeanFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ), 
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yMeanFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        (double)m_meanFilterQValue.getFloatValue( ) );
                
                }
            }
            
            if ( m_filterFamilyType.getSelectedIndex( ) == 1 )
            {
                if ( m_minFilter.isSelected( ) == true )
                {
                    m_osfFilters = new OrderStatisticsFilter( this );
                    
                    m_osfFilters.applyOSFMinimumFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ), 
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );
                
//                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
//                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
//                    m_dicomCanvas.update( true );
                    
                }

                if ( m_maxFilter.isSelected( ) == true )
                {
                    m_osfFilters = new OrderStatisticsFilter( this );
                    
                    m_osfFilters.applyOSFMaximumFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ),
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ) );
                
//                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
//                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
//                    m_dicomCanvas.update( true );

                }
                if ( m_medianFilter.isSelected( ) == true )
                {
                    m_osfFilters = new OrderStatisticsFilter( this );
                    
                    m_osfFilters.applyOSFMedianFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData,
                                                        m_fileInfo.getImageWidth( ),
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );
                
//                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
//                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
//                    m_dicomCanvas.update( true );
                }
                
                if ( m_midPointFilter.isSelected( ) == true )
                {
                    m_osfFilters = new OrderStatisticsFilter( this );
                    
                    m_osfFilters.applyOSFMidPointFilterInSwingWorkerThread(
                                                        m_parent,
                                                        m_filteredPixelData, 
                                                        m_fileInfo.getImageWidth( ), 
                                                        m_fileInfo.getImageHeight( ),
                                                        m_numberOfImages,
                                                        new Integer( (String)m_xOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                        new Integer( (String)m_yOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( )  );
                
//                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
//                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
//                    m_dicomCanvas.update( true );
                }
                
                if ( m_alphatrimmedMeanFilter.isSelected( ) == true )
                {
                     
                     m_osfFilters = new OrderStatisticsFilter( this );
                    
                     m_osfFilters.applyOSFAlphaTrimmedMeanFilterInSwingWorkerThread(
                                                         m_parent,
                                                         m_filteredPixelData, 
                                                         m_fileInfo.getImageWidth( ),
                                                         m_fileInfo.getImageHeight( ),
                                                         m_numberOfImages,
                                                         new Integer( (String)m_xOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                         new Integer( (String)m_yOrderStatisticsFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                         (int)m_orderStatisticsFilterAlphaValue.getFloatValue( ) );
                
//                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
//                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
//                    m_dicomCanvas.update( true );
                }
                
            }
            
            if ( m_filterFamilyType.getSelectedIndex( ) == 2 )
            {
                if ( m_adaptiveMeanRadioButton.isSelected( ) == true )
                {
                    
                    if ( m_dicomCanvas.getVarianceOfSelectedPolygon( ) <= 0f )
                    {
                        // notify to the user that the algorithm needs a noise estimate larger than zero!
                        MessageBox msgBox = new MessageBox( this );
                        msgBox.showModal( "Filter needs drawn ROI in a noisy are of the image", 
                                          "No noise ROI defined ", 
                                          MessageBox.MB_OK );
                    }
                    else
                    {
                        System.out.println( "Variance : " + m_dicomCanvas.getVarianceOfSelectedPolygon( ) );
                        
                        AdaptiveFilter adaptiveFilter = new AdaptiveFilter( );
                        
                        m_filteredPixelData = adaptiveFilter.applyAdaptiveMeanFilter( m_filteredPixelData, 
                                                                m_fileInfo.getImageWidth( ), 
                                                                m_fileInfo.getImageHeight( ),
                                                                m_numberOfImages,
                                                                new Integer( (String)m_xAdaptiveFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                                new Integer( (String)m_yAdaptiveFilterSizeCombobox.getSelectedItem( ) ).intValue( ),
                                                                m_dicomCanvas.getVarianceOfSelectedPolygon( ) );
                        
                        m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
                        m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
                        m_dicomCanvas.update( true );
                    }
                }
                
                if ( m_adaptiveMedianRadioButton.isSelected( ) == true )
                { 
                        
                    AdaptiveFilter adaptiveFilter = new AdaptiveFilter( );
                        
                    m_filteredPixelData = adaptiveFilter.applyAdaptiveMedianFilter( m_filteredPixelData, 
                                                            m_fileInfo.getImageWidth( ), 
                                                            m_fileInfo.getImageHeight( ),
                                                            m_numberOfImages,
                                                            new Integer( (String)m_xAdaptiveFilterSizeCombobox.getSelectedItem( ) ).intValue( ) );
                    
                    m_dicomCanvas.setImageSeriesPixelData( m_filteredPixelData );
                    m_dicomCanvas.displayDicomImage( m_dicomCanvas.getCurrentlyDisplayedImageNumber( ) );
                    m_dicomCanvas.update( true );
                    
                }
            }
        }
    }
    
    /**
     * Adds an action listener to the list of action listeners who are notified
     * when something happens with the spin box.
     * @param e
     */
    public void addActionListener( ActionListener e )
    {
	m_listeners.add( e );
    }
    
    /**
     * Notify all actionListeners.
     */
    private void notifyActionListeners( int eventType )
    {
        Iterator it = m_listeners.iterator( );
        while( it.hasNext( ) )
        {
            ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, eventType, "" ) );
        }
    }
    
    
    
    /**
     * Returns the filtered dataset to the called.
     * @return 
     */
    public int[] getFilteredImageData( )
    {
        return m_filteredPixelData;
    }
    
    
    // Listeners to changes in the segmentation gui control
    private final LinkedList<ActionListener> m_listeners = new LinkedList<>( );
    
    private final JFrame m_parent;
    
    private final JPanel m_imagePanel;
    
    private JRadioButton m_HarmonicMeanFilter;
    private JRadioButton m_arithmeticMeanRadioButton;
    private JRadioButton m_contraHarmonicMeanFilter;
    private JRadioButton m_geometricMeanFilter;
    private ButtonGroup m_filterTypeButtonGroup;
    private JLabel      m_xMeanFilterSizeLabel;
    private JComboBox   m_xMeanFilterSizeCombobox;
    
    private JLabel      m_yMeanFilterSizeLabel;
    private JComboBox   m_yMeanFilterSizeCombobox;
    
    private JLabel              m_QValueLabel;
    private DecimalTextField    m_meanFilterQValue;
    
    private JPanel m_meanFilterCollectionPanel;
    private JPanel m_meanFilterRButtonPanel;
    private JPanel m_meanFilterParameterPanel;
    
    private JPanel m_meanFilterQParameterPanel;
    
    private final String[] m_meanFilterSizes = { "0", "1", "2", "3", "4", "5", "6", "7" };
    
    private final String[] m_adaptiveFilterSizes = { "0", "1", "2", "3", "4", "5", "6", "7" };
    
    private JTabbedPane m_filterFamilyType;
    
    private JPanel m_selectionPanel;
    
    private JButton m_applyButton;
    private JButton m_applyAndQuitButton;
    private JButton m_undoAllButton;
    private JButton m_cancelButton;
    
    private final JPanel m_collectionPanel;
    
    private final DicomSliceCanvasPolyDraw m_dicomCanvas;
    private final FileInfo m_fileInfo;
    
    private float m_appliedZoom;
    
    private int[] m_originalPixelData;
    private int[] m_filteredPixelData;
    
    // Attributes for order statistics frame
    private JPanel m_orderStatisticsCollectionPanel;
    private JPanel m_orderStatisticsRButtonPanel;
    private JRadioButton m_minFilter;
    private JRadioButton m_maxFilter;
    private JRadioButton m_medianFilter;
    private JRadioButton m_midPointFilter;
    private JRadioButton m_alphatrimmedMeanFilter;
    
    private JPanel m_orderStatisticsFilterParameterPanel;
    
    private JLabel m_xOrderStatisticsFilterSizeLabel;
    private JLabel m_yOrderStatisticsFilterSizeLabel;
    private JComboBox m_xOrderStatisticsFilterSizeCombobox;
    private JComboBox m_yOrderStatisticsFilterSizeCombobox;
    
    private JPanel m_orderStatisticsFilterAlphaParameterPanel;
    
    private JPanel m_meanFilterAlphaParameterPanel;
    private JLabel m_AlphaValueLabel;
    private DecimalTextField m_orderStatisticsFilterAlphaValue;
    
    // Attributes for order statistics frame
    private JPanel m_adaptiveFilterCollectionPanel;
    private JPanel m_adaptiveFilterRButtonPanel;
    
    private JPanel m_adaptiveFilterParameterPanel;
    
    private JRadioButton m_adaptiveMeanRadioButton;
    private JRadioButton m_adaptiveMedianRadioButton;
    private ButtonGroup m_adaptiveFilterTypeButtonGroup;
    private JPanel m_adaptiveFilterVarianceParameterPanel;
    
    private JLabel m_xAdaptiveFilterSizeLabel;
    private JLabel m_yAdaptiveFilterSizeLabel;
    private JComboBox m_xAdaptiveFilterSizeCombobox;
    private JComboBox m_yAdaptiveFilterSizeCombobox;
    private JLabel m_VarianceValueLabel;
    
    private DecimalTextField m_VarianceValue;
    
    // Event IDs raised by this control
    public final static int APPLY_FILTER = 0;
    
    private final int               m_numberOfImages;
    private MeanFilters             m_meanFilters;
    private OrderStatisticsFilter   m_osfFilters;

    // Must be declarated somehow
    private static final long serialVersionUID = 34874988952319393L;
}
