/*
 * File: $Id: MeanFilters.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 07.12.2004 Hans Slotboom  Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/*
 * MeanFilters.java
 *
 * Created on December 7, 2004, 11:00 AM
 */

package org.spectrim.imageprocessing;

import org.spectrim.lib.DualProgressBar;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.LinkedList;
import javax.swing.JFrame;

/**
 * 
 * This class implements simple mean filters:
 * 
 *  1. arithmetic filter
 *  2. Geometric filter
 *  3. Harmonic filter
 *  4. Contra-harmonic filter
 * @author hans
 */

public class MeanFilters 
{
    
    /**
     * Creates a new instance of MeanFilters.
     */
    public MeanFilters( ActionListener listener )
    {
        m_listeners.add( listener );
    }

   /**
     * Applies the arithmetic mean filter to the input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyArithmeticMeanFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                              final int imageWidth, final int imageHeight ,
                                                              final int numberOfImages,
                                                              final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying arithmetic mean filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying arithmetic mean filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            public Object construct( )
            {
                m_images = applyArithmeticMeanFilter( x, imageWidth, imageHeight, numberOfImages,
                                                      xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }

     /**
     * Notify all actionListeners.
     */
    private void notifyActionListeners( int eventType )
    {
        Iterator it = m_listeners.iterator( );
        while( it.hasNext( ) )
        {
            ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, eventType, "" ) );
        }
    }

    /**
     * Applies the arithmetic mean filter to the input image x.
     * @param x
     * @param imageWidth
     * @param numberOfImages
     * @param imageHeight
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyArithmeticMeanFilter( int[] x, 
                                            int imageWidth, int imageHeight , int numberOfImages,
                                            int xFilterSize, int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float scaleFactor = ( 2 * xFilterSize + 1 ) * ( 2 * yFilterSize + 1 );

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying arithmetic mean filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    float sum = 0f;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                            sum += x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                        }
                    }
                    procImage[ ima * imageWidth * imageHeight + coord ] = Math.round( sum / scaleFactor );
                }
            }
        }

        return procImage;
    }



    /**
     * Applies the arithmetic mean filter to the input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyGeometricMeanFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                             final int imageWidth, final int imageHeight ,
                                                             final int numberOfImages,
                                                             final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying geometric mean filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying geometric mean filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyGeometricMeanFilter( x, imageWidth, imageHeight, numberOfImages,
                                                     xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }



    /**
     * Applies the geometric mean filter to the input image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyGeometricMeanFilter( int[] x, 
                                           int imageWidth, int imageHeight , int numberOfImages,
                                           int xFilterSize, int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float scaleFactor = ( 2f * xFilterSize + 1f ) * ( 2f * yFilterSize +1f );

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying geometric mean filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    double product = 1.0;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                            product *= x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                        }
                    }
                
                    procImage[ ima * imageWidth * imageHeight + coord ] = (int)Math.round( Math.pow( product , 1.0 / scaleFactor ) );
                
                }
            }
        }

        return procImage;
    }


    /**
     * Applies the arithmetic mean filter to the input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyHarmonicMeanFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                            final int imageWidth, final int imageHeight ,
                                                            final int numberOfImages,
                                                            final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying harmonic mean filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying harmonic mean filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyHarmonicMeanFilter( x, imageWidth, imageHeight, numberOfImages,
                                                    xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }

    /**
     * Applies the Harmonic mean filter to the input image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyHarmonicMeanFilter( int[] x, 
                                          int imageWidth, int imageHeight, int numberOfImages,
                                          int xFilterSize, int yFilterSize )
    {
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float scaleFactor = ( 2f * xFilterSize + 1f ) * ( 2f * yFilterSize +1f );
        
        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying harmonic mean filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    double oneOverSum = 0.0;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            if ( x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ] != 0 )
                            {
                                oneOverSum += ( 1.0 / x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ] );
                            }
                            else
                            {
                                oneOverSum = Double.POSITIVE_INFINITY;
                                break;
                            }
                        
                        }
                        if ( oneOverSum == Double.POSITIVE_INFINITY )
                        {
                            break;
                        }
                    }
                    if ( ( oneOverSum != Double.POSITIVE_INFINITY ) && ( oneOverSum != 0.0 ) )
                    {
                        procImage[ ima * imageWidth * imageHeight + coord ] = (int)Math.round( scaleFactor / oneOverSum );
                    }
                    else
                    {
                        procImage[ ima * imageWidth * imageHeight + coord ] = 0;
                    }
                
                }
            }
        }
        
        return procImage;
    }

    /**
     * Applies the Harmonic mean filter to the input image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @param Q
     * @return 
     */
    public int[] applyContraHarmonicMeanFilter( int[] x, 
                                                int imageWidth, int imageHeight , int numberOfImages,
                                                int xFilterSize, int yFilterSize, double Q )
    {
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float scaleFactor = ( 2f * xFilterSize + 1f ) * ( 2f * yFilterSize +1f );

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying conta harmonic mean filter to image number : " + ima + "</HTML>" );


            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                
                    double nummerator = 0.0;
                    double dominator = 0.0;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            nummerator += Math.pow( x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ], Q + 1.0 );
                            dominator += Math.pow( x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ], Q );
                        }
                    }
                
                    if ( dominator != 0 )
                    {
                        procImage[ ima * imageWidth * imageHeight + coord ] = (int)Math.round( nummerator / dominator );
                    }
                    else
                    {
                        procImage[ ima * imageWidth * imageHeight + coord ] = 0;
                    }
                }
            }
        }

        return procImage;
    }

    /**
     * Applies the arithmetic mean filter to the input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @param Q
     */
    public void applyContraHarmonicMeanFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                                  final int imageWidth, final int imageHeight ,
                                                                  final int numberOfImages,
                                                                  final int xFilterSize, final int yFilterSize,
                                                                  final double Q )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying contra harmonic mean filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying contra harmonic mean filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyContraHarmonicMeanFilter( x, imageWidth, imageHeight, numberOfImages,
                                                          xFilterSize, yFilterSize, Q );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }


    /**
     * Returns the array with image pixel information.
     * @return
     */
    public int[] getFilteredImageData( )
    {
        return m_images;
    }


    // Attributes
    private DualProgressBar m_dualProgressBar;
    private int[] m_images;

    // Listeners to changes in the segmentation gui control
    private final LinkedList<ActionListener> m_listeners = new LinkedList<>( );

    public final static int FILTERING_APPLIED = 0;

}
