/*
 * File: $Id: MorphologicOperator.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 10.03.2003 Hans Slotboom  Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.imageprocessing;

/**
 *
 * Class that handles morphologic operations on mask images:
 *  + erode
 *  + dilate
 * The class also allows the definition of a circular shaped kernel.
 *
 * @author  slotboom
 */

public class MorphologicOperator 
{
    
    /** 
     * Creates a new instance of MorphologicOperator with 
     * default constructor. 
     */
    public MorphologicOperator( ) 
    {
        m_kernelWidth = 1;
        
        m_kernelHeight = 1;
        
        m_kernelLength = 9;
        
        m_kernel = new int[ m_kernelLength ];
        
        m_kernel[0]=0;
        m_kernel[1]=1;
        m_kernel[2]=0;
        m_kernel[3]=1;
        m_kernel[4]=1;
        m_kernel[5]=1;
        m_kernel[6]=0;
        m_kernel[7]=1;
        m_kernel[8]=0;
        
    }
    
    /** 
     * Creates a new instance of MorphologicOperator with known kernel having
     * width 'kernalWidth' and height 'kernalHeight'.
     * @param kernelWidth
     * @param kernelHeight
     */
    public MorphologicOperator( int kernelWidth, int kernelHeight )
    {
        m_kernelWidth = kernelWidth;
        
        m_kernelHeight = kernelHeight;
        
        m_kernelLength = ( 2 * kernelWidth + 1 ) * ( 2 * kernelHeight + 1 );
        
        m_kernel = new int[ m_kernelLength ];
        
        for ( int i = 0; i <= 2 * m_kernelWidth; i++ )
        {
            for ( int j = 0; j <= 2 * m_kernelHeight; j++ )
            {
                m_kernel[ i * m_kernelWidth + j ] = 1;
            }
        }
    }

    /** 
     * Creates a new instance of MorphologicOperator with known kernel having
     * width 'kernalWidth' and height 'kernalHeight'.
     * @param kernelRadius
     */
    public MorphologicOperator( int kernelRadius )
    {
        m_kernelWidth   = kernelRadius;
        m_kernelHeight  = kernelRadius;
        m_kernelRadius  = kernelRadius;
        
        m_kernel = new int[ ( 2 * m_kernelWidth + 1 ) * ( 2* m_kernelHeight + 1 ) ];
        
        // Calculate the kernel
        for ( int i = 0; i <= 2 * m_kernelWidth ; i++ )
        {
            for ( int j = 0; j <= 2* m_kernelHeight; j++ )
            {
                int radiusSquare = ( i - kernelRadius ) * ( i - kernelRadius ) +
                                   ( j - kernelRadius ) * ( j - kernelRadius );
                if ( radiusSquare > kernelRadius * kernelRadius )
                {
                    m_kernel[ i * m_kernelWidth + j ] = 0;
                }
                else
                {
                    m_kernel[ i * m_kernelWidth + j ] = 1;
                }
            }
        }
    }

    /** 
     * Applies the erode function with kernel 'm_kernal' to the image 'image'
     * and returns the eroded image.
     * @param imageWidth
     * @param imageHeight
     * @param image
     * @return 
     */
    public int[] erode( int imageWidth, int imageHeight, int[] image )
    {
        int[] processedImage = new int[ image.length] ;
        if ( imageWidth * imageHeight == image.length )
        {        
            for ( int i = 0; i < imageWidth; i++ )
            {
                for ( int j = 0; j < imageHeight; j++ )
                {
                    // assume that all pixel are 1 of mask and image
                    processedImage[ i * imageHeight + j ] = erodePixel( i, j, imageWidth, imageHeight, image );
                }
            }
        }
        else
        {
            System.out.println("Error: image array size does not fit with specified image sizes");
        }
        return processedImage;
    }

    /**
     * Contains the actual erosion functionality.
     */
    private int erodePixel( int i, int j, int imageWidth, 
                            int imageHeight, int[] image )
    {
        for ( int m = 0 ; m <= 2 * m_kernelWidth; m++ )
        {
            for ( int n = 0; n <= 2 * m_kernelHeight; n++ )
            {
                int ii = ( i + m - m_kernelWidth + imageWidth ) % imageWidth;
                int jj = ( j + n - m_kernelHeight + imageHeight ) % imageHeight;
                
                if ( m_kernel[ m * m_kernelHeight + n ] == 1 )
                {
                    // in this case the kernal != 0 so image mask must be everywhere !=0
                    // for non zero result
                    if ( image[ ii * imageWidth + jj ] == 0 ) 
                    {
                        return 0;
                    }
                }
            }
        }
        return image[ i * imageWidth + j ];
    }
    
    
    /** 
     * Applies the erode function with kernel 'm_kernal' to the image 'image'
     * and returns the eroded image.
     * @param imageWidth
     * @param imageHeight
     * @param image
     * @return 
     */
    public boolean[][] erode( int imageWidth, int imageHeight, boolean[][] image )
    {
        boolean[][] processedImage = new boolean[imageWidth][imageHeight];
        
               
        for ( int i = 0; i < imageWidth; i++ )
        {
            for ( int j = 0; j < imageHeight; j++ )
            {
                // assume that all pixel are 1 of mask and image
                processedImage[i][j] = erodePixel( i, j, imageWidth, imageHeight, image );
            }
        }
        
        return processedImage;
    }

    /**
     * Contains the actual erosion functionality.
     */
    private boolean erodePixel( int i, int j, int imageWidth, 
                                int imageHeight, boolean[][] image )
    {
        for ( int m = 0 ; m <= 2 * m_kernelWidth; m++ )
        {
            for ( int n = 0; n <= 2 * m_kernelHeight; n++ )
            {
                int ii = ( i + m - m_kernelWidth + imageWidth ) % imageWidth;
                int jj = ( j + n - m_kernelHeight + imageHeight ) % imageHeight;
                
                if ( m_kernel[ m * m_kernelHeight + n ] == 1 )
                {
                    // in this case the kernal != 0 so image mask must be everywhere !=0
                    // for non zero result
                    if ( image[ii][jj] == false ) 
                    {
                        return false;
                    }
                }
            }
        }
        return image[i][j];
    }
    
    /** 
     * Applies the dilate function with kernel 'm_kernal' to the image 'image'
     * and returns the dilated image.
     * @param imageWidth
     * @param imageHeight
     * @param image
     * @return 
     */
    public int[] dilate( int imageWidth, int imageHeight, int[] image )
    {
        int[] processedImage = new int[ image.length] ;
        if ( imageWidth * imageHeight == image.length )
        {        
            for ( int i = 0; i < imageWidth; i++ )
            {
                for ( int j = 0; j < imageHeight; j++ )
                {
                    // assume that all pixel are 1 of mask and image
                    processedImage[ i * imageHeight + j ] = dilatePixel( i, j, imageWidth, imageHeight, image );
                }
            }
        }
        else
        {
            System.out.println( "Error: image array size does not fit with specified image sizes" );
        }
        return processedImage;
    }
    
    /**
     * Contains the actual dilate functionality.
     */
    private int dilatePixel( int i, int j, int imageWidth, 
                             int imageHeight, int[] image )
    {
        
        for ( int m = 0 ; m <= 2 * m_kernelWidth; m++ )
        {
            for ( int n = 0; n <= 2 * m_kernelHeight; n++ )
            {
                int ii = ( i + m - m_kernelWidth + imageWidth ) % imageWidth;
                int jj = ( j + n - m_kernelHeight + imageHeight ) % imageHeight;
                
                if ( m_kernel[ m * m_kernelHeight + n ] == 1 )
                {
                    // in this case the kernel != 0 so image mask must be everywhere !=0
                    // for non zero result
                    if ( image[ ii * imageWidth + jj ] != 0 ) 
                    {
                        return image[ ii * imageWidth + jj ];
                    }
                }
            }
        }
        return 0;
    }
    
    /** 
     * Applies the dilate function with kernel 'm_kernal' to the image 'image'
     * and returns the dilated image.
     * @param imageWidth
     * @param imageHeight
     * @param image
     * @return 
     */
    public boolean[][] dilate( int imageWidth, int imageHeight, boolean[][] image )
    {
        
        boolean[][] processedImage = new boolean[imageWidth][imageHeight];
                
        for ( int i = 0; i < imageWidth; i++ )
        {
            for ( int j = 0; j < imageHeight; j++ )
            {
                // assume that all pixel are 1 of mask and image
                processedImage[i][j] = dilatePixel( i, j, imageWidth, imageHeight, image );
            }
        }
        
        return processedImage;
    }
    
    /**
     * Contains the actual dilate functionality.
     */
    private boolean dilatePixel( int i, int j, int imageWidth, 
                                 int imageHeight, boolean[][] image )
    {
        
        for ( int m = 0 ; m <= 2 * m_kernelWidth; m++ )
        {
            for ( int n = 0; n <= 2 * m_kernelHeight; n++ )
            {
                int ii = ( i + m - m_kernelWidth + imageWidth ) % imageWidth;
                int jj = ( j + n - m_kernelHeight + imageHeight ) % imageHeight;
                
                if ( m_kernel[ m * m_kernelHeight + n ] == 1 )
                {
                    // in this case the kernel != 0 so image mask must be everywhere !=0
                    // for non zero result
                    if ( image[ii][jj] != false ) 
                    {
                        
                        return image[ii][jj];
                    }
                }
            }
        }
       
        return false;
    }
    
    
    /**
     * 
     * Test the functionality of the erode and dilate function.
     * 
     * @param args
     */
    public static void main( String[] args )
    {
        MorphologicOperator morph = new MorphologicOperator( 1, 1 );
        //test code int
        int[] image = new int[25];
        int[] imageOut = new int[25];
        int imageWidth = 5;
        int imageHeight = 5;
       
        image[0]=0;  image[1]=0;  image[2]=0;  image[3]=0;  image[4]=0;
        image[5]=0;  image[6]=0;  image[7]=1;  image[8]=0;  image[9]=0;
        image[10]=0; image[11]=0; image[12]=1; image[13]=0; image[14]=0;
        image[15]=0; image[16]=0; image[17]=1; image[18]=0; image[19]=0;
        image[20]=0; image[21]=0; image[22]=0; image[23]=0; image[24]=0;
       
        System.out.println( "Input Image" );
       
        for ( int i = 0; i < imageHeight; i++ ) 
        {
           
            for( int j = 0; j < imageWidth; j++ )
            {
                System.out.print(image[ i * imageWidth + j ]);
            }
           
            System.out.println( );
        }
       
       
        imageOut = morph.dilate( 5, 5, image );
       
       
        System.out.println( "Output Image" );
       
        for ( int i = 0; i < imageHeight; i++ ) 
        {
            for( int j = 0; j < imageWidth; j++ )
            {
                System.out.print( imageOut[i*imageWidth + j] );
            }
            System.out.println( );
        }
       
        //test code boolean
        boolean[][] imageB = new boolean[5][5];
        boolean[][] imageOutB = new boolean[5][5];
        
        imageB[0][0]=false;  imageB[0][1]=false; imageB[0][2]=false; imageB[0][3]=false; imageB[0][4]=false;
        imageB[1][0]=false;  imageB[1][1]=false; imageB[1][2]=true;  imageB[1][3]=false; imageB[1][4]=false;
        imageB[2][0]=false;  imageB[2][1]=false; imageB[2][2]=true;  imageB[2][3]=false; imageB[2][4]=false;
        imageB[3][0]=false;  imageB[3][1]=false; imageB[3][2]=true;  imageB[3][3]=false; imageB[3][4]=false;
        imageB[4][0]=false;  imageB[4][1]=false; imageB[4][2]=false; imageB[4][3]=false; imageB[4][4]=false;
       
        for ( int i = 0; i < imageHeight; i++ ) 
        {  
            for( int j = 0; j < imageWidth; j++ )
            {
                if ( imageB[i][j] == true )
                {
                    System.out.print("true  ");
                }
                else
                {
                    System.out.print("false ");
                }
            }
            System.out.println( );
        }
       
        imageOutB = morph.dilate( 5, 5, imageB );
       
        System.out.println( "Output Image" );
       
        for ( int i = 0; i < imageHeight; i++ ) 
        {  
            for( int j = 0; j < imageWidth; j++ )
            {
                if ( imageOutB[i][j] == true )
                {
                    System.out.print( "true  " );
                }
                else
                {
                    System.out.print( "false " );
                }
            }
            System.out.println( );
        }
    }
    
    
    // Attributes
    int m_kernelWidth = 0;
    int m_kernelHeight = 0;
    int m_kernelRadius = 0;
    int m_kernelLength = 0;
    
    int[] m_kernel;
    
    
}
