/*
 * File: $Id: OrderStatisticsFilter.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 07.12.2004 Hans Slotboom  
 * Changes: 07.04.2010 a. introduce images stacks +
 *                     b. returns filtered images data
 *
 * Copyright (C) 2003-2010 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/**
 * 
 * OrderStatisticsFilter.java
 *
 * Created on December 10, 2004, 8:22 AM
 */

package org.spectrim.imageprocessing;

import org.spectrim.lib.DualProgressBar;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.LinkedList;
import javax.swing.JFrame;

/** 
 * This class implements order statistics filtering for image reconstructions:
 * 
 *  1. Minimum filter
 *  2. Maximum filter
 *  3. Median filter
 *  4. Midpoint filter
 *  5. Alpha trimmed mean filter
 *
 * @author  hans
 */
public class OrderStatisticsFilter 
{
    
    /** 
     * Creates a new instance of OrderStatisticsFilter.
     * @param listener
     */
    public OrderStatisticsFilter( ActionListener listener )
    {
        m_listeners.add( listener );
    }

    /**
     * Applies the OSF Minimum Filter to the input integer value image  x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyOSFMinimumFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                          final int imageWidth, final int imageHeight ,
                                                          final int numberOfImages,
                                                          final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying order statistic minimum filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic minimum filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyOSFMinimumFilter( x, imageWidth, imageHeight, numberOfImages,
                                                  xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }

     /**
     * Notifies all actionListeners.
     */
    private void notifyActionListeners( int eventType )
    {
        Iterator it = m_listeners.iterator( );
        while( it.hasNext( ) )
        {
            ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, eventType, "" ) );
        }
    }


    /**
     * 
     * Applies the OSF Minimum Filter to the input integer image x.
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyOSFMinimumFilter( int[] x,
                                        int imageWidth, int imageHeight, 
                                        int numberOfImages, int xFilterSize, 
                                        int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;

        
        float filterSize = xFilterSize * yFilterSize;

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic minimum filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    float minimum = Float.MAX_VALUE;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                            if ( x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ] < minimum )
                            {
                                minimum = x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            }
                        }
                    }
                
                    procImage[ ima * imageWidth * imageHeight + coord ] = Math.round( minimum );
                }
            }
        }

        return procImage;

    }


    /**
     * Applies the OSF Maximum Filter to the integer input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyOSFMaximumFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                          final int imageWidth, final int imageHeight ,
                                                          final int numberOfImages,
                                                          final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying order statistic maximum filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic maximum filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyOSFMaximumFilter( x, imageWidth, imageHeight, numberOfImages,
                                                  xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }


    /**
     * Applies the OSF maximum filter to the input integer image x.
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyOSFMaximumFilter( int[] x,
                                        int imageWidth, int imageHeight , int numberOfImages,
                                        int xFilterSize, int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float filterSize = xFilterSize * yFilterSize;
        for ( int ima = 0; ima < numberOfImages; ima++ )
        {
            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic maximum filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    float maximum = Float.MIN_VALUE;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                            if ( x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ] > maximum )
                            {
                                maximum = x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            }
                        }
                    }
                
                    procImage[ ima * imageWidth * imageHeight + coord ] = Math.round( maximum );
                }
            }
        }
        
        return procImage;
    }

    /**
     * Applies the OSF mid point filter to the input integer image x.
     * 
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyOSFMidPointFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                           final int imageWidth, final int imageHeight ,
                                                           final int numberOfImages,
                                                           final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying order statistic mid point filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic mid point filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyOSFMidPointFilter( x, imageWidth, imageHeight, numberOfImages,
                                                   xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }



    /**
     * Applies the maximum filter to the input integer image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyOSFMidPointFilter( int[] x,
                                         int imageWidth, int imageHeight , int numberOfImages,
                                         int xFilterSize, int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        float filterSize = xFilterSize * yFilterSize;
        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic mid point filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    float minimum = Float.MAX_VALUE;
                    float maximum = Float.MIN_VALUE;
                
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            if ( x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ] < minimum )
                            {
                                minimum = x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            }
                        
                            if ( x[  ima * imageWidth * imageHeight + ii + imageWidth * jj ] > maximum )
                            {
                                maximum = x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            }
                        }
                    }
                
                    procImage[ ima * imageWidth * imageHeight + coord ] = Math.round( 0.5f * ( minimum + maximum ) );
                }
            }
        }
        
        return procImage;
    }

    /**
     * Applies the OSF Median filter to the input integer image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     */
    public void applyOSFMedianFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                         final int imageWidth, final int imageHeight ,
                                                         final int numberOfImages,
                                                         final int xFilterSize, final int yFilterSize )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying order statistic median filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic median filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyOSFMedianFilter( x, imageWidth, imageHeight, numberOfImages,
                                                 xFilterSize, yFilterSize );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }



    /**
     * Applies the OSF median filter to the input integer image x.
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @return 
     */
    public int[] applyOSFMedianFilter( int[] x,
                                       int imageWidth, int imageHeight, int numberOfImages,
                                       int xFilterSize, int yFilterSize )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        QuickSort qSort = new QuickSort( );
        
        float filterSize = xFilterSize * yFilterSize;

        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic median filter to image number : " + ima + "</HTML>" );

            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    int[] values = new int[ ( 2 * xFilterSize + 1 ) * ( 2 * yFilterSize + 1 ) ];
                
                    int cntr = 0;
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                    
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            values[cntr] = x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            cntr++;
                        }
                    }
                
                    values = QuickSort.quickSort( values );
                
                    procImage[ ima * imageWidth * imageHeight + coord ] = Math.round( values[ values.length / 2 ] );
                }
            }
        }

        return procImage;
    }


    /**
     * Applies the Alpha Trimmed Mean filter to the input image x.
     * @param parent
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @param alpha
     */
    public void applyOSFAlphaTrimmedMeanFilterInSwingWorkerThread( final JFrame parent, final int[] x,
                                                                   final int imageWidth, final int imageHeight ,
                                                                   final int numberOfImages,
                                                                   final int xFilterSize, final int yFilterSize,
                                                                   final int alpha )
    {

        m_dualProgressBar = new DualProgressBar( parent, 100, 100, new Dimension( 900, 150 ), false );
        m_dualProgressBar.setTitle( "Applying order statistic alpha trimmed mean filtering"  );

        m_dualProgressBar.setUpperProgress( 0 );
        m_dualProgressBar.setUpperLabel( "        Total progress " );
        m_dualProgressBar.setUpperProgressVisible( false );
        m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic alpha trimmed mean filter to image number : " + 0 + "</HTML>" );

        final org.spectrim.lib.SwingWorker worker = new org.spectrim.lib.SwingWorker( )
        {

            @Override
            public Object construct( )
            {
                m_images = applyOSFAlphaTrimmedMeanFilter( x, imageWidth, imageHeight, numberOfImages,
                                                           xFilterSize, yFilterSize, alpha );
                return null;
            }

            // runs in the event dispatching thread
            @Override
            public void finished( )
            {
                // After calculation set enabled again
                m_dualProgressBar.setVisible( false );
                notifyActionListeners( FILTERING_APPLIED );
            }
        };

        worker.start( );
    }



    /**
     * Applies the maximum filter to the integer input image x.
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param numberOfImages
     * @param xFilterSize
     * @param yFilterSize
     * @param alpha
     * @return 
     */
    public int[] applyOSFAlphaTrimmedMeanFilter( int[] x,
                                                 int imageWidth, int imageHeight, int numberOfImages,
                                                 int xFilterSize, int yFilterSize, int alpha )
    {
        
        int[] procImage = new int[ imageWidth * imageHeight * numberOfImages ];
        int ii, jj;
        
        QuickSort qSort = new QuickSort( );
        for ( int ima = 0; ima < numberOfImages; ima++ )
        {

            m_dualProgressBar.setLowerProgress( (int)( 100.0 * (double)ima / numberOfImages ) );
            m_dualProgressBar.setLowerLabel( "<HTML><BR>     Applying order statistic alpha trimmed mean filter to image number : " + ima + "</HTML>" );


            for ( int i = 0 ; i < imageHeight ; i++ )
            {
                for ( int j = 0 ; j < imageWidth; j++ )
                {
                    // current coordinate
                    int coord = i * imageWidth + j;
                    int[] values = new int[ ( 2 * xFilterSize + 1 ) * ( 2 * yFilterSize + 1 ) ];
                
                    int cntr = 0;
                    for ( int k = j - yFilterSize ; k <= j + yFilterSize; k++ )
                    {
                    
                        for ( int l = i - xFilterSize ; l <= i + xFilterSize; l++ )
                        {
                            ii = ( k + imageWidth ) % imageWidth;
                            jj = ( l + imageHeight ) % imageHeight;
                        
                            values[cntr] = x[ ima * imageWidth * imageHeight + ii + imageWidth * jj ];
                            cntr++;
                        }
                    }
                
                    values = QuickSort.quickSort( values );
                    int sum = 0;
                    for ( int iii = alpha - 1 ; iii < values.length - alpha -1; iii++ )
                    {
                        sum += Math.round( values[ iii ] );
                    }
                
                    if ( ( values.length - 2 * alpha ) > 0 )
                    {
                        procImage[ ima * imageWidth * imageHeight + coord ] = sum / ( values.length - 2 * alpha );
                    }
                    else
                    {
                        System.out.println( "Illegal filter parameter alpha!" );
                        break;
                    }
                }
            }
        }
        
        return procImage;
    }

    /**
     * Returns the array with image pixel information.
     * @return
     */
    public int[] getFilteredImageData( )
    {
        return m_images;
    }


    // Attributes
    private DualProgressBar m_dualProgressBar;
    private int[] m_images;

    // Listeners to changes in the segmentation gui control
    private final LinkedList<ActionListener> m_listeners = new LinkedList<>( );

    public final static int FILTERING_APPLIED = 0;
    
}