/*
 * SimpleProbabilisticImageSegmentationCalculator.java
 *
 * Created on April 23, 2008, 2:55 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package org.spectrim.imageprocessing;

import org.spectrim.dicom.DicomSliceCanvasPolyDraw;
import org.spectrim.imagegraphicobjects.RegioPolygon;
import java.util.LinkedList;
import java.util.Vector;

/**
 *
 * This class handles simple probabilistic image segmentation based 
 * on user provided prior knowledge on the different tissue types to
 * discriminate.
 *   
 * @author hans
 */
public class SimpleProbabilisticImageSegmentationCalculator 
{
    

    
    /**
     * 
     * Creates a new instance of SimpleProbabilisticImageSegmentationCalculator
     * @param binaryPixelIntensityThresholdOverlay
     * @param image
     * @param probeSizeX
     * @param probeSizeY
     * @param permutations
     * @param numberOfProbingPixels
     * @param pixelRangeDeterminationType
     * @param pixelRangeMultiplier
     * @param minimumImageNumberToFocus
     * @param maximumImageNumberToFocus
     * @param nordicScalingFactor
     */
    public SimpleProbabilisticImageSegmentationCalculator( boolean[][][] binaryPixelIntensityThresholdOverlay,
                                                           DicomSliceCanvasPolyDraw image,
                                                           int probeSizeX, int probeSizeY, 
                                                           int permutations, int numberOfProbingPixels,
                                                           int pixelRangeDeterminationType,
                                                           float pixelRangeMultiplier, 
                                                           int minimumImageNumberToFocus,
                                                           int maximumImageNumberToFocus,
                                                           float nordicScalingFactor )
                                                            
    {
        
        m_binaryPixelIntensityThresholdOverlay = binaryPixelIntensityThresholdOverlay;
        m_currentView = image;
        m_nordicScalingFactor = nordicScalingFactor;

        m_minimumImageNumberToFocus = minimumImageNumberToFocus;
        m_maximumImageNumberToFocus = maximumImageNumberToFocus;

        m_tissueTypesPriorPolygonList = m_currentView.getPolygonList( );
        
        m_probeSizeX = probeSizeX;
        m_probeSizeY = probeSizeY;
        m_permutations = permutations;
        m_numberOfProbingPixels = numberOfProbingPixels;
        
        m_imageWidth = (int)( image.getWidth( ) / image.getZoom( ) );
        m_imageHeight = (int)( image.getHeight( ) / image.getZoom( ) );
        m_numberOfImages = image.getNumberOfImages( );
        m_currentImageNumber = image.getCurrentlyDisplayedImageNumber( );
        m_zoom = image.getZoom( );
        
        m_pixelRangeDeterminationType = pixelRangeDeterminationType;
        m_pixelRangeMultiplier = pixelRangeMultiplier;
    }
     
    
    
    /**
     *  Returns the int-value of the most probable tissue type .
     */
    public float getProbabilityForTissueType( int tissueType, float[] testArray )
    {
        float pVal = -1f;
        
        if ( m_regionTTesters == null )
        {
            initializeTTesters( );
            pVal = m_regionTTesters.get( tissueType ).calculateTTest( testArray );
        }
        else
        {
            // compute the TTest for a region around pixel(i,j) defined by 
            // "testArray" with m_regionTTesters[tissueType]
            pVal = m_regionTTesters.get( tissueType ).calculateTTest( testArray );
            System.out.println( "Tissuetype  : " + tissueType + " pVal = " + pVal  );
            
        }
        return pVal;
    }
    
    /**
     * Computes the array of float-value located near (i,j) 
     */
    public float[] computeTestArray( int posX, int posY )
    {
        
        float[] retArray = new float[9]; 
        int[] image = m_currentView.getRawData( );
        
        int arrayIndex = 0;
        int counter = 0;
        
        for ( int i = posX - 1; i <= posX + 1; i++ )
        {
            for ( int j = posY - 1; j <= posY + 1; j++ )
            {
                arrayIndex = m_imageWidth * j + i;
                retArray[counter] = image[arrayIndex];
                counter++;
            }
        }
        
        return retArray;
    }
    
    /**
     * Initializes our battery of Region TTester objects
     */
    private void initializeTTesters( )
    {
        
        m_regionTTesters = new Vector<RegionTTester>( );
        
        for ( int i = 0 ; i < m_tissueTypesPriorPolygonList.size( ) ; i++ )
        {
            RegioPolygon regPol = m_tissueTypesPriorPolygonList.get( i );
            
            RegionTTester regTTest = new RegionTTester( regPol, m_binaryPixelIntensityThresholdOverlay , 
                                                        m_currentView.getRawData( ), 
                                                        m_imageWidth, m_imageHeight,  
                                                        m_zoom, m_probeSizeX, m_probeSizeY,
                                                        m_permutations, m_numberOfProbingPixels, m_nordicScalingFactor );
            m_regionTTesters.add( regTTest );
        }
    }
    
    
    /**
     * Performs the segmentation for tissue type having name "tissueName"
     */ 
    public boolean[][] performSegmentationForSpecificTissueType( String tissueName )
    {
        int selectedTissueType = -1;
        boolean[][] imageSegment = null;
        RegioPolygon selectedRegPol = null;
        
        for ( int i = 0 ; i < m_tissueTypesPriorPolygonList.size( ) ; i++ )
        {
            RegioPolygon regPol = m_tissueTypesPriorPolygonList.get( i );
            if ( tissueName.compareTo( regPol.getLabel( ) ) == 0 )
            {
                selectedTissueType = i;
                selectedRegPol = regPol; 
                break;
            }
        }
        
        RegionTTester regTTest = new RegionTTester( selectedRegPol, m_binaryPixelIntensityThresholdOverlay , 
                                                    m_currentView.getRawData( ), 
                                                    m_imageWidth, m_imageHeight,  
                                                    m_zoom, m_probeSizeX, m_probeSizeY,
                                                    m_permutations, m_numberOfProbingPixels, m_nordicScalingFactor );

        m_pValueMap = regTTest.computeTissueSpecificProbabilityMaps( m_currentImageNumber, RegionTTester.FULL_PROBING_ROIS );
        
        m_tValueMap = regTTest.getTMap( );
        
        imageSegment = new boolean[m_imageWidth][m_imageHeight];
        
        for ( int i = 0; i < m_imageWidth; i++ )
        {
            for ( int j = 0; j < m_imageHeight; j++ )
            {
                if ( m_pValueMap[i][j] > 0.0f )
                {
                    imageSegment[i][j] = true;
                }
            }
        }
        
        return imageSegment;
    }
    
    /**
     * Returns the map (image) pixel values of the probabilities
     */
    float[][] getPValueMap( )
    {
        return m_pValueMap;
    }
    
    /**
     * Returns the map (image) pixel values of the probabilities
     */
    float[][] getTValueMap( )
    {
        return m_tValueMap;
    }
    
    /**
     *  Classifies tissues based on the mean and variances of tissue priors
     *  defined by the user
     * 
     * THIS ROUTINE MUST BE FURTHER DEVELOPED!! HS 20.10.2008
     * 
     */
    public boolean[][] getSegmentedTissueMapWithSimpleClassifier( int sliceNumber, String tissueName )
    {
        int selectedTissueType = -1;
        boolean[][] imageSegment = null;
        RegioPolygon selectedRegPol = null;
        
        for ( int i = 0 ; i < m_tissueTypesPriorPolygonList.size( ) ; i++ )
        {
            RegioPolygon regPol = m_tissueTypesPriorPolygonList.get( i );
            if ( tissueName.compareTo( regPol.getLabel( ) ) == 0 )
            {
                selectedTissueType = i;
                selectedRegPol = regPol; 
                break;
            }
        }
        
        RegionTTester regTTest = new RegionTTester( selectedRegPol, m_binaryPixelIntensityThresholdOverlay , 
                                                    m_currentView.getRawData( ), 
                                                    m_imageWidth, m_imageHeight,  
                                                    m_zoom, m_probeSizeX, m_probeSizeY,
                                                    m_permutations, m_numberOfProbingPixels, 
                                                    m_nordicScalingFactor );
        
        
        float centerPixelValue;
        float pixelValueRange;
        
        if ( m_pixelRangeDeterminationType == ProbingRegionPropertiesPanel.PIXEL_RANGE_DEFINED_BY_STANDARD_DEVIATION )
        {
            centerPixelValue = regTTest.getMeanOfRegion( );
            pixelValueRange = (float)Math.sqrt( regTTest.getSampleVarianceOfRegion( ) );
        }
        else
        {
            centerPixelValue = regTTest.getCenterValueOfRegion( );
            pixelValueRange =  regTTest.getMinimumMaximumRangeOfRegion( );
        }
        
        imageSegment = new boolean[m_imageWidth][m_imageHeight];
        
        int[] image = m_currentView.getImagePixelData( );
                
        int arrayIndex = 0;
        
        for ( int i = 0; i < m_imageWidth; i++ )
        {
            for ( int j = 0; j < m_imageHeight; j++ )
            {
                arrayIndex = sliceNumber * m_imageWidth * m_imageHeight + m_imageWidth * j + i;
                if ( m_binaryPixelIntensityThresholdOverlay[m_currentImageNumber][i][j] == true )
                {
                    if ( ( image[arrayIndex] <= centerPixelValue + m_pixelRangeMultiplier * pixelValueRange ) &&
                         ( image[arrayIndex] >= centerPixelValue - m_pixelRangeMultiplier * pixelValueRange ) )
                    {
                        imageSegment[i][j] = true;
                    }
                }
            }
        }
        
        return imageSegment;
    }
    
    
    /**
     *  Classifies tissues based on the mean and variances of tissue priors
     *  defined by the user
     * @param sliceNumber
     * @param tissueProperties
     * @return 
     */
    public boolean[][] getSegmentedTissueMapWithSimpleClassifier( int sliceNumber, ImageParametersWithinARegioPolygon tissueProperties )
    {

        boolean[][] imageSegment = null;

        
        // float centerPixelValue = tissueProperties.getMeanPixelValue( );
        // float sampleStDev = tissueProperties.getStandardDeviationOfPixelValues( );
        
        
        float centerPixelValue;
        float pixelValueRange;
        
        if ( m_pixelRangeDeterminationType == ProbingRegionPropertiesPanel.PIXEL_RANGE_DEFINED_BY_STANDARD_DEVIATION )
        {
            centerPixelValue = tissueProperties.getMeanRawPixelValue( );
            pixelValueRange =  tissueProperties.getStandardDeviationRawOfPixelValues( );
        }
        else
        {
            centerPixelValue = tissueProperties.getCenterPixelRawRangeValueOfRegion( );
            pixelValueRange  = tissueProperties.getMinimumMaximumRawRangeOfRegion( );
        }
        
        imageSegment = new boolean[m_imageWidth][m_imageHeight];
        
        int[] image = m_currentView.getImagePixelData( );
                
        int arrayIndex = 0;
        
        for ( int i = 0; i < m_imageWidth; i++ )
        {
            for ( int j = 0; j < m_imageHeight; j++ )
            {
                arrayIndex = sliceNumber * m_imageWidth * m_imageHeight + m_imageWidth * j + i;
                if ( m_binaryPixelIntensityThresholdOverlay[m_currentImageNumber][i][j] == true )
                {
                    
                    if ( ( image[arrayIndex] <= centerPixelValue + m_pixelRangeMultiplier * pixelValueRange ) &&
                         ( image[arrayIndex] >= centerPixelValue - m_pixelRangeMultiplier * pixelValueRange ) )
                    {
                        imageSegment[i][j] = true;
                    }
                }
            }
        }
        
        return imageSegment;
    }

    
    // Attributes of this class
    private LinkedList<RegioPolygon> m_tissueTypesPriorPolygonList;  // contains the polygons enclosing tissues of different types
    private int m_probeSizeX, m_probeSizeY;
    private DicomSliceCanvasPolyDraw m_currentView;

    private int m_permutations;

    private int m_numberOfProbingPixels;
    private boolean[][][] m_binaryPixelIntensityThresholdOverlay;
    private float[][]   m_pValueMap;
    private float[][]   m_tValueMap;
    
    // we need a battery of TTesters in order to be able to test quickly 
    // of which type a pixel is..
    private Vector<RegionTTester> m_regionTTesters;
    
    private int m_imageWidth;
    private int m_imageHeight;
    private int m_numberOfImages;
    private int m_currentImageNumber;
    
    private float m_zoom;
    
    // determines the way the pixel intensity range should be determined
    private int   m_pixelRangeDeterminationType;
    private float m_pixelRangeMultiplier;


    private int m_minimumImageNumberToFocus = -1;
    private int m_maximumImageNumberToFocus = -1;
    
    private float m_nordicScalingFactor = 1f;
    
}
