/*
 * File: $Id: RegFile.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 30. Januar 2002, 12:47  Authors:  Martin Buerki & Hans Slotboom
 *
 * Changes:
 *              24-04-2003 moved from package Regio to  package Lib
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.lib;

import java.io.RandomAccessFile;
import java.io.IOException;

import java.util.Date;
import java.util.StringTokenizer;

import org.spectrim.imagegraphicobjects.RegioPolygon;


/**
 * 
 * The original regio file in which region polygons are stored in a file.
 * 
 * @author Johannes Slotboom <johannes.slotboom at insel.ch>
 */
public class RegFile 
{

    public RegHeaderStruct hdr;
    private RandomAccessFile f;
    private String fileMode = "";

    /**
     * Constructor of the .reg file type.
     * @param fname
     * @param w
     * @param h 
     */
    public RegFile( String fname, int w, int h ) 
    {
        hdr = new RegHeaderStruct( fname, w, h );
    }

    /**
     * Opens the .reg file in io-mode 'mode'.
     * @param mode
     * @return 
     */
    public boolean open( String mode ) 
    {
        try
        {
            close( );
            f = new RandomAccessFile( hdr.fileName, mode );
        }
        catch ( IOException e )
        {
            System.out.println( "File " + hdr.fileName + " not found" );
            return false;
        }
        catch ( IllegalArgumentException e )
        {
            System.out.println( "Illegal file mode: " + mode );
            return false;
        }
        fileMode = mode;
        return true;
    }

    /**
     * Closes the .reg file.
     */
    public void close( )
    {
        if ( f == null ) return;
        try
        {
            f.close( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
        f = null;
        fileMode = "";
    }

    /**
     * Writes the header of the .reg file.
     */
    public void writeHeader( )
    {
        if ( f == null ) return;
        try
        {
            f.seek(0);
            f.writeBytes( "#$regions$\n");
            f.writeBytes( "# " + hdr.fileName + "\n" );
            Date date = new Date( );
            f.writeBytes( "# automatically created at " + date.toString() + "\n" );
            //Process proc = Runtime.getRuntime().exec("/bin/hostname");
            //BufferedReader in = new BufferedReader(new InputStreamReader(proc.getInputStream()));
            //String machine = in.readLine();
            String machine = "<machine>";
            f.writeBytes( "# on " + machine + " " +
                          System.getProperty( "os.name" ) + " " +
                          System.getProperty( "os.arch" ) +
                          " from " +
                          System.getProperty("user.name") + "\n" );
            f.writeBytes( "# \n");
            f.writeBytes( "imagefile=" + hdr.fileName + "\n" );
            f.writeBytes( "width=" + hdr.imageWidth + "\n" );
            f.writeBytes( "height=" + hdr.imageHeight + "\n" );
            f.writeBytes( "normal={-0|0|1}\n" );
            f.writeBytes( "pos={0|0|<slice pos>}\n" );
            f.writeBytes( "slice=<slice id>\n" );
            f.writeBytes( "# \n" );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }
    
    /**
     * Writes the RegioPolygon information to the .reg file.
     * @param p
     * @param zoom 
     */
    public void writePolygonData( RegioPolygon p, float zoom ) 
    {
        if ( f == null ) return;
        try
        {
            f.writeBytes( "# \n" );
            f.writeBytes( "regio=region " + p.m_index + ":::\n" );
            f.writeBytes( "label=" + p.m_label + ":::\n" );
            f.writeBytes( "Poly" );
            for ( int k=0; k < p.npoints; k++ )
            {
                f.writeBytes( "{" + (int)( p.xpoints[k] / zoom ) + ";" + (int)( p.ypoints[k] / zoom ) + "}" );
                if ( k % 8 == 7 ) f.writeBytes( "\n\t" );
            }
            f.writeBytes( "\n\n" );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }

    /**
     * Finishes the file-IO.
     * @return 
     */
    public boolean finish( )
    {
        if ( f == null ) return false;
        try
        {
            f.writeBytes( "# end\n" );
            close( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
            return false;
        }
        return true;
    }

    /**
     * Reads the header of the .reg file.
     */
    public void readHeader( )
    {
        if ( f == null ) return;
        String line;
        try
        {
            f.seek( 0 );
            skipComments( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }

    /**
     * Reads the RegioPolygon information from the .reg file, directly zooming
     * the RegioPolygon object with the required zooming factor.
     * @param zoom
     * @return 
     */
    public RegioPolygon readPolygonData( float zoom ) 
    {
        RegioPolygon p;
        String line;
        try
        {
            String indexString = gotoNextRegioEntry( );
            
            String polygonLabel = readRegioLabel( );
            
            if ( indexString == null ) return null;
            int index = Integer.parseInt(indexString.substring( 7 ) );
            p = new RegioPolygon( index );
            
            p.setLabel( polygonLabel );
            
            while ( true )
            {
                line = f.readLine( );
                if ( line.trim( ).equals( "" ) ) break;
                StringTokenizer sTok = new StringTokenizer( line.substring(line.indexOf("{")), "{};" );
                while ( sTok.hasMoreTokens() )
                {
                    int x = Integer.parseInt( sTok.nextToken( ) );
                    int y = Integer.parseInt( sTok.nextToken( ) );
                    p.addPoint( (int)( x * zoom ), (int)( y * zoom ) );
                }
            }
        }
        catch ( IOException e )
        {
            System.out.println( e );
            return ( RegioPolygon )null;
        }
        if ( p.npoints == 0 ) return null;
        else return p;
    }
    
    /**
     * Skips all lines starting with a '#' - 
     * these lines are comments.
     */
    private void skipComments( ) throws IOException 
    {
        String line;
        do
        {
            line = f.readLine( );
        }
        while ( line.startsWith( "#" ) );
    }
    
    /**
     * Returns the line starting with 'regio='.
     */
    private String gotoNextRegioEntry( ) throws IOException 
    {
        String line;
        do
        {
            line = f.readLine( );
        } while ( line != null && !line.startsWith( "regio=" ) );

        if ( line == null )
        {
            return line;
        }
        else
        {
            return line.substring( line.indexOf( "=" ) + 1, line.indexOf( ":::" ) );
        }
    }
    
    /**
     * Reads the line with contains regio m_label information.
     */
    private String readRegioLabel( ) throws IOException 
    {
        String line;
        do
        {
            line = f.readLine( );
        }
        while ( line != null && !line.startsWith( "label=" ) );
        if ( line == null )
        {
            return line;
        }
        else
        {
            return line.substring( line.indexOf( "=" ) + 1, line.indexOf( ":::" ) );
        }
    }
}

class RegHeaderStruct 
{
    String fileName;
    int imageWidth;
    int imageHeight;

    public RegHeaderStruct( String name, int w, int h ) 
    {
        fileName = name;
        imageWidth = w;
        imageHeight = h;
    }
}
