/*
 * File: $Id: Tiff.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 30.11.2002 Authors: Martin Buerki  Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.lib;

/*
 * Tiff image struct declaration
 *
 */

import java.io.RandomAccessFile;
import java.io.IOException;

import java.awt.Color;

public class Tiff 
{

    /**
     *
     * return values
     *
     */

    public final static int TIFF_UNKNOWNTAG_WARN = -2;
    public final static int TIFF_MULTIIMAGE_WARN = -1;
    public final static int TIFF_OK	= 0;
    public final static int TIFF_IO_ERROR = 1;
    public final static int TIFF_NOT_A_TIFF = 2;
    public final static int TIFF_NOT_SUPPORTED = 3;
    public final static int TIFF_FILE_CORRUPT = 4;
    public final static int TIFF_TAG_ERROR = 5;
    public final static int TIFF_INCONSISTENT_DATA = 6;
    

    /**
     * constants
     *
     */
    
    public final static int TAG_TYPE_BYTE = 1;
    public final static int TAG_TYPE_ASCII = 2;
    public final static int TAG_TYPE_SHORT = 3;
    public final static int TAG_TYPE_LONG = 4;
    public final static int TAG_TYPE_RATIONAL = 5;
    public final static int TAG_TYPE_SBYTE = 6;
    public final static int TAG_TYPE_UNDEFINED = 7;
    public final static int TAG_TYPE_SSHORT = 8;
    public final static int TAG_TYPE_SLONG = 9;
    public final static int TAG_TYPE_SRATIONAL = 10;
    public final static int TAG_TYPE_FLOAT = 11;
    public final static int TAG_TYPE_DOUBLE = 12;

    public final static int COMPRESSION_NONE = 1;
    public final static int COMPRESSION_LZW = 2;
    public final static int COMPRESSION_PACKBITS = 32773;
    
    public final static int TIFF_COLORTYPE_WHITE = 0;
    public final static int TIFF_COLORTYPE_BLACK = 1;
    public final static int TIFF_COLORTYPE_RGB = 2;
    public final static int TIFF_COLORTYPE_PALETTE = 3;
    
    public final static int TIFF_UNIT_NONE = 1;
    public final static int TIFF_UNIT_INCH = 2;
    public final static int TIFF_UNIT_CM = 3;

    public final static int HDR_SIZE = 8;
    public final static int TAG_SIZE = 12;

    public final static int SHORT_SIZE = 2;
    public final static int RATIO_SIZE = 8;

    /**
     * structures
     *
     */

    public class TiffHdr 
    {
	String endian;
	int magic;
	long ifdOffset;
    }

    public class TiffTag 
    {
	int tag;
	int type;
	long numVals;
	long values;
    }

	
    public class Ratio 
    {
	long numerator;
	long denominator;
    }


    public class TiffInfo 
    {
	short isInitialized;
	short numStrips;
	long[] stripOffsets;
	long[] stripByteCounts;
	long rowsPerStrip;
    }

    /**
     *
     * member variables
     *
     */

    RandomAccessFile f;

    int width;
    int height;
    short depth;			// Bits per sample
    short samplesPerPixel;
    short compression;
    short colortype;			// photometric interpretation
    StringBuffer description;
    Ratio xResolution;
    Ratio yResolution;
    StringBuffer resolutionUnit;	// "none","inch" or "cm  "
    int numColors;
    short bigendian;			// 0 or 1
    Color[] color;
    int[] data;

    TiffInfo tinf;
    TiffTag tag;
    TiffHdr tHdr;
    short type_size[] = { 0, 1, 1, 2, 4, 8 };  // for types dummy(0),BYTE(1),ASCII(2),SHORT(3),LONG(4),RATIONAL(5)

    // Constructor(s)
    public Tiff( ) 
    {
	xResolution = new Ratio( );
	xResolution.numerator = 1;
	xResolution.denominator = 1;
	yResolution = new Ratio( );
	yResolution.numerator = 1;
	yResolution.denominator = 1;
	depth = 8;
	resolutionUnit = new StringBuffer( "cm  " );
    }

    public void setImage( int[] d, int width, int height ) 
    {
	this.width = width;
	this.height = height;
	if ( data == null || data.length != d.length ) 
	    data = new int[d.length];
	System.arraycopy( d, 0, data, 0, data.length );
    }
    
    public void setColormap( Color[] c ) 
    {
	color = c.clone( );
	numColors = color.length;
	setBitsPerSample( );
    }

    public void setColormap( String type, int size ) 
    {
	if ( type.toLowerCase( ).equals( "gray" ) ) 
        {
	    color = new Color[size];
	    float d = 1f / (float)(size - 1);
	    for ( int k = 0; k < size; k++ )
            {
		float v = k * d;
		color[k] = new Color( v, v, v );
	    }
	}
	numColors = size;
	setBitsPerSample( );
    }

    private void setBitsPerSample( ) 
    {
	if ( numColors > 16 )
        { 
            depth = 8;
        }
	else 
        { 
            depth = 4;
        }
    }

    public void setXResolution( long numerator, long denominator ) 
    {
	xResolution.numerator = numerator;
	xResolution.denominator = denominator;
    }

    public void setYResolution( long numerator, long denominator ) 
    {
	yResolution.numerator = numerator;
	yResolution.denominator = denominator;
    }

    public int writeGrayLevel( String filename, int numLevels ) 
    {

	/* save in tiff with the simplest possible structure, i.e.
	   byte	data
	   ------------------------------------------------
	   0   - 7	header
	   8   - 9	number of tags (=11)
	   10  - 153	11 tags
	   154 - 169	two rationals for X and Y resolution
	   170 - eof	image data
	*/

	if ( numLevels != 256 ) 
        {
	    System.out.println( "Warning: only 256 gray levels supported." );
	    numColors = 256;
	}
	else numColors = numLevels;

	setBitsPerSample( );
	int status = TIFF_OK;
	short numTags = 11;
	
	// file header
	tHdr = new TiffHdr( );
	tHdr.endian = "MM";		// number format is bigendian
	tHdr.magic = 42;		// the magic number for tiff
	tHdr.ifdOffset = 8;		// the first ifd follows immediately
	// after the header
	
	long tagOffset = tHdr.ifdOffset + SHORT_SIZE;
	long ratio1Offset = 4 + tagOffset + numTags * TAG_SIZE;
	long ratio2Offset = ratio1Offset + RATIO_SIZE;
	long dataOffset = ratio2Offset + RATIO_SIZE;
	long imsize = width * height;

	// create file
	try
        {
	    f = new RandomAccessFile( filename, "rw" );
	    // write the header
	    f.writeBytes( tHdr.endian );
	    f.writeChar( tHdr.magic );
	    fwriteLong( tHdr.ifdOffset );
	    f.writeChar( numTags );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException at file creation: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}
	
	// the needed tags
	if ( tag == null ) tag = new TiffTag( );

	// image width
	tag.tag = 256;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = width;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	
	// image height
	tag.tag = 257;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = height;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// bits per sample
	tag.tag = 258;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = depth;
	if ( !saveTag( ) ) 
        { 
            System.out.println( "Error writing tag " + tag.tag );
        }

	// image compression
	tag.tag = 259;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = COMPRESSION_NONE;
	
        if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	
	// photometric interpretation
	tag.tag = 262;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values =  TIFF_COLORTYPE_BLACK;
	
        if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// strip offsets
	tag.tag = 273;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values =  dataOffset;
	
        if ( !saveTag())
        { 
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// rows per strip
	tag.tag = 278;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = (long)height;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// strip byte counts
	tag.tag = 279;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = imsize;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// x resolution
	tag.tag = 282;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio1Offset;
	if ( !saveTag() ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// y resolution
	tag.tag = 283;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio2Offset;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	
        // resolution units
	tag.tag = 296;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	if ( resolutionUnit.equals( "inch" ) )
        {
	    tag.values = TIFF_UNIT_INCH;
        }
	else if ( resolutionUnit.equals( "cm  " ) )
        {
	    tag.values = TIFF_UNIT_CM;
        }
	else
        {
	    tag.values = TIFF_UNIT_NONE;
        }
        
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// no more IFDs
	fwriteLong( 0l );

	fwriteLong( xResolution.numerator );
	fwriteLong( xResolution.denominator );
	fwriteLong( yResolution.numerator );
	fwriteLong( yResolution.denominator );

	try 
        {
	    for ( int k = 0; k < imsize; k++ )
            {
                f.writeByte( data[k] );
            }
	    f.close( );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException while writing data: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}

	return status;
    }

    public int writePalette( String filename ) 
    {
	/* save in tiff with the simplest possible structure, i.e.
	   byte	data
	   ------------------------------------------------
	   0   - 7	header
	   8   - 9	number of tags (=12)
	   10  - 153	12 tags
	   154 - 169	two rationals for X and Y resolution
	   170 - ???	colormap
	   ??? - eof	image data
	*/
	
	int status = TIFF_OK;
	short numTags = 12;
	
	// file header
	tHdr = new TiffHdr( );
	tHdr.endian = "MM";		// number format is bigendian
	tHdr.magic = 42;		// the magic number for tiff
	tHdr.ifdOffset = 8;		// the first ifd follows immediately
	// after the header
	
	long tagOffset = tHdr.ifdOffset + SHORT_SIZE;
	long ratio1Offset = 4 + tagOffset + numTags * TAG_SIZE;
	long ratio2Offset = ratio1Offset + RATIO_SIZE;
	long cmapOffset = ratio2Offset + RATIO_SIZE;
	long dataOffset = cmapOffset + 3 * SHORT_SIZE * numColors;

	long imsize = width * height;

	// the colormap
	depth = (short)Math.ceil( Math.log( (double)numColors) / Math.log( 2.0 ) );
	int cmsize = 3 << depth;
	if ( cmsize != 3 * numColors ) 
        {
	    System.out.print( "colormap size mismatch..." );
	    return TIFF_INCONSISTENT_DATA;
	}
	int[] cmap = new int[cmsize];
	
	for ( int k = 0; k < numColors; k++ ) 
        {
	    cmap[k] = color[k].getRed( ) * 256;
	    cmap[numColors+k] = color[k].getGreen( ) * 256;
	    cmap[2*numColors+k] = color[k].getBlue( ) * 256;
	}
	
	// create file
	try 
        {
	    f = new RandomAccessFile( filename, "rw" );
	    // write the header
	    f.writeBytes( tHdr.endian );
	    f.writeChar( tHdr.magic );
	    fwriteLong( tHdr.ifdOffset );
	    f.writeChar( numTags );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException at file creation: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}
	
	// the needed tags
	if ( tag == null )
        {
            tag = new TiffTag( );
        }
        
	// image width
	tag.tag = 256;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = width;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// image height
	tag.tag = 257;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = height;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// bits per sample
	tag.tag = 258;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = depth;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	
        // image compression
	tag.tag = 259;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = COMPRESSION_NONE;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// photometric interpretation
	tag.tag = 262;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values =  TIFF_COLORTYPE_PALETTE;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// strip offsets
	tag.tag = 273;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values =  dataOffset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// rows per strip
	tag.tag = 278;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = (long)height;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// strip byte counts
	tag.tag = 279;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = imsize;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// x resolution
	tag.tag = 282;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio1Offset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// y resolution
	tag.tag = 283;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio2Offset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// resolution units
	tag.tag = 296;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	if ( resolutionUnit.equals( "inch" ) )
	    tag.values = TIFF_UNIT_INCH;
	else if ( resolutionUnit.equals( "cm  " ) )
	    tag.values = TIFF_UNIT_CM;
	else
	    tag.values = TIFF_UNIT_NONE;
	if ( !saveTag( ) ) 
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	//colormap
	tag.tag = 320;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = cmsize;
	tag.values = cmapOffset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// no more IFDs
	fwriteLong(0l);

	fwriteLong( xResolution.numerator );
	fwriteLong( xResolution.denominator );
	fwriteLong( yResolution.numerator );
	fwriteLong( yResolution.denominator );

	try 
        {
	    for ( int k=0; k < cmsize; k++ )
            {
                f.writeChar( cmap[k] );
            }
	    for ( int k=0; k < imsize; k++ )
            {
                f.writeByte( data[k] );
            }
	    f.close( );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException while writing data: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}

	return status;
	
    }

    public int writeFullRGB( String filename ) 
    {
	
	/* save in tiff with the simplest possible structure, i.e.
	   byte	data
	   ------------------------------------------------
	   0   - 7	header
	   8   - 9	number of tags (=12)
	   10  - 153	12 tags
	   154 - 169	two rationals for X and Y resolution
	   170 - eof	image data
	*/
	
	int status = TIFF_OK;
	short numTags = 12;
	
	// file header
	tHdr = new TiffHdr( );
	tHdr.endian = "MM";		// number format is bigendian
	tHdr.magic = 42;		// the magic number for tiff
	tHdr.ifdOffset = 8;		// the first ifd follows immediately
					// after the header
	
	long tagOffset = tHdr.ifdOffset + SHORT_SIZE;
	long bitsPerSampleOffset = tagOffset + numTags*TAG_SIZE + 4;
	long ratio1Offset = bitsPerSampleOffset + 3*SHORT_SIZE;
	long ratio2Offset = ratio1Offset + RATIO_SIZE;
	long dataOffset = ratio2Offset + RATIO_SIZE;

	long imsize = width * height;
	
	// create file
	try 
        {
	    f = new RandomAccessFile( filename, "rw" );
	    // write the header
	    f.writeBytes( tHdr.endian );
	    f.writeChar( tHdr.magic );
	    fwriteLong( tHdr.ifdOffset );
	    f.writeChar( numTags );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException at file creation: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}
	
	// the needed tags
	if ( tag == null )
        {
            tag = new TiffTag( );
        }
        
	// image width
	tag.tag = 256;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = width;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// image height
	tag.tag = 257;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = height;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// bits per sample
	tag.tag = 258;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 3;
	tag.values = bitsPerSampleOffset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// image compression
	tag.tag = 259;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = COMPRESSION_NONE;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// photometric interpretation
	tag.tag = 262;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values =  TIFF_COLORTYPE_RGB;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// strip offsets
	tag.tag = 273;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values =  dataOffset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	
        // samples per pixel
	tag.tag = 277;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	tag.values = 3;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// rows per strip
	tag.tag = 278;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = (long)height;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// strip byte counts
	tag.tag = 279;
	tag.type = TAG_TYPE_LONG;
	tag.numVals = 1;
	tag.values = imsize*3;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
        
	// x resolution
	tag.tag = 282;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio1Offset;
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// y resolution
	tag.tag = 283;
	tag.type = TAG_TYPE_RATIONAL;
	tag.numVals = 1;
	tag.values = ratio2Offset;
	if (!saveTag()) System.out.println( "Error writing tag " + tag.tag );
	
	// resolution units
	tag.tag = 296;
	tag.type = TAG_TYPE_SHORT;
	tag.numVals = 1;
	if ( resolutionUnit.equals( "inch" ) )
        {
	    tag.values = TIFF_UNIT_INCH;
        }
	else if ( resolutionUnit.equals( "cm  " ) )
        {
	    tag.values = TIFF_UNIT_CM;
        }
	else
        {
	    tag.values = TIFF_UNIT_NONE;
        }
        
	if ( !saveTag( ) )
        {
            System.out.println( "Error writing tag " + tag.tag );
        }
	// no more IFDs
	fwriteLong(0l);

	// the values for BitsPerSample
	for ( int k=0; k<3; k++ )
        {
            fwriteShort(8);
        }
        
	// the values for the x- and y-resolution
	fwriteLong( xResolution.numerator );
	fwriteLong( xResolution.denominator );
	fwriteLong( yResolution.numerator );
	fwriteLong( yResolution.denominator );

	try 
        {
	    for ( int k = 0; k < imsize; k++ ) 
            {
		f.writeByte( ( data[k] & 0x00FF0000) >> 16 );
		f.writeByte( ( data[k] & 0x0000FF00) >> 8 );
		f.writeByte( ( data[k] & 0x000000FF ) );
	    }
	    f.close( );
	}
	catch ( IOException e ) 
        {
	    System.out.println( "IOException while writing data: " + e.toString( ) );
	    return TIFF_IO_ERROR;
	}

	return status;
    }

    private boolean saveTag( ) 
    {
	boolean status;
	status = fwriteShort( tag.tag );
	status &= fwriteShort( tag.type );
	status &= fwriteLong( tag.numVals );
	switch ( tag.type ) 
        {
            case TAG_TYPE_SHORT:
                if ( tag.numVals == 1 ) 
                {
                    status &= fwriteShort( (int)tag.values );
                    status &= fwriteShort( 0 );
                }
                else status &= fwriteLong( tag.values );
                break;
            case TAG_TYPE_LONG:
                status &= fwriteLong( tag.values );
                break;
            case TAG_TYPE_RATIONAL:
                status &= fwriteLong( tag.values );
                break;
	}
	return status;
    }

    private boolean fwriteShort( int v ) 
    {
	try
        {
	    f.writeChar( v );
	}
	catch ( IOException e ) 
        {
	    return false;
	}
	return true;
    }

    private boolean fwriteLong( long v ) 
    {
	try
        {
	    f.writeChar( (int)( v & 0xFFFFFFFF ) >> 16 );
	    f.writeChar( (int)( v & 0xFFFF ) );
	}
	catch ( IOException e ) 
        {
	    return false;
	}
	return true;
    }

    private short swap_short( short number ) 
    {
   
	short lo,hi;
   
	lo = (short)( (number & 0xFF00 ) >> 8 );
	hi = (short)( (number & 0x00FF ) << 8 );
	
	return (short)(hi | lo);
    }

    private long swap_long( long number )
    {

	long b0, b1, b2, b3;
    
	b0 = ( number & 0xFF000000 ) >> 24;
	b1 = ( number & 0x00FF0000 ) >> 8;
	b2 = ( number & 0x0000FF00 ) << 8;
	b3 = ( number & 0x000000FF ) << 24;
	
	return (b3 | b2 | b1 | b0);
    }

    private short swap_string( StringBuffer text ) 
    {

	text.reverse( );

	return 0;
    }


    private int read_tag( long offset ) 
    {

// 	short asOffset = 0;
// 	int k;
// 	short *reds,*greens,*blues;
	
// 	fseek(f,offset,SEEK_SET);
// 	fread(&tt, sizeof(tt), 1, f);
// 	if (!image->bigendian) {
// 	    tt.tag = swap_short(tt.tag);
// 	    tt.type = swap_short(tt.type);
// 	    tt.numVals = swap_long(tt.numVals);
// 	    tt.values = swap_long(tt.values);
// 	}
	
// 	if (tt.numVals*type_size[tt.type] > 4) asOffset = 1;
	
// 	switch (tt.tag) {
// 	case 256 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) image->width = (tt.values & 0xFFFF0000) >> 16;
// 		else image->width = tt.values & 0x0000FFFF;
// 	    }
// 	    else if (tt.type==4) image->width = tt.values;
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 257 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) image->height = (tt.values & 0xFFFF0000) >> 16;
// 		else image->height = tt.values & 0x0000FFFF;
// 	    }
// 	    else if (tt.type==4) image->height = tt.values;
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 258 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) image->depth = (tt.values & 0xFFFF0000) >> 16;
// 		else image->depth = tt.values & 0x0000FFFF;
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 259 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) image->compression = (tt.values & 0xFFFF0000) >> 16;
// 		else image->compression = tt.values & 0x0000FFFF;
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 262 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) image->colortype = (tt.values & 0xFFFF0000) >> 16;
// 		else image->colortype = tt.values & 0x0000FFFF;
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 273 :
// 	    if (ti->isInitialized) {
// 		if (tt.numVals != ti->numStrips) return TIFF_TAG_ERROR;
// 	    }
// 	    else {
// 		ti->numStrips = tt.numVals;
// 		ti->stripOffsets = (long int*) malloc(tt.numVals*sizeof(long int));
// 		ti->stripByteCounts = (long int*) malloc(tt.numVals*sizeof(long int));
// 		ti->isInitialized = 1;
// 	    }
// 	    if (tt.type==3) {
// 		if (asOffset) {
// 		    fseek(f,tt.values,SEEK_SET);
// 		    for (k=0; k<tt.numVals; k++) {
// 			fread(ti->stripOffsets+k, sizeof(short), 1, f);
// 			if (image->bigendian) ti->stripOffsets[k] = swap_short((short) ti->stripOffsets[k]);
// 		    }
// 		}
// 		else {
// 		    ti->stripOffsets[0] = (tt.values & 0xFFFF0000) >> 16;
// 		    if (tt.numVals>1) ti->stripOffsets[1] = tt.values & 0x0000FFFF;
// 		}
// 	    }
// 	    else if (tt.type==4) {
// 		if (asOffset) {
// 		    fseek(f,tt.values,SEEK_SET);
// 		    fread(ti->stripOffsets, sizeof(long int), tt.numVals,f);
// 		}
// 		else {
// 		    ti->stripOffsets[0] = tt.values;
// 		}
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 278 :
// 	    if (tt.type==3) {
// 		if (image->bigendian) ti->rowsPerStrip = (tt.values & 0xFFFF0000) >> 16;
// 		else ti->rowsPerStrip = tt.values & 0x0000FFFF;
// 	    }
// 	    else if (tt.type==4) ti->rowsPerStrip = tt.values;
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 279 :
// 	    if (ti->isInitialized) {
// 		if (tt.numVals != ti->numStrips) return TIFF_TAG_ERROR;
// 	    }
// 	    else {
// 		ti->numStrips = tt.numVals;
// 		ti->stripOffsets = (long int*) malloc(tt.numVals*sizeof(long int));
// 		ti->stripByteCounts = (long int*) malloc(tt.numVals*sizeof(long int));
// 		ti->isInitialized = 1;
// 	    }
// 	    if (tt.type==3) {
// 		if (asOffset) {
// 		    fseek(f,tt.values,SEEK_SET);
// 		    for (k=0; k<tt.numVals; k++) {
// 			fread(ti->stripByteCounts+k, sizeof(short), 1, f);
// 			if (image->bigendian) ti->stripByteCounts[k] = swap_short((short) ti->stripByteCounts[k]);
// 		    }
// 		}
// 		else {
// 		    ti->stripByteCounts[0] = (tt.values & 0xFFFF0000) >> 16;
// 		    if (tt.numVals>1) ti->stripByteCounts[1] = tt.values & 0x0000FFFF;
// 		}
// 	    }
// 	    else if (tt.type==4) {
// 		if (asOffset) {
// 		    fseek(f,tt.values,SEEK_SET);
// 		    fread(ti->stripByteCounts, sizeof(long int), tt.numVals,f);
// 		}
// 		else {
// 		    ti->stripByteCounts[0] = tt.values;
// 		}
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 282 :
// 	    if (tt.type==5 && tt.numVals==1) {
// 		fseek(f,tt.values,SEEK_SET);
// 		fread(&image->xResolution.numerator, sizeof(long int), 1, f);
// 		fread(&image->xResolution.denominator, sizeof(long int), 1, f);
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 283 :
// 	    if (tt.type==5 && tt.numVals==1) {
// 		fseek(f,tt.values,SEEK_SET);
// 		fread(&image->yResolution.numerator, sizeof(long int), 1, f);
// 		fread(&image->yResolution.denominator, sizeof(long int), 1, f);
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 296 :
// 	    if (tt.type==3 && tt.numVals==1) {
// 		if (image->bigendian) tt.values = (tt.values && 0xFFFF0000) >> 16;
// 		if (tt.values==2) strcpy(image->resolutionUnit, "inch");
// 		else if (tt.values==3) strcpy(image->resolutionUnit, "cm  ");
// 		else strcpy(image->resolutionUnit, "none");
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	case 320 :
// 	    if (tt.type==3) {
// 		image->numColors = tt.numVals / 3;
// 		image->color = (XColor*) malloc(image->numColors*sizeof(XColor));
// 		reds = (short*) malloc(tt.numVals*sizeof(short));
// 		greens = (short*) malloc(tt.numVals*sizeof(short));
// 		blues = (short*) malloc(tt.numVals*sizeof(short));
// 		fseek(f,tt.values,SEEK_SET);
// 		fread(reds,sizeof(short),tt.numVals,f);
// 		fread(greens,sizeof(short),tt.numVals,f);
// 		fread(blues,sizeof(short),tt.numVals,f);
// 		for (k=0; k<image->numColors; k++) {
// 		    image->color[k].red = reds[k];
// 		    image->color[k].green = greens[k];
// 		    image->color[k].blue = blues[k];
// 		}
// 		free(reds);
// 		free(greens);
// 		free(blues);
// 	    }
// 	    else return TIFF_TAG_ERROR;
// 	    break;
// 	default  :
// 	    fprintf(stderr, "Warning: unknown tag %d\n", tt.tag);
// 	    return TIFF_UNKNOWNTAG_WARN;
// 	    break;
// 	}
	
	return TIFF_OK;
	
    }

    
    public int tiff_read( String filename ) 
    {
    
// 	FILE* f;
// 	char end0,end1;
// 	short magic;
// 	short bigendian;
// 	long int ifdOffset;
// 	long int dataOffset;
// 	short numTags;
// 	int k,res;
// 	tiffInfo ti;
	
// 	f = fopen(filename, "r");
// 	if (!f) return TIFF_IO_ERROR;
	
// 	k = fscanf(f,"%c%c", &end0, &end1);
// 	if (end0 != end1) {
// 	    fclose(f);
// 	    return TIFF_NOT_A_TIFF;
// 	}
// 	switch (end0) {
// 	case 'I' :
// 	    bigendian = 0;
// 	    break;
// 	case 'M' :
// 	    bigendian = 1;
// 	    break;
// 	default  :
// 	    fclose(f);
// 	    return TIFF_NOT_A_TIFF;
// 	    break;
// 	}
// 	this.bigendian = bigendian;
// 	k = fread(&magic, sizeof(short), 1, f);
// 	if (!bigendian) magic = swap_short(magic);
// 	if (magic != 42) {
// 	    fclose(f);
// 	    return TIFF_NOT_A_TIFF;
// 	}
	
// 	fread(&ifdOffset, sizeof(long int), 1,f);
// 	if (!bigendian) ifdOffset = swap_long(ifdOffset);
// 	fseek(f,ifdOffset,SEEK_SET);
	
// 	// read the tags
	
// 	fread(&numTags, sizeof(short), 1, f);
// 	if (!bigendian) numTags = swap_short(numTags);
	
// 	ti.isInitialized = 0;
	
// 	for (k=0; k<numTags; k++) {
// 	    res = read_tag(f,ifdOffset+2+12*k,image, &ti);
// 	    if (res) fprintf(stderr, "Tag error or warning %d (Tag no %d)\n", res, k);
// 	}
// 	fseek(f, ifdOffset+2+12*numTags, SEEK_SET);
// 	fread(&ifdOffset, sizeof(long int), 1,f);
// 	if (!bigendian) ifdOffset = swap_long(ifdOffset);
	
// 	// read the image data
	
// 	dataOffset = 0;
// 	image->data = (unsigned char*) malloc(image->width*image->height*sizeof(unsigned char));
// 	for (k=0; k<ti.numStrips; k++) {
// 	    fseek(f, ti.stripOffsets[k], SEEK_SET);
// 	    fread(image->data + dataOffset, sizeof(short), (size_t) ti.stripByteCounts[k], f);
// 	    dataOffset += ti.stripByteCounts[k];
// 	}
	
// 	fclose(f);
	
// 	free(ti.stripOffsets);
// 	free(ti.stripByteCounts);
	
// 	if (ifdOffset > 0) return TIFF_MULTIIMAGE_WARN;
	/*else*/ return TIFF_OK;
	
    }
    
}
