/*
 * File: $Id: MRNumericPattern.java,v 1.1.1.1 01.02.2017 09:38:49 hans Exp $
 * History
 * Date: 01.02.2017, 11:09  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2017 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */
 
package org.spectrim.tdfdfit;

import java.io.File;
import java.io.FileNotFoundException;
import org.spectrim.signalprocessing.FFT;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import javax.swing.JOptionPane;
import org.spectrim.model.GlobalModel;
import org.spectrim.spectra.Spectra;
import org.spectrim.spectra.Spectrum;

/**
 * Class handling all aspects related to MRNumerical pattern.
 * A "numeric pattern" is a complex basis set signal which can be used by
 * by TDFDFit for quantification. The response is used as is during the fitting
 * but has a maximum of 5 parameters (one instance of a MRFittingParameters-
 * class object) by which the signal can be apodized (Lorentzian and Gaussian), 
 * phased, frequency shifted, and/or amplified.
 * @author  hans
 */

public class MRNumericPattern 
{
     
    /** 
     * Creates a new default instance of MRNumericPattern. 
     */
    public MRNumericPattern( )
    {
        m_fittingParameters = new MRFittingParameters( m_numericPatternName, MRFittingParameters.NUMERIC_PATTERN_MODEL );
        m_numPatModifierSpectralResponse = new MRLineModelSpectralResponse( m_fittingParameters , MRSignalProperties.getInstance( ) );
    }
    
    /** 
     * Copy constructor of MRNumericPattern. 
     * @param numPatToCopy
     */
    public MRNumericPattern( MRNumericPattern numPatToCopy )
    {
        // First copy remaining attributes ..
        m_numberOfPoints = numPatToCopy.getNumberOfPoints( );
        m_fileReference = numPatToCopy.getFileReference( );
        m_acquisitionBandwidth = numPatToCopy.getAcquisitionBandwidth( );
        
        m_numericPatternName = numPatToCopy.getName( );
        
        m_staticNumericPatternSpectralResponse    =   new MRLineModelSpectralResponse( numPatToCopy.getStaticSpectralResponse( MRSignalProperties.getInstance( ) ) );
        m_dynamicNumericPatternSpectralResponse   =   new MRLineModelSpectralResponse( numPatToCopy.getDynamicSpectralResponse( MRSignalProperties.getInstance( ) ) );
        
        m_fittingParameters = new MRFittingParameters( numPatToCopy.getName( ), 
                                                       MRFittingParameters.NUMERIC_PATTERN_MODEL );
        
        setFittingParameters( numPatToCopy.getFittingParameters( ) );
        
    }
    
    /**
     * Creates a new instance of MRParametricPattern with name 'patternName'. 
     * @param patternName
     */
    public MRNumericPattern( String patternName )
    {
        m_numericPatternName = patternName;
        m_fittingParameters = new MRFittingParameters( patternName, MRFittingParameters.NUMERIC_PATTERN_MODEL );
        m_numPatModifierSpectralResponse = new MRLineModelSpectralResponse( m_fittingParameters , MRSignalProperties.getInstance( ) );
    }
    
    /**
     * Method that enables construction of a numerical pattern from a mrui 
     * typed file.
     * @param mruiFileReference 
     */
    void setMRUIFileAsSource( File mruiFileReference ) 
    {
       JMRUIFile mruiFile = new JMRUIFile( mruiFileReference.toString( ) );
       if ( mruiFile.loadMRUITypedData( ) == true )
       {
            m_absorptionChannelTD = mruiFile.getAbsorbtionTDResponse( );
            m_dispersionChannelTD = mruiFile.getDispersionTDResponse( );
            m_numberOfPoints = m_dispersionChannelTD.length;
            
            // Set the static spectral response
            m_staticNumericPatternSpectralResponse = new MRLineModelSpectralResponse( );
            m_staticNumericPatternSpectralResponse.setMRSignalProperties( MRSignalProperties.getInstance( ) );
            
            m_staticNumericPatternSpectralResponse.addTimeDomainResponse( m_absorptionChannelTD, m_dispersionChannelTD );
            m_staticNumericPatternSpectralResponse.normalizeOnTDzeroPart( );
       }
    }
    
    /**
     * Returns currently only the name of the numeric pattern.
     * @return 
     */
    @Override
    public String toString( )
    {
        return m_numericPatternName;
    }
    
    /**
     * Sets the name of the numeric pattern.
     * @param name
     */
    public void setName( String name )
    {
        m_numericPatternName = name;
        m_fittingParameters.setName( name );
    }
    
    /**
     * Returns the name of the numeric pattern.
     * @return 
     */
    public String getName( )
    {
        return m_numericPatternName;
    }
    
    /**
     * Sets or updates the MRFittingParameter attribute of the numeric pattern.
     * @param fitPars
     */
    public void setFittingParameters( MRFittingParameters fitPars )
    {
        m_fittingParameters.setName( fitPars.getName( ) );
        m_fittingParameters.setModelType( fitPars.getModelType( ) );
        m_fittingParameters.setFileReferenceName( fitPars.getFileReferenceName( ) );
        m_fittingParameters.setLineGroup( fitPars.getLineGroup( ) );
        m_fittingParameters.setMetabolite( fitPars.getMetabolite( ) );
        m_fittingParameters.setFitWithArbitraryLineShapeFunction( fitPars.getFitWithArbitraryLineShapeFunction( ) );
        m_fittingParameters.setFitWithReferenceLineShape( fitPars.getFitWithReferenceLineShape( ) );
        
        // fit parameter values
        m_fittingParameters.setArea( fitPars.getArea( ) );
        m_fittingParameters.setOffset( fitPars.getOffset( ) );
        m_fittingParameters.setGaussianWidth( fitPars.getGaussianWidth( ) );
        m_fittingParameters.setPhase( fitPars.getPhase( ) );
        m_fittingParameters.setLorentzianWidth( fitPars.getLorentzianLineWidth( ) );
        
        // minimum parameter values
        m_fittingParameters.setMinimumArea( fitPars.getMinimumArea( ) );
        m_fittingParameters.setMinimumOffset( fitPars.getMinimumOffset( ) );
        m_fittingParameters.setMinimumGaussianWidth( fitPars.getMinimumGaussianWidth( ) );
        m_fittingParameters.setMinimumPhase( fitPars.getMinimumPhase( ) );
        m_fittingParameters.setMinimumLorentzianWidth( fitPars.getMinimumLorentzianWidth( ) );
        
        // maximum parameter values
        m_fittingParameters.setMaximumArea( fitPars.getMaximumArea( ) );
        m_fittingParameters.setMaximumOffset( fitPars.getMaximumOffset( ) );
        m_fittingParameters.setMaximumGaussianWidth( fitPars.getMaximumGaussianWidth( ) );
        m_fittingParameters.setMaximumPhase( fitPars.getMaximumPhase( ) );
        m_fittingParameters.setMaximumLorentzianWidth( fitPars.getMaximumLorentzianWidth( ) );
        
        // prior knowledge references
        m_fittingParameters.setAreaReference( fitPars.getAreaReference( ) );
        m_fittingParameters.setOffsetReference( fitPars.getOffsetReference( ) );
        m_fittingParameters.setGaussianWidthReference( fitPars.getGaussianWidthReference( ) );
        m_fittingParameters.setPhaseReference( fitPars.getPhaseReference( ) );
        m_fittingParameters.setLorentzianWidthReference( fitPars.getLorentzianLineWidthReference( ) );
     
        // The m_numPatModifierSpectralResponse (Voigt typed) 
        // MRLineModelSpectralResponse element is meant to modify the 
        // m_staticNumericPatternSpectralResponse object. The 
        // m_numPatModifierSpectralResponse elements get fitted by the TDFDFit. 
        m_numPatModifierSpectralResponse = new MRLineModelSpectralResponse( m_fittingParameters , MRSignalProperties.getInstance( ) );
    }
    
    /**
     * Returns the MRFittingParameter attribute of the numeric pattern.
     * @return 
     */
    public MRFittingParameters getFittingParameters( )
    {
        return m_fittingParameters;
    }
    
    
    /**
     * Writes the attributes from a RandomAccesFile.
     */
    void write( RandomAccessFile out )
    {
        try
        {
            System.out.println( "Writing numeric pattern start!");
            System.out.println( "File reference : " + m_fileReference );
            
            out.writeUTF( "<MRNumericPattern>" );
            
            out.writeUTF( "<MRNumericPatternHeader>" );
            out.writeUTF( m_numericPatternName );           // Name of the pattern the user sees
            out.writeUTF( m_fileReference );                // Writes the file name of the numeric part
            out.writeInt( m_numberOfPoints );               // Number of points of the response
            out.writeFloat( m_acquisitionBandwidth );       // Writes the bandwidth of the response
            out.writeUTF( "</MRNumericPatternHeader>" );
            
            m_fittingParameters.write( out );
            
            out.writeUTF( "<MRNumericPatternResponse>" );
            
            writeMRNumericPatternAsResponse( out );         // This writes the response to a disk
            
            out.writeUTF( "</MRNumericPatternResponse>" );
            
            out.writeUTF( "</MRNumericPattern>" );
        }
        catch ( IOException e )
        {
            System.out.println( "Error writing numeric pattern!" );
        }
    }
    
    /**
     * Reads MRNumericPattern-attributes from RandomAccessFile "in".
     */
    void read( RandomAccessFile in )
    {
        try
        {
            String tag = in.readUTF( );         // "<MRNumericPattern>"
            if ( tag.compareTo( "<MRNumericPatternHeader>" ) != 0 )
            {
                tag = in.readUTF( );            // "<MRNumericPatternHeader>"
            }
            
            m_numericPatternName = in.readUTF( );
            m_fileReference = in.readUTF( );
            m_numberOfPoints = in.readInt( );
            m_acquisitionBandwidth = in.readFloat( );
            
            tag = in.readUTF( );                // "</MRNumericPatternHeader>"
            
            if ( tag.compareTo( "</MRNumericPatternHeader>" ) == 0 )
            { 
                tag = in.readUTF( );
                m_fittingParameters.read( in );
            }
            
            // read the opening tag
            tag = in.readUTF( );
            if ( tag.compareTo( "<MRNumericPatternResponse>" ) == 0 )
            {
                readTDFDFitFormattedResponseFile( in, m_numberOfPoints );
            }
            tag = in.readUTF( );
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot read TDFDFit-formatted MRNumericPattern response file." );
            e.printStackTrace( );
        }  
    }
    
    /**
     * Clears data structures.
     */
    public void clear( )
    {
        // Nothing to clear
    }
    
    /**
     * Sets the TDFDFit file reference to the file where the numeric pattern
     * is stored.
     * @param fileRef
     */
    public void setNumericPatternFileReference( String fileRef )
    {
        m_fileReference = fileRef;
    }

    /**
     * Sets the file reference to the file where the numeric pattern
     * is stored.
     * @return 
     */
    public String getNumericPatternFileReference( )
    {
        return m_fileReference;
    }
    
    /**
     * Writes the current file in a TDFDFit file format.
     * @param output
     */
    public void writeMRNumericPatternAsResponse( RandomAccessFile output )
    {
        
        m_absorptionChannelFD = m_staticNumericPatternSpectralResponse.getAbsorptionChannelFD( );
        m_dispersionChannelFD = m_staticNumericPatternSpectralResponse.getDispersionChannelFD( );
        
        try
        {
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_absorptionChannelFD[i] );
                
                output.write( buf.get( 3 ) );
                output.write( buf.get( 2 ) );
                output.write( buf.get( 1 ) );
                output.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_dispersionChannelFD[i] );                
                
                output.write( buf1.get( 3 ) );
                output.write( buf1.get( 2 ) );
                output.write( buf1.get( 1 ) );
                output.write( buf1.get( 0 ) );                   
            }            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write TDFDFit-formatted MRNumericPattern response file" );
        }
    }
    
    
    /**
     * Writes the current file in a TDFDFit file format;
     * Which means it is written in frequency domain.
     * @param fileName
     */
    public void writeAsTdfdfitFile( String fileName )
    {
        
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_staticNumericPatternSpectralResponse.m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_staticNumericPatternSpectralResponse.m_absorptionChannelFD[i] );
                
                writer.write( buf.get( 3 ) );
                writer.write( buf.get( 2 ) );
                writer.write( buf.get( 1 ) );
                writer.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_staticNumericPatternSpectralResponse.m_dispersionChannelFD[i] );                
                
                writer.write( buf1.get( 3 ) );
                writer.write( buf1.get( 2 ) );
                writer.write( buf1.get( 1 ) );
                writer.write( buf1.get( 0 ) );         
            }

            writer.close( );

        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write TDFDFit-formatted MRNumericPattern response file" );
        }
    }
    
    /**
     * Writes the current file to TDFDFit file format.
     * @param path
     * @param fileName
     */
    public void writeTDFDFitFormattedResponseFile( String path, String fileName )
    {
        
        if ( path != null )
        {
            fileName = path + System.getProperty( "file.separator" ) + fileName;
        }
        
        System.out.println( "fileName : " + fileName );
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_absorptionChannelFD[i] );
                
                writer.write( buf.get( 3 ) );
                writer.write( buf.get( 2 ) );
                writer.write( buf.get( 1 ) );
                writer.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_dispersionChannelFD[i] );                
                
                writer.write( buf1.get( 3 ) );
                writer.write( buf1.get( 2 ) );
                writer.write( buf1.get( 1 ) );
                writer.write( buf1.get( 0 ) );
                                
            }

            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write TDFDFit-formatted MRNumericPattern response file." );
        }
    }
    
    /**
     * Sets the FD response part of the MRNumericPattern object.
     * @param td_abs
     * @param td_disp 
     */
    public void setTDResponse( float[] fd_abs, float[] fd_disp )
    {
        m_numberOfPoints = fd_abs.length;
        m_absorptionChannelTD = new float[m_numberOfPoints];
        m_dispersionChannelTD = new float[m_numberOfPoints];
        System.arraycopy( fd_abs, 0, m_absorptionChannelTD, 0, fd_abs.length );
        System.arraycopy( fd_disp, 0, m_dispersionChannelTD, 0, fd_abs.length );
        
        // The frequency domain response is read so generate the time 
        // domain response to get full representation
        calculateFDRepresentation( );
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     */
    protected void calculateFDRepresentation( )
    {
        float[][] fftArray = new float[ m_absorptionChannelTD.length][2];
        
        for ( int i = 0 ; i < m_absorptionChannelTD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelTD[i];
            fftArray[i][1] = m_dispersionChannelTD[i];
        }
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        m_absorptionChannelFD = new float[m_absorptionChannelTD.length];
        m_dispersionChannelFD = new float[m_dispersionChannelTD.length];

        for ( int i = 0 ; i < m_absorptionChannelTD.length ; i++ )
        {
            int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelTD.length;
            m_absorptionChannelFD[d] = fftArray[i][0];
            m_dispersionChannelFD[d] = fftArray[i][1];
        }
    
        fftArray = null;
        
        // Bring the spectral response into sync
        m_staticNumericPatternSpectralResponse = new MRLineModelSpectralResponse( );
        m_staticNumericPatternSpectralResponse.setMRSignalProperties( MRSignalProperties.getInstance( ) );
        m_staticNumericPatternSpectralResponse.addTimeDomainResponse( m_absorptionChannelTD, m_dispersionChannelTD ); 
    }
     
    
    /**
     * Reads a TDFDFit file format binary file and stores its contents in the
     * appropriate member variables.
     * @param numberOfComplexPoints
     * @param fileRef
     */
    public void readTDFDFitFormattedResponseFile( int numberOfComplexPoints, String fileRef )
    {
        m_absorptionChannelFD = new float[ numberOfComplexPoints ];
        m_dispersionChannelFD = new float[ numberOfComplexPoints ];
        m_numberOfPoints = numberOfComplexPoints;
        try
        {          
            RandomAccessFile reader = new RandomAccessFile( fileRef, "rw" );
            
            ByteBuffer buf;
            byte byte0;
            byte byte1;
            byte byte2;
            byte byte3;
            float readFloat;
            
            for ( int i = 0; i < numberOfComplexPoints; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_absorptionChannelFD[i] = readFloat;
                
                System.out.println( "i = " + i + "FD-Absorption-Binfile = " + readFloat );
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_dispersionChannelFD[i] = readFloat;
                
            }
            
            // The frequency domain response is read so generate the time 
            // domain response to get full representation
            calculateTDRepresentation( );
            
            reader.close( );
            
        }
        catch ( IOException e )
        {
            e.printStackTrace( );
            System.out.println( "Cannot read TDFDFit-formatted MRNumericPattern response file" );
        }
    }
    
    /**
     * Reads a TDFDFit file format binary file and stores its contents in the
     * appropriate member variables.
     * @param in
     * @param nPoints
     */
    public void readTDFDFitFormattedResponseFile( RandomAccessFile in, int nPoints )
    {
        m_absorptionChannelFD = new float[nPoints];
        m_dispersionChannelFD = new float[nPoints];
        
        m_absorptionChannelTD = new float[nPoints];
        m_dispersionChannelTD = new float[nPoints];
        
        m_numberOfPoints = nPoints;
        
        try
        {           
            ByteBuffer  buf;
            byte        byte0;
            byte        byte1;
            byte        byte2;
            byte        byte3;
            float       readFloat;
            
            for ( int i = 0; i < nPoints; i++ )
            {
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = in.readByte( );
                byte1 = in.readByte( );
                byte2 = in.readByte( );
                byte3 = in.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_absorptionChannelFD[i] = readFloat;
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = in.readByte( );
                byte1 = in.readByte( );
                byte2 = in.readByte( );
                byte3 = in.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_dispersionChannelFD[i] = readFloat;
                
            }
            
            // The frequency domain response is read so generate the time 
            // domain response to get full representation
            calculateTDRepresentation( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot read TDFDFit-formatted MRNumericPattern response file" );
        }
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels of the static, non change-able
     * response.
     */
    private void calculateTDRepresentation( )
    {
        float[][] fftArray = new float[m_absorptionChannelFD.length][2];
        float[] bufAbs = new float[m_absorptionChannelFD.length];
        float[] bufDisp = new float[m_absorptionChannelFD.length];
               
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
            fftArray[i][1] = m_dispersionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
             bufAbs[ i ] = fftArray[i][0];
             bufDisp[ i ] = fftArray[i][1];           
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            int d = ( i + 1 + m_absorptionChannelFD.length / 2 ) % m_absorptionChannelFD.length;
            //int d = ( i + m_absorptionChannelFD.length / 2 ) % m_absorptionChannelFD.length;
            fftArray[d][0] = bufAbs[i];
            fftArray[d][1] = bufDisp[i];
        }
        
        bufAbs = null;
        bufDisp = null;
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        m_absorptionChannelTD = new float[m_absorptionChannelFD.length];
        m_dispersionChannelTD = new float[m_dispersionChannelFD.length];

        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            //int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelFD.length;
            m_absorptionChannelTD[i] = fftArray[i][0] / m_absorptionChannelTD.length;
            m_dispersionChannelTD[i] = fftArray[i][1] / m_absorptionChannelTD.length;
        }
        
        // Bring the spectral response into sync
        m_staticNumericPatternSpectralResponse = new MRLineModelSpectralResponse( );
        m_staticNumericPatternSpectralResponse.setMRSignalProperties( MRSignalProperties.getInstance( ) );
        m_staticNumericPatternSpectralResponse.addTimeDomainResponse( m_absorptionChannelTD, m_dispersionChannelTD );
        
        fftArray = null;
    }
        
    /**
     * Returns the number of points of the response.
     * @return the number of time sample points
     */
    public int getNumberOfPoints( ) 
    {
        return m_numberOfPoints;
    }
    
    /**
     * Sets the number of points of the response.
     * @param numberOfPoints
     */
    public void setNumberOfPoints( int numberOfPoints ) 
    {
        m_numberOfPoints = numberOfPoints;
    }
    
    /**
     * Returns the acquisition band width.
     * @return 
     */
    public float getAcquisitionBandwidth( ) 
    {
        return m_acquisitionBandwidth;
    }
    
    /**
     * Sets the acquisition band width. 
     * @param acquisitionBandwidth
     */
    public void setAcquisitionBandwidth( float acquisitionBandwidth ) 
    {
        m_acquisitionBandwidth = acquisitionBandwidth;
    }
    
    /**
     * Returns the file reference attribute; that is the path where the file is
     * stored on in the file system.
     * @return 
     */
    public String getFileReference( ) 
    {
        return m_fileReference;
    }
    
    /**
     * Sets the file reference attribute; that is the location (path) 
     * where the file is stored on in the file system.
     * 
     * @param fileReference 
     */
    public void setFileReference( String fileReference ) 
    {
        m_fileReference = fileReference;
    }
    
    /**
     * Sets the MRLineModelSpectralResonse attribute of the numerical pattern;
     * Method is used when converting an object to an MRNumericPattern typed
     * object.
     * @param currentResponse.   
     */
    void setMRLineModelSpectralResponse( MRLineModelSpectralResponse spectralResponse ) 
    {
        
        float normalizationFactor = spectralResponse.getNormalizationFactor( );
        
        m_staticNumericPatternSpectralResponse = spectralResponse;
        m_absorptionChannelFD = m_staticNumericPatternSpectralResponse.getAbsorptionChannelFD( );
        m_dispersionChannelFD = m_staticNumericPatternSpectralResponse.getDispersionChannelFD( );
        
        m_absorptionChannelTD = m_staticNumericPatternSpectralResponse.getAbsorptionChannelTD( );
        m_dispersionChannelTD = m_staticNumericPatternSpectralResponse.getDispersionChannelTD( );
        
        if ( normalizationFactor!= 0f )
        {
            for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
            {
                m_absorptionChannelFD[i] = m_absorptionChannelFD[i]/normalizationFactor;
                m_dispersionChannelFD[i] = m_dispersionChannelFD[i]/normalizationFactor;
                m_absorptionChannelTD[i] = m_absorptionChannelTD[i]/normalizationFactor;
                m_dispersionChannelTD[i] = m_dispersionChannelTD[i]/normalizationFactor;
            }
        }
        else
        {
            JOptionPane.showMessageDialog( null, "Cannot normalize numeric pattern \n Cause: Devision by zero." );
        }
    }
    
    /**
     * Returns the spectral response of the parametric pattern.
     * @param m_MRSignalProperties
     * @return 
     */
    public MRLineModelSpectralResponse getStaticSpectralResponse( MRSignalProperties m_MRSignalProperties )
    {
        return m_staticNumericPatternSpectralResponse;
    }
    
    /**
     * Returns the dynamic (modulator) response of this MRNumericPattern.
     * @param m_MRSignalProperties
     * @return 
     */
    public MRLineModelSpectralResponse getDynamicSpectralResponse( MRSignalProperties m_MRSignalProperties )
    {
        m_dynamicNumericPatternSpectralResponse = new MRLineModelSpectralResponse( m_staticNumericPatternSpectralResponse );
        m_dynamicNumericPatternSpectralResponse.applyMRLineModelSpectralResponse( m_numPatModifierSpectralResponse.getFittingParameters( ) );
        return m_dynamicNumericPatternSpectralResponse;
    }
    
    /**
     * Writes the numerical pattern to a binary .bin file which can be read
     * by TDFDFit.
     * @param numPatFileName 
     */
    void writeNumericPatternAsTDFDFitDotBinFile( String numPatFileName, boolean reverseFDOrder ) 
    {
        m_staticNumericPatternSpectralResponse.writeAsTdfdfitFileForFitting( numPatFileName, reverseFDOrder );
    }
    
    // Name and file reference of the numerical pattern
    private String m_numericPatternName = "Num. pattern";
    private String m_fileReference      = "File reference";

    // Object to store the fitting paramters into
    private final MRFittingParameters m_fittingParameters;
    
    private int     m_numberOfPoints;
    private float   m_acquisitionBandwidth = 1.0f;
    
    // Objects to store the response into
    private float[] m_absorptionChannelFD;
    private float[] m_dispersionChannelFD;
    private float[] m_absorptionChannelTD;
    private float[] m_dispersionChannelTD;
    
    MRLineModelSpectralResponse m_staticNumericPatternSpectralResponse;  
    MRLineModelSpectralResponse m_numPatModifierSpectralResponse;
    MRLineModelSpectralResponse m_dynamicNumericPatternSpectralResponse;

}