/*
 * File: $Id: MRParametricPattern.java,v 1.1.1.1 2006/04/19 09:38:49 hans Exp $
 * History
 * Date: 03.06.2003, 11:10  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


package org.spectrim.tdfdfit;
import org.spectrim.signalprocessing.FFT;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.nio.ByteBuffer;

import javax.swing.tree.DefaultMutableTreeNode;

import java.util.LinkedList;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import static org.spectrim.tdfdfit.TDFDFittingDefaults.DEFAULT_TDFDFIT_FIT_MODEL_NAME;

/**
 * Class that defines all aspects of a MRParametricPattern used by TDFDFit.
 * @author  hans
 */
public class MRParametricPattern
{
    
    /** 
     * Creates a new instance of MRParametricPattern. 
     */
    public MRParametricPattern( )
    {
        m_parPatFittingParameters = new MRFittingParameters( m_parametricPatternName, MRFittingParameters.PARAMETRIC_PATTERN_MODEL );
    }
    
    /**
     * Copy constructor of MRParametricPattern.
     * @param parPatCopy
     */
    public MRParametricPattern( MRParametricPattern parPatCopy )
    {
        m_parPatFittingParameters = new MRFittingParameters( parPatCopy.getParametricPatternName( ), MRFittingParameters.PARAMETRIC_PATTERN_MODEL ); 
        setFittingParameters( parPatCopy.getFittingParameters( ) );
        
        // Deep copy of simple lines elements of the simple line list ...
        LinkedList simpleLineList = parPatCopy.getSimpleLineList( );
        
        Iterator it = simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            m_simpleLineList.add( new MRLineModelParameters( (MRLineModelParameters)it.next( ) ) );
        }  
    }   
    
    /** 
     * Creates a new instance of MRParametricPattern with name 'patternName'. 
     * @param patternName
     */
    public MRParametricPattern( String patternName )
    {
        m_parametricPatternName = patternName;
        m_parPatFittingParameters = new MRFittingParameters( m_parametricPatternName, MRFittingParameters.PARAMETRIC_PATTERN_MODEL );
    }
    
    /**
     * Returns currently only the name of the parametric pattern.
     * @return 
     */
    @Override
     public String toString( )
     {
         return m_parametricPatternName;
     }
     
     
    public String getParametricPatternName( ) 
    {
        return m_parametricPatternName;
    }
     
     /**
      * Sets the name of the parametric pattern.
      */
     public void setName( String name  )
     {
         m_parametricPatternName = name;
         m_parPatFittingParameters.setName( name );
     }
     
     /**
      * Sets the name of the parametric pattern.
      */
     public String getName( )
     {
         return m_parametricPatternName;
     }  
     
    /**
     * Sets the MRFitting parameter attribute of the parametric pattern.
     */
    public void setFittingParameters( MRFittingParameters fitPars )
    {
        
        m_parPatFittingParameters.setName( fitPars.getName( ) );
        m_parPatFittingParameters.setModelType( fitPars.getModelType( ) );
        m_parPatFittingParameters.setFileReferenceName( fitPars.getFileReferenceName( ) );
        m_parPatFittingParameters.setLineGroup( fitPars.getLineGroup( ) );
        m_parPatFittingParameters.setMetabolite( fitPars.getMetabolite( ) );
        m_parPatFittingParameters.setFitWithArbitraryLineShapeFunction( fitPars.getFitWithArbitraryLineShapeFunction( ) );
        m_parPatFittingParameters.setFitWithReferenceLineShape( fitPars.getFitWithReferenceLineShape( ) );
        
        // fit parameter values
        m_parPatFittingParameters.setArea( fitPars.getArea( ) );
        m_parPatFittingParameters.setOffset( fitPars.getOffset( ) );
        m_parPatFittingParameters.setGaussianWidth( fitPars.getGaussianWidth( ) );
        m_parPatFittingParameters.setPhase( fitPars.getPhase( ) );
        m_parPatFittingParameters.setLorentzianWidth( fitPars.getLorentzianLineWidth( ) );
        
        // minimum parameter values
        m_parPatFittingParameters.setMinimumArea( fitPars.getMinimumArea( ) );
        m_parPatFittingParameters.setMinimumOffset( fitPars.getMinimumOffset( ) );
        m_parPatFittingParameters.setMinimumGaussianWidth( fitPars.getMinimumGaussianWidth( ) );
        m_parPatFittingParameters.setMinimumPhase( fitPars.getMinimumPhase( ) );
        m_parPatFittingParameters.setMinimumLorentzianWidth( fitPars.getMinimumLorentzianWidth( ) );
        
        // maximum parameter values
        m_parPatFittingParameters.setMaximumArea( fitPars.getMaximumArea( ) );
        m_parPatFittingParameters.setMaximumOffset( fitPars.getMaximumOffset( ) );
        m_parPatFittingParameters.setMaximumGaussianWidth( fitPars.getMaximumGaussianWidth( ) );
        m_parPatFittingParameters.setMaximumPhase( fitPars.getMaximumPhase( ) );
        m_parPatFittingParameters.setMaximumLorentzianWidth( fitPars.getMaximumLorentzianWidth( ) );
        
        // prior knowledge references
        m_parPatFittingParameters.setAreaReference( fitPars.getAreaReference( ) );
        m_parPatFittingParameters.setOffsetReference( fitPars.getOffsetReference( ) );
        m_parPatFittingParameters.setGaussianWidthReference( fitPars.getGaussianWidthReference( ) );
        m_parPatFittingParameters.setPhaseReference( fitPars.getPhaseReference( ) );
        m_parPatFittingParameters.setLorentzianWidthReference( fitPars.getLorentzianLineWidthReference( ) );
    }
    
    /**
     * Returns the MRFittingParameter attribute of the parametric pattern.
     */
    public MRFittingParameters getFittingParameters( )
    {
        return m_parPatFittingParameters;
    }
    
     /**
      * Sets the name of the parametric pattern.
      */
     public void setFileReferenceName( String name  )
     {
         m_fileReference = name;
     }
     
     /**
      * Sets the name of the parametric pattern.
      */
     public String getFileReferenceName( )
     {
         return m_fileReference;
     }  
     
    /** 
     * Adds a list element MRLineModelParameters.
     */
    public void addSimpleLine( MRLineModelParameters simpleLineModel )
    {
        m_simpleLineList.add( simpleLineModel );
    }

    /** 
     * Removes a simple line from the linked list m_simpleLineList.
     */
    public void deleteSimpleLine( MRLineModelParameters simpleLineModel )
    {
        m_simpleLineList.remove( simpleLineModel );
    } 
    
    /**
     * Reads the attributes from a RandomAccesFile.
     */
    void read( RandomAccessFile in )
    {
        
        clear( );
        try
        {
            String tag = in.readUTF( );
            // System.out.println( "tag = " + tag );
            
            // First string after openening tag is the name of the parametric pattern
            m_parametricPatternName = tag;
            
            tag = in.readUTF( );
            m_parPatFittingParameters.read( in );
            
            tag = in.readUTF( );
            // System.out.println( "tag = " + tag );
            
            while ( tag.compareTo( "<MRFittingParameters>" ) == 0 )
            {

                MRLineModelParameters linePars = new MRLineModelParameters( );
                linePars.read( in );
                m_simpleLineList.add( linePars );    

                tag = in.readUTF( );
                // System.out.println( "tag = " + tag );
            }
            
            // Read the sum response of the parametric pattern ..
            m_parPatModelSpectralResponse = new MRLineModelSpectralResponse( );
            m_parPatModelSpectralResponse.read( in, true );
            
            tag = in.readUTF( ); // Should be the </MRParametricPattern> tag.
            
            // System.out.println( "tag = " + tag );
            
        }
        catch ( IOException e )
        {
            e.printStackTrace( );
        }
    }

    /**
     * Writes attributes of this object a RandomAccessFile to a database of
     * ParametricPatterns; this file is importable in other projects.  
     */
    void writeModelFile( )
    {
        try
        {
            TDFDFittingDefaults defaults = TDFDFittingDefaults.getInstance( );
            defaults.load( );
            
            String path = defaults.getProperty( TDFDFittingDefaults.DEFAULT_USED_TDFDFIT_MODEL_PATH );
            
            String parametricPatternDir = System.getProperty( "file.separator" ) + 
                                          "Parametric_Patterns" + 
                                          System.getProperty( "file.separator" );
            
            String directory = path + parametricPatternDir;
            
            String name = getName( );
            
            if ( name.endsWith( ".tfm" ) == true )
            {
                name = name.substring( 0,name.lastIndexOf( ".tfm" ) );
            }
            
            String fileRef = path + parametricPatternDir + name + ".ppf";
            // System.out.println( "Parametric pattern is written seperately in file : " + fileRef );
            
            File outfile = new File( fileRef );
            RandomAccessFile writer = new RandomAccessFile( outfile, "rw" );
            
            // Should the fitting reference object be wriiten as well ??
            
            writer.writeUTF( "<MRParametricPattern>" );
            writer.writeUTF( m_parametricPatternName );
            // the factors should be written as well!!
            m_parPatFittingParameters.write( writer );
            
            writeSimpleLines( writer );
            
            m_parPatModelSpectralResponse.write( writer );
            writer.writeUTF( "</MRParametricPattern>" );
        }
        catch ( IOException e )
        {
            System.out.println( "Could not write parametric file." );
        }
    }
    
    
    /**
     * Writes attributes of this object a RandomAccessFile to a database of
     * ParametricPatterns; this file is importable in other projects.  
     * @param fileReference
     */
    public void writeModelFile( String fileReference )
    {
        try
        {        
            File outfile = new File( fileReference );
            RandomAccessFile writer = new RandomAccessFile( outfile, "rw" );
            
            writer.writeUTF( "<MRParametricPattern>" );
            writer.writeUTF( m_parametricPatternName );
            
            // the factors should be written as well!!
            m_parPatFittingParameters.write( writer );
            
            writeSimpleLines( writer );
            writer.writeUTF( "</MRParametricPattern>" );
        }
        catch ( IOException e )
        {
            System.out.println( "Could not write parametric file." );
        }
    }    
    
    
    /**
     * Reads attributes of this object a RandomAccessFile to a database of
     * ParametricPatterns; these files are exchangeable between TDFDFitModels.  
     * @param node
     * @param patternFileReference
     */
    public void readModelFile( DefaultMutableTreeNode node, String patternFileReference )
    {
        try
        {
            clear( );   
            RandomAccessFile reader = new RandomAccessFile( patternFileReference, "rw" );
            
            // Should the fitting reference object be written as well ??
            String nextTag = reader.readUTF( );
            nextTag = reader.readUTF( );

            setName( nextTag );
            
            nextTag = reader.readUTF( );
            
            int cntr = 0;
            
            while ( nextTag.compareTo( "</MRParametricPattern>" ) != 0 )
            {
                if ( cntr == 0 )
                {
                    m_parPatFittingParameters.read( reader );
                }
                
                nextTag = reader.readUTF( );
                if ( nextTag.compareTo( "<MRFittingParameters>" ) == 0 )
                {   
                    MRLineModelParameters linPar = new MRLineModelParameters( ) ;
                    linPar.read( reader );
                    DefaultMutableTreeNode childNode = new DefaultMutableTreeNode( linPar );
                    node.add( childNode );
                    addSimpleLine( linPar );
                }
                if ( nextTag.compareTo( "<SingleVoxelSpectralData>" ) == 0 )
                {
                    m_parPatModelSpectralResponse = new MRLineModelSpectralResponse( );
                    m_parPatModelSpectralResponse.read( reader, true );
                }
                cntr++;
            }
        }
        catch ( IOException e )
        {
            System.out.println( "Could not read parametric pattern file : " + patternFileReference );
        }
    }

    /**
     * Writes the attributes of the MRParametricPattern to a RandomAccessFile.
     */
    void write( RandomAccessFile out )
    {   
        
        // when called from save model OK
        
        // Sync the response with the current model
        m_parPatModelSpectralResponse = getSpectralResponse( MRSignalProperties.getInstance( ) );
        
        // Write a seperate autonomous file that contains the model parameters of
        // this numeric pattern
        writeModelFile( );
        
        // Write it also to the RandomAccessFile that is indicated by the caller of this
        // method
        try
        {
            out.writeUTF( "<MRParametricPattern>" );
            
            out.writeUTF( m_parametricPatternName );
            
            m_parPatFittingParameters.write( out );
            
            writeSimpleLines( out );
            
            m_parPatModelSpectralResponse.write( out );
            
            out.writeUTF( "</MRParametricPattern>" );
        }
        catch ( IOException e )
        {
            System.out.println( "Error writing spectral response to file." );
        }
    }
    
    /**
     * Removes all the lines from m_simpleLineList.
     */
    public void clear( )
    {
        m_simpleLineList.clear( );
        m_simpleLineList = null;
        m_simpleLineList = new LinkedList<MRLineModelParameters>( );
    }
    
    /**
     * Writes all metabolites to a random access file.
     */    
    private void writeSimpleLines( RandomAccessFile out )
    {      
        // get a reference to all the keys of the list containing 
        // the simple line information
        Iterator iter = m_simpleLineList.iterator( );
        // Iterate of all elements of the list
        while ( iter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)iter.next( );
            simpleLine.write( out );
        }
    }
    
    /**
     * Returns the LinkedList containing the MRLineModelParameters.
     * @return 
     */
    public LinkedList getSimpleLineList( )
    {
        return m_simpleLineList;
    }

    /**
     * Sets the LinkedList containing the MRLineModelParameters.
     * @param lineList
     */
    public void setSimpleLineList( LinkedList<MRLineModelParameters> lineList )
    {
        m_simpleLineList = lineList;
    }
    
    /**
     * Writes the current parametric pattern object as numerical pattern
     * the fitting parameters object of the numerical pattern is set to
     * numerical values -> ( 1.00, 0.00, 0.00, 0.00, 0.00 ).
     */
    public void exportParametricPatternAsNumericalPattern( )
    {
        MRLineModelSpectralResponse currentResponse = getSpectralResponse( MRSignalProperties.getInstance( ) );
        
        m_absorptionChannelFD = currentResponse.getAbsorptionChannelFD( );
        m_dispersionChannelFD = currentResponse.getDispersionChannelFD( );
        
        TDFDFittingDefaults defaults = TDFDFittingDefaults.getInstance( );
        defaults.load( );
        
        //String path = defaults.getProperty( DEFAULT_TDFDFIT_FIT_MODEL_NAME );
        String path = defaults.getProperty( TDFDFittingDefaults.DEFAULT_USED_TDFDFIT_MODEL_PATH );
        
        String numericPatternDir = System.getProperty( "file.separator" ) + 
                                                       "Numeric_Patterns" + 
                                                       System.getProperty( "file.separator" );
        
        String tdfdfitResponseDir = System.getProperty( "file.separator" ) + 
                                                       "tdfdfit_responses" + 
                                                       System.getProperty( "file.separator" );
        String directory = path + numericPatternDir;
        
        MRNumericPattern numPat = new MRNumericPattern( );
        numPat.setNumberOfPoints( m_absorptionChannelFD.length );
        numPat.setFileReference( directory + getName( ) + "_AsNumpat.npf" );
        numPat.setMRLineModelSpectralResponse( currentResponse );
        numPat.setName( getName( ) + "_num_pat" );
        
        try 
        {
            try ( RandomAccessFile out = new RandomAccessFile( directory + getName( ) + "_AsNumpat.npf", "rw" ) ) 
            {
                numPat.write( out );
            }
        } 
        catch ( FileNotFoundException ex )
        {
            Logger.getLogger( MRParametricPattern.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
        catch ( IOException ex )
        {
            Logger.getLogger( MRParametricPattern.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
        
        writeTDFDFitFormattedResponseFile( path + tdfdfitResponseDir, getName( ) + ".bin" );
    
    }
    
    /**
     * Calculates the spectral representation of the complete 
     * parametric pattern.
     */
    public MRLineModelSpectralResponse getSpectralResponse( MRSignalProperties mrSignalProperties )
    {
        
        // Create a model spectral response per default constructor
        MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( );
        
        m_parPatModelSpectralResponse = null;
        m_parPatModelSpectralResponse = new MRLineModelSpectralResponse( mrSignalProperties.getNumberOfSamplePoints( ) );
        m_parPatModelSpectralResponse.setMRSignalProperties( mrSignalProperties );
        
        // Iterate over the m_simpleLineList containing the simple lines
        Iterator iter = m_simpleLineList.iterator( );
        while ( iter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)iter.next( );
            
            // Calculate the corresponding resonance line
            resp = new MRLineModelSpectralResponse( (MRFittingParameters)simpleLine, mrSignalProperties );
            m_parPatModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
        }
        
        m_parPatModelSpectralResponse.applyMRLineModelSpectralResponse( m_parPatFittingParameters );
        
        return m_parPatModelSpectralResponse;
    } 
    
    /**
     * Writes the current file to TDFDFit file format.
     */
    public void writeTDFDFitFormattedResponseFile( String path, String fileName )
    {
        
        if ( path != null )
        {
            fileName = path + fileName;
        }
        
        System.out.println( "fileName : " + fileName );
        try
        {     
            try ( RandomAccessFile writer = new RandomAccessFile( fileName, "rw" ) ) 
            {
                ByteBuffer buf, buf1;
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    buf = ByteBuffer.allocate( 4 );
                    
                    buf.putFloat( m_absorptionChannelFD[i] );
                    
                    writer.write( buf.get( 3 ) );
                    writer.write( buf.get( 2 ) );
                    writer.write( buf.get( 1 ) );
                    writer.write( buf.get( 0 ) );
                    
                    buf1 = ByteBuffer.allocate( 4 );
                    
                    buf1.putFloat( m_dispersionChannelFD[i] );
                    
                    writer.write( buf1.get( 3 ) );
                    writer.write( buf1.get( 2 ) );
                    writer.write( buf1.get( 1 ) );
                    writer.write( buf1.get( 0 ) );
                    
                }
            }
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile." );
        }
    }
    
    /**
     * Reads a TDFDFit file format binary file and stores its contents in the
     * appropriate member variables.
     * @param numberOfComplexPoints
     * @param fileRef
     */
    public void readTDFDFitFormattedResponseFile( int numberOfComplexPoints, String fileRef )
    {
        
        try
        {          
            RandomAccessFile reader = new RandomAccessFile( fileRef, "rw" );
            
            ByteBuffer buf;
            byte byte0;
            byte byte1;
            byte byte2;
            byte byte3;
            float readFloat;
            
            for ( int i = 0; i < numberOfComplexPoints; i++ )
            {
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_absorptionChannelFD[i] = readFloat;
                
                buf = ByteBuffer.allocate( 4 );
                
                byte0 = reader.readByte( );
                byte1 = reader.readByte( );
                byte2 = reader.readByte( );
                byte3 = reader.readByte( );
                  
                buf.put( byte3 );
                buf.put( byte2 );
                buf.put( byte1 );
                buf.put( byte0 );
                
                readFloat = buf.getFloat( 0 );
                buf.clear( );
                
                m_dispersionChannelFD[i] = readFloat;
                
            }
            
            // the frequency domain response is read so generate the time 
            // domain response to get full representation
            calculateTDRepresentation( );
            
            reader.close( );
            
        }
        catch ( IOException e )
        {
            e.printStackTrace( );
            System.out.println( "2. Cannot read tdfdfit-formatted measured datafile" );
        }
    }
    
    /**
     * This method writes the pattern to a text file. 
     * @param file_reference 
     */
    public void writeParPatternAsTDFDFitDotPatFile( String file_reference )
    {
        TextFileWriter writer = null;
        MRSignalProperties mrSignalProperties = MRSignalProperties.getInstance( );
        
        float frequencyscale = mrSignalProperties.getFrequencyScale( );
        int samplePoints = mrSignalProperties.getNumberOfSamplePoints( );
        
        float bandwidth = samplePoints;
        
        try 
        {    
            writer = new TextFileWriter( new FileWriter( file_reference ) );
            
            writer.setNumberFormat( "0.00000000E00" );
            writer.writeStringLn( "start_pattern" );
            writer.writeString( "number_of_lines\t" );
            writer.writeInteger( m_simpleLineList.size( ) );
            writer.writeStringLn( "" );
            writer.writeStringLn( "parameters" );
            writer.writeStringLn( "name\t\tline_type\tarea\t\tfrequency\tgauss_width\tzero_phase\tlorentz_width" );
            
            for ( int i = 0; i < m_simpleLineList.size( ); i++ )
            {
                
                MRFittingParameters fitPars = m_simpleLineList.get( i );
                
                writer.writeString( fitPars.getName( ) + "\t\t" );
                
                // set default line model to voigt ..
                // TODO make this user setable in a later stage
                writer.writeString( "voigt" + "\t\t" );
                writer.writeFloat( fitPars.getArea( ) );
                writer.writeString( "\t" );
                writer.writeFloat( (fitPars.getOffset( ) + bandwidth/2f) );
                writer.writeString( "\t" );
                writer.writeFloat( fitPars.getGaussianWidth( ) );
                writer.writeString( "\t" );
                writer.writeFloat( fitPars.getPhase( ) );
                writer.writeString( "\t" );
                writer.writeFloat( fitPars.getLorentzianLineWidth( ) );
                writer.writeStringLn( "" );
            }
            writer.writeStringLn( "end_of_pattern" );
            writer.flush( );
            writer.close( );
        } 
        catch ( IOException ex ) 
        {
            Logger.getLogger( MRParametricPattern.class.getName(  ) ).log( Level.SEVERE, null, ex );
        }
    }
    
    
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     */
    private void calculateTDRepresentation( )
    {
        float[][] fftArray = new float[m_absorptionChannelFD.length][2];
        float[] bufAbs = new float[m_absorptionChannelFD.length];
        float[] bufDisp = new float[m_absorptionChannelFD.length];
               
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
            fftArray[i][1] = m_dispersionChannelFD[ m_absorptionChannelFD.length - 1 - i ];
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
             bufAbs[ i ] = fftArray[i][0];
             bufDisp[ i ] = fftArray[i][1];           
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            int d = ( i + 1 + m_absorptionChannelFD.length / 2 ) % m_absorptionChannelFD.length;
            fftArray[d][0] = bufAbs[i];
            fftArray[d][1] = bufDisp[i];
        }
        
        bufAbs = null;
        bufDisp = null;
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        m_absorptionChannelTD = new float[m_absorptionChannelFD.length];
        m_dispersionChannelTD = new float[m_dispersionChannelFD.length];

        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            //int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelFD.length;
            m_absorptionChannelTD[i] = fftArray[i][0] / m_absorptionChannelTD.length;
            m_dispersionChannelTD[i] = fftArray[i][1] / m_absorptionChannelTD.length;
        }
    
        fftArray = null;
    }
    
    // Attributes
    private String m_parametricPatternName  = "Par_Pattern";
    private String m_fileReference          = "File reference";
    private MRFittingParameters m_parPatFittingParameters;
    
    // Contains a linkedlist of < simple line >
    private LinkedList<MRLineModelParameters> m_simpleLineList = new LinkedList<>( );
    
    // Attribute that holds the sum spectrum of all simple lines
    // that are defined in the list
    private MRLineModelSpectralResponse m_parPatModelSpectralResponse;
    
    private float[] m_absorptionChannelFD;
    private float[] m_dispersionChannelFD;
    private float[] m_absorptionChannelTD;
    private float[] m_dispersionChannelTD;
    
}