/*
 * File: $Id: OrderedBaseLinePointsDataModel.java,v 1.1.1.1 2006/04/19 09:38:51 hans Exp $
 * History
 * Date: 17.03.2006, 12:06  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2006 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

/**
 * OrderedBaseLinePointsDataModel.java
 *
 * Created on March 17, 2006, 9:37 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package org.spectrim.tdfdfit;

import java.io.RandomAccessFile;
import java.io.IOException;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * This class handles a OrderedBaseLinePointsDataModel.
 * The class contains the spline nodes to be be used to define the 
 * baseline spline. Only the offset values are relevant, the y-values
 * will be determined individually for the case to be corrected.
 * The offsets are explicitly defined in ppm units.
 * @author hans
 */
public class OrderedBaseLinePointsDataModel 
{
    
    /** 
     * Singleton class OrderedBaseLinePointsDataModel.
     */
    private OrderedBaseLinePointsDataModel( )
    {
        m_arraySize = 0;
        TDFDFittingDefaults fdef = TDFDFittingDefaults.getInstance( );
        fdef.load( );
        
        String selectedSplineNodeModelsPath = fdef.getProperty( TDFDFittingDefaults.BASELINE_SPLINE_MODEL_PATH );
        String selectedSplineNodeModelName = fdef.getProperty( TDFDFittingDefaults.BASELINE_SPLINE_MODEL_NAME );
        
        if ( selectedSplineNodeModelName != null){
            if ( !selectedSplineNodeModelName.endsWith( "snm" ) )
            {
                selectedSplineNodeModelName = selectedSplineNodeModelName + ".snm";
            }
        }
        
        if ( selectedSplineNodeModelName != null  )
        {
            RandomAccessFile in = null;
            try 
            {
                File tempFile = new File( selectedSplineNodeModelsPath + System.getProperty( "file.separator" ) + selectedSplineNodeModelName );
                boolean exists = tempFile.exists( );
          
                if ( exists )
                {
                    in = new RandomAccessFile( selectedSplineNodeModelsPath + System.getProperty( "file.separator" ) + selectedSplineNodeModelName, "r" );
                    read( in );
                    try 
                    {
                        in.close( );
                    } 
                    catch ( IOException ex ) 
                    {
                        Logger.getLogger(OrderedBaseLinePointsDataModel.class.getName()).log( Level.SEVERE, null, ex );
                    }
                }
            } 
            catch ( FileNotFoundException ex ) 
            {
                Logger.getLogger( OrderedBaseLinePointsDataModel.class.getName( ) ).log( Level.SEVERE, null, ex );
            }
        
        }
    }
    
    /**
     * Returns the one and only instance of OrderedBaseLinePointsDataModel.
     * @return 
     */
    public static OrderedBaseLinePointsDataModel getInstance( )
    {
        if ( theInstance == null )
        {
            theInstance = new OrderedBaseLinePointsDataModel( );
        }
        return theInstance;
    }
    
    public void setArraySize( int arraySize )
    {
        m_arraySize = arraySize;
    }
    
    /**
     * Deletes all points of the model.
     */
    public void deleteAllPoints( )
    {
        m_arraySize = 0;
        m_splineNodesVector.clear( );
    }
    
    /** 
     * Deletes the "index"-number entry from  OrderedBaseLinePointsDataModel. 
     * @param index
     */
    public void deletePointNumber( int index )
    {
        m_splineNodesVector.removeElementAt( index );
    }
   
    /** 
     * Creates a new Point instance to the BaseLinePointsDataModel. 
     * @param newPoint 
     */
    public void addPoint( Point2D.Double newPoint )
    {
        
        Point2D.Double newPointAdd = new Point2D.Double( (float)newPoint.getX( ), 
                                                         (float)newPoint.getY( ) );
        
        if ( m_splineNodesVector.isEmpty( ) )
        {
            m_splineNodesVector.add( 0, newPointAdd );
        }
        else
        {
            if ( m_splineNodesVector.size( ) == 1 )
            {
                if ( newPointAdd.getX( ) > m_splineNodesVector.get( 0 ).getX( ) )
                {
                    m_splineNodesVector.add( newPointAdd );
                }
            }
            else
            {
                boolean addedNothing = true;
                Point2D.Double shiftedPoint = new Point2D.Double( newPointAdd.getX( ) + ( m_arraySize / 2 ),  newPointAdd.getY( ) );
                int currentSize = m_splineNodesVector.size( );
                // There are already at least two point defined (min and max). 
                for ( int i = 0; i < currentSize; i++ )
                {
                    if ( ( shiftedPoint.getX( ) > m_splineNodesVector.get( i ).getX( ) ) )
                    {
                        // do nothing
                    }
                    else
                    {
                        addedNothing = false;
                        m_splineNodesVector.insertElementAt( shiftedPoint, i );
                    }
                }
                if ( addedNothing == true )
                {
                    m_splineNodesVector.add( shiftedPoint );
                }
            }
        }
    }
     
    
    public Vector<Point2D.Double> getSplineNodeVector( )
    {
        return m_splineNodesVector;
    }    
    /**
     * Returns the Point object corresponding to number "pointNumber".
     * @param pointNumber
     * @return 
     */
    public Point2D.Double getPoint( int pointNumber )
    {
        return m_splineNodesVector.get( pointNumber );
    }
    
    /**
     * Returns the size of the vector of points in the list.
     * @return 
     */ 
    public int getSize( )
    {
        return m_splineNodesVector.size( );
    }
    
    
    /**
     * Writes attributes to RandomAccessFile.
     * @param out
     */
    public void write( RandomAccessFile out )
    {
        try
        {
            out.writeUTF( "<OrderedBaseLinePointsModel>" );
            
            // writes number of points
            out.writeInt( m_splineNodesVector.size( ) );
            
            // writes all points
            for ( int i = 0; i < m_splineNodesVector.size( ); i++ )
            {
                out.writeUTF( "<DoublePoint>" );
                out.writeDouble( m_splineNodesVector.get( i ).getX( ) );
                out.writeDouble( m_splineNodesVector.get( i ).getY( ) );
                out.writeUTF( "</DoublePoint>" );
            }
            
            out.writeUTF( "</OrderedBaseLinePointModel>" );
        }
        catch ( IOException e )
        {
        }
    }
    
    /**
     * Reads attributes from RandomAccessFile.
     * @param in
     */
    public void read( RandomAccessFile in )
    {
        // When reading an existing file first delete the elements that may
        // be contained already ...
        m_arraySize = -1;
        m_splineNodesVector.removeAllElements( );
        
        try
        {
           // read the opening tag
            String tag = in.readUTF( ); 
            
            int numberOfPoints = in.readInt( );
            
            for ( int i = 0 ; i < numberOfPoints ; i++ )
            {
                // reads openening tag
                tag = in.readUTF( );
                
                double x_coord = in.readDouble( );
                double y_coord = in.readDouble( );
                
                m_splineNodesVector.add( new Point2D.Double( x_coord, y_coord ) );
                
                // reads closing tag
                tag = in.readUTF( );
            }
            
            // read the closing tag
            tag = in.readUTF( );   
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot read OrderedBaseLinePointsDataModel from file!" );
            e.printStackTrace( );
        }
    }
    
    // Attributes
    // Static private attribute is the instance itself (Singleton!)
    private static  OrderedBaseLinePointsDataModel theInstance = null;
    private int m_arraySize = -1;
    private Vector<Point2D.Double> m_splineNodesVector = new Vector<Point2D.Double>( );
    
}
