/*
 * File: $Id: SingleVoxelMeasuredSpectralData.java,v 1.2 2006/04/26 14:54:11 hans Exp $
 * History
 * Date: 22.07.2003, 14:28  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.tdfdfit;

import java.awt.Color;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.nio.ByteBuffer;
import org.dcm4che2.data.DicomObject;
import org.spectrim.data.CSIData;
import org.spectrim.io.dicom.DicomHandler;
import org.spectrim.io.dicom.DicomHeader;

import org.spectrim.lib.Convert;
import org.spectrim.lib.DicomStringReformatter;
import org.spectrim.spectra.Spectra;


/**
 * This class extends the SingleVoxelSpectralData class which is intended to
 * contain time and frequency domain representations of single voxel 
 * spectroscopic data which is measured; Currently only Siemens data is 
 * supported; this class can further be extended to support other data formats
 * as well.
 * @author  hans
 */

public class SingleVoxelMeasuredSpectralData extends SingleVoxelSpectralData
{
    // Member necessary to finally get rid of the Siemens specific header 
    // information.
    private String m_studyDate;
    private String m_seriesNumber;
    private String m_seriesDescription;
    private String m_patientDOB;
    private String m_patientName;
    private String m_instanceNumber;
    
    /** 
     * Creates a new instance of SingleVoxelMeasuredSpectralData. 
     */
    public SingleVoxelMeasuredSpectralData( )
    {
    }
    
    /** 
     * Copy constructor SingleVoxelMeasuredSpectralData. 
     * @param toCopy
     */
    public SingleVoxelMeasuredSpectralData( SingleVoxelMeasuredSpectralData toCopy )
    {
        
        super( (SingleVoxelSpectralData)toCopy );
        
        m_patientName = toCopy.getPatientName( );
        m_patientDOB = toCopy.getPatientBirthDate( );
        m_seriesDescription = toCopy.getSeriesDescription( );
        m_seriesNumber = toCopy.getSeriesNumber( );
        m_instanceNumber = toCopy.getInstanceNumber( );
        m_studyDate = toCopy.getAcquisitionDate( );
        m_TE = toCopy.getTE( );
        m_TR = toCopy.getTR( );
   
    }
    
    
    /**
     * Creates a new instance of SingleVoxelMeasuredData and reserves N sample
     * points for the response.
     * @param samplePoints
     */
    public SingleVoxelMeasuredSpectralData( int samplePoints )
    {
       allocateArrays( samplePoints );
    }
    
    /**
     * Siemens specific constructor based on the siemensDicomData byte array
     * data.
     * @param siemensDicomData 
     */
    public SingleVoxelMeasuredSpectralData( byte[] siemensDicomData )
    {
        setSiemensRawData( siemensDicomData );
    }
     
    /**
     * Constructor to be called if TD abs and disp channels are known.
     * @param abs
     * @param disp
     */ 
    public SingleVoxelMeasuredSpectralData( float[] abs, float[] disp )
    {
        m_bothChannelsTD        = new float[ 2 * abs.length ];
        m_absorptionChannelTD   = new float[ abs.length ];
        m_dispersionChannelTD   = new float[ disp.length ];
        this.setNumberOfSamplePoints( abs.length );
        for ( int i = 0; i < abs.length; i++ )
        {
            m_bothChannelsTD[2*i] = abs[i];
            m_bothChannelsTD[2*i+1] = disp[i];
            m_absorptionChannelTD[i] = abs[i];
            m_dispersionChannelTD[i] = disp[i];
        }
        
        calculateFDRepresentation( );
        
        autoSetXaxis( abs.length );
    }

    /**
     * Alternative constructor.
     * @param m_numberOfPoints
     * @param red 
     */
    SingleVoxelMeasuredSpectralData( int m_numberOfPoints, Color color ) 
    {
       super( m_numberOfPoints, color );
    }
    
    /**
     * 
     * @param abs
     * @param disp
     * @param domain 
     */
    public SingleVoxelMeasuredSpectralData( float[] abs, float[] disp, int domain )
    {
        super( abs, disp, domain );
    }
    
    
    /**
     * This method is the interface between the old loading of data as was done
     * on Siemens data of SCANalyze and SpectrIm way of loading Siemens or 
     * Philips data; It returns a SingleVoxelMeasuredSpectralData object.
     * @param selectedSVSFileReference 
     * @return  
     */
     public SingleVoxelMeasuredSpectralData getSVSData( String selectedSVSFileReference )
     {
        
        SingleVoxelMeasuredSpectralData svsData = null;
        
        CSIData csidata;
        DicomObject object;
        String filetype;
        String imagetype;
        String classUID;
        boolean canLoadSVS       = false;
        boolean isSiemensMRSData = false;
        
        if ( DicomHandler.isDicom( selectedSVSFileReference ) ==  true ) 
        {
            object = DicomHandler.getDicomObject( selectedSVSFileReference );
            imagetype = DicomHandler.getImageType( object );
            classUID = DicomHandler.getSOPClassUID( object );
            switch ( classUID ) 
            {
                case "1.2.840.10008.5.1.4.1.1.4": // Image
                    canLoadSVS = false;
                    break;
                case "1.2.840.10008.5.1.4.1.1.4.2": // Spectro std
                    canLoadSVS = true;
                    isSiemensMRSData = false; // this is the case of Philips data and Siemens data follwoing the Dicom standard.
                    break;
                case "1.3.12.2.1107.5.9.1": // Spectro/Tensor Siemens
                    if ( imagetype.indexOf( "TENSOR" ) >= 0 ) 
                    {
                        canLoadSVS = false;
                    } 
                    else 
                    {
                        // we are dealing with Siemens spectroscopy data
                        canLoadSVS = true;
                        isSiemensMRSData = true;
                    }
                    break;
                case "1.2.840.10008.5.1.4.1.1.7": // Perfusion derived maps
                    canLoadSVS = false;
                    break;
                case "1.2.840.10008.5.1.4.1.1.4.1":
                    //enhanced MR image
                    canLoadSVS = false;
                    break;                        
                default: // Unknown format
                    canLoadSVS = false;
                    break;
            }
            
            DicomHeader header = DicomHandler.getHeader( object );
                
            if ( isSiemensMRSData == true ) 
            {
                DicomHandler.readSiemensData( object, header );
            }
            
            if ( canLoadSVS ) 
            {
               	csidata = new CSIData( header, selectedSVSFileReference );
                // csidata.conjugate( );
                if ( csidata != null )
                {
            
                    Spectra spectra = csidata.getSpectra( );
                    String key = spectra.getData( ).keySet( ).iterator( ).next( );
                    
                    float[] absTD  = csidata.getSpectra( ).getSpectrumData( key, Spectra.Domain.TD, Spectra.Mode.ABS );
                    float[] dispTD = csidata.getSpectra( ).getSpectrumData( key, Spectra.Domain.TD, Spectra.Mode.DISP );
                    
                    for ( int i = 0; i < dispTD.length; i++ )
                    {
                        dispTD[i] = -dispTD[i];
                    }
                    // Read the single voxel spectral data response
                    svsData = new SingleVoxelMeasuredSpectralData( absTD, dispTD );
                    DicomStringReformatter dformat = new DicomStringReformatter( );
            
                    svsData.setPatientName( dformat.reformatPatientName( csidata.getPatientName( ).replaceAll( "�","_" ) ) );
                    svsData.setPatientBirthDate( dformat.getDateMonthYear( csidata.getPatientDOB( ) ) );
                    svsData.setSeriesDescription( csidata.getSeriesDescription( ) );
                    svsData.setSeriesNumber( csidata.getSeriesNumber( ) );
                    svsData.setInstanceNumber( csidata.getInstanceNumber( ) );
                    svsData.setAcquisitionDate( dformat.getDateMonthYear( csidata.getStudyDate( ) ) );
                    svsData.setTE( csidata.getTE( ) );
                    svsData.setTR( csidata.getTR( ) ); 
            
                    double[] excVoxelDimensions = csidata.csigrid.getDimsExcitedVolume( );
            
                    svsData.set_sVOIThickness( excVoxelDimensions[0] );
                    svsData.set_sVoIPhaseFOV( excVoxelDimensions[1] );
                    svsData.set_sVoIReadoutFOV( excVoxelDimensions[2] );
            
                    // m_MRSpectrumModellerPanel.setMeasuredClinicalSVSSpectrum( clinicalSpectrumData.getFile( ), m_singleVoxelClinicalData );             
           
                }
            }
        }
        
         return svsData;
         
     }  
    
    /** 
     * Sets the absorption and dispersion arrays and number of spectral points
     *
     * @param siemensDicomRawData byte array as obtained by simply reading the 
     *                            Siemens DICOM file containing spectroscopic 
     *                            data.
     */
    public void setSiemensRawData( byte[] siemensDicomRawData )
    {
        byte[] dummy = new byte[4];
        
        m_bothChannelsTD = new float[ siemensDicomRawData.length / 4 ];
        m_absorptionChannelTD = new float[ siemensDicomRawData.length / 8 ];
        m_dispersionChannelTD = new float[ siemensDicomRawData.length / 8 ];
        
        for ( int i = 0; i < siemensDicomRawData.length / 4; i++ )
        {
            dummy[ 0 ] = siemensDicomRawData[ 4 * i + 3 ];
            dummy[ 1 ] = siemensDicomRawData[ 4 * i + 2 ];
            dummy[ 2 ] = siemensDicomRawData[ 4 * i + 1 ];
            dummy[ 3 ] = siemensDicomRawData[ 4 * i + 0 ];
                
            m_bothChannelsTD[ i ] = Convert.byte2float( dummy, 0 );
        }
        
        for ( int i = 0; i < m_bothChannelsTD.length / 2; i++ )
        {
            m_absorptionChannelTD[ i ] = m_bothChannelsTD[ 2 * i ];
            m_dispersionChannelTD[ i ] = m_bothChannelsTD[ 2 * i + 1 ];
        }
        
        calculateFDRepresentation( );
        
        autoSetXaxis( siemensDicomRawData.length / 8 );
    }
    
    /**
     * Writes the current data file to a TDFDFit formatted binary file.
     * @param fileName
     * @return 
     */
    public String writeAsTdfdfitFileForFitting( String fileName )
    {
        
        TDFDFittingDefaults fitDef = TDFDFittingDefaults.getInstance( );
        fitDef.load( );
        
        String path = fitDef.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH );
        
         // Remove the file extension
        while ( fileName.contains( "." ) == true )
        {
            int index = fileName.lastIndexOf( "." );
            fileName = fileName.substring( 0, index );
        }
        
        String dataFileName = path + System.getProperty( "file.separator" ) + fileName;
        
        // System.out.println( "Data filename : " + dataFileName );
        
        try
        {
            
            //m_absorptionChannelTD[0] = m_absorptionChannelTD[0] / 2f;
            //m_dispersionChannelTD[0] = m_dispersionChannelTD[0] / 2f;
            
            this.calculateFDRepresentation( );
            
            try ( RandomAccessFile writer = new RandomAccessFile( path + System.getProperty( "file.separator" ) + fileName + ".bin", "rw" ) ) 
            {
                ByteBuffer buf, buf1;
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    buf = ByteBuffer.allocate( 4 );
                    
                    buf.putFloat( m_absorptionChannelFD[i] );
                    
                    
                    writer.write( buf.get( 3 ) );
                    writer.write( buf.get( 2 ) );
                    writer.write( buf.get( 1 ) );
                    writer.write( buf.get( 0 ) );
                    
                    buf1 = ByteBuffer.allocate( 4 );
                    
                    buf1.putFloat( m_dispersionChannelFD[i] );
                    
                    writer.write( buf1.get( 3 ) );
                    writer.write( buf1.get( 2 ) );
                    writer.write( buf1.get( 1 ) );
                    writer.write( buf1.get( 0 ) );
                    
                }   //m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2f;
                //m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2f;

            calculateFDRepresentation( );
            }
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile" );
        }
        return fileName + ".bin";
    }
    
    /**
     * Writes the current file to TDFDFit file format.
     */
    @Override
    public void writeAsTdfdfitFile( String path, String fileName )
    {
        
        fileName = path + System.getProperty( "file.separator" ) + fileName;
        System.out.println( "fileName : " + fileName );
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            
            for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
            {
                buf = ByteBuffer.allocate( 4 );
                
                buf.putFloat( m_absorptionChannelFD[i] );
                
                writer.write( buf.get( 3 ) );
                writer.write( buf.get( 2 ) );
                writer.write( buf.get( 1 ) );
                writer.write( buf.get( 0 ) );
                
                buf1 = ByteBuffer.allocate( 4 );
                
                buf1.putFloat( m_dispersionChannelFD[i] );                
                
                writer.write( buf1.get( 3 ) );
                writer.write( buf1.get( 2 ) );
                writer.write( buf1.get( 1 ) );
                writer.write( buf1.get( 0 ) );
                                
            }

            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile" );
        }
    }

    /**
     * Applies the QUALITY algorithm to the measured data.
     * @param applyFullQuality
     * @param reference
     */
    public void applyQuality( boolean applyFullQuality, 
                              SingleVoxelMeasuredSpectralData reference )
    {
        // Determine the phase of the water refererence signal
        float[] re = reference.getAbsorptionChannelTD( );
        float[] im = reference.getDispersionChannelTD( );
        
        float[] norm = new float[re.length];
        
        // importing spectrum halves the first time point
        float ref_scale = 2.0f * ( float )Math.sqrt( re[0] * re[0] + im[0] * im[0] );
        float[] waternorm = new float[re.length];
        float[] referencePhase = new float[re.length];
        float[] correctedSignalPhase = new float[re.length];
        
        for ( int i= 0; i < re.length; i++ )
        {
            norm[i] = (float)Math.sqrt( m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + m_dispersionChannelTD[i] * m_dispersionChannelTD[i] );
            waternorm[i] =(float)Math.sqrt( re[i] * re[i] + im[i] * im[i] ) / ref_scale;
            
            referencePhase[i] = (float)Math.atan2( im[i], re[i] );
            correctedSignalPhase[i] = (float)Math.atan2( m_dispersionChannelTD[i], m_absorptionChannelTD[i] ) - referencePhase[i];
        }
        
        TDFDFittingDefaults def = TDFDFittingDefaults.getInstance( );
        def.load( );
        
        if ( def.getProperty( TDFDFittingDefaults.QUALITY_TYPE ).compareTo( TDFDFittingDefaults.QUALITY_TYPE_AMPLITUDE_AND_PHASE ) == 0 )        
        {
            /*
             * This quality type corrects phase and amplitude
             */
            for ( int i = 0; i < re.length; i++ )
            {
                m_absorptionChannelTD[i] = (float)( norm[i] * Math.cos( correctedSignalPhase[i] ) ) / waternorm[i];
                m_dispersionChannelTD[i] = (float)( norm[i] * Math.sin( correctedSignalPhase[i] ) ) / waternorm[i];
            }
        }
        else
        {
            /*
             * This quality type only corrects the phase
             */
            for ( int i = 0; i < re.length; i++ )
            {
                m_absorptionChannelTD[i] = (float)( norm[i] * Math.cos( correctedSignalPhase[i] ) );
                m_dispersionChannelTD[i] = (float)( norm[i] * Math.sin( correctedSignalPhase[i] ) );
            }
        }
        
        // If response is loaded and set halving the first data point is done automatically
        m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2.0f;
        m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2.0f;
        System.out.println( "SingleVoxelMeasuredSpectralData::applyQuality() : Is doubling the amplitide OK ??????????????????????????????????????");
        
        System.out.println("SingleVoxelMeasuredSpectralData::ApplyQuality - First point is doubled to get it in sync!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        this.setSpectralDataPointsTD( m_absorptionChannelTD, m_dispersionChannelTD );
    }
    
    /**
     * Applies the QUALITY algorithm to the measured data.
     * @param applyFullQuality
     * @param realWater
     * @param imagWater
     */
    public void applyQuality( boolean applyFullQuality, 
                              double[] realWater,
                              double[] imagWater )
    {
        
        double[] norm = new double[realWater.length];
        
        // importing spectrum halves the first time point
        double ref_scale = 2.0f * ( double )Math.sqrt( realWater[0] * realWater[0] + imagWater[0] * imagWater[0] );
        double[] waternorm = new double[realWater.length];
        double[] referencePhase = new double[realWater.length];
        double[] correctedSignalPhase = new double[realWater.length];
        
        for ( int i= 0; i < realWater.length; i++ )
        {
            norm[i] = (double)Math.sqrt( m_absorptionChannelTD[i] * m_absorptionChannelTD[i] + m_dispersionChannelTD[i] * m_dispersionChannelTD[i] );
            waternorm[i] =(double)Math.sqrt( realWater[i] * realWater[i] + imagWater[i] * imagWater[i] ) / ref_scale;
            
            referencePhase[i] = (double)Math.atan2( imagWater[i], realWater[i] );
            correctedSignalPhase[i] = (double)Math.atan2( m_dispersionChannelTD[i], m_absorptionChannelTD[i] ) - referencePhase[i];
        }
        
        TDFDFittingDefaults def = TDFDFittingDefaults.getInstance( );
        def.load( );
        
        if ( def.getProperty( TDFDFittingDefaults.QUALITY_TYPE ).compareTo( TDFDFittingDefaults.QUALITY_TYPE_AMPLITUDE_AND_PHASE ) == 0 )        
        {
            /*
             * This quality type corrects phase and amplitude
             */
            for ( int i = 0; i < realWater.length; i++ )
            {
                m_absorptionChannelTD[i] = (float)( norm[i] * Math.cos( correctedSignalPhase[i] ) / waternorm[i] );
                m_dispersionChannelTD[i] = (float)( norm[i] * Math.sin( correctedSignalPhase[i] ) / waternorm[i] );
            }
        }
        else
        {
            /*
             * This quality type only corrects the phase.
             */
            for ( int i = 0; i < realWater.length; i++ )
            {
                m_absorptionChannelTD[i] = (float)( norm[i] * Math.cos( correctedSignalPhase[i] ) );
                m_dispersionChannelTD[i] = (float)( norm[i] * Math.sin( correctedSignalPhase[i] ) );
            }
        }
        
        // If response is loaded and set halving the first data point is done automatically
//      m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2.0f;
//      m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2.0f;
        System.out.println( "SingleVoxelMeasuredSpectralData::applyQuality() : Is doubling the amplitide OK ??????????????????????????????????????");
        
        System.out.println( "SingleVoxelMeasuredSpectralData::ApplyQuality - First point is doubled to get it in sync!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        this.setSpectralDataPointsTD( m_absorptionChannelTD, m_dispersionChannelTD );
    }
    
    /**
     * Sets the patient name.
     * @param patientName 
     */
    public void setPatientName( String patientName ) 
    {
        m_patientName = patientName;
    }
    
    /**
     * Sets the patients' date of birth.
     * @param patientDOB 
     */
    public void setPatientBirthDate( String patientDOB ) 
    {
        m_patientDOB = patientDOB;
    }

    /**
     * Sets the series description string.
     * @param seriesDescription 
     */
    public void setSeriesDescription( String seriesDescription ) 
    {
        m_seriesDescription = seriesDescription;
    }
    /**
     * Sets the series number the dataset was taken from.
     * @param seriesNumber 
     */
    public void setSeriesNumber( String seriesNumber ) 
    {
        m_seriesNumber = seriesNumber;
    }
  
    /**
     * 
     * @param instanceNumber 
     */
    public void setInstanceNumber( String instanceNumber ) 
    {
        m_instanceNumber = instanceNumber;
    }
    
    /**
     * Sets the study date variable of the study. 
     * @param studyDate 
     */
    public void setAcquisitionDate( String studyDate )
    {
        m_studyDate = studyDate;
    }
    
    /**
     * Returns the patient name. 
     * @return  
     */
    public String getPatientName( ) 
    {
        return m_patientName;
    }
    
    /**
     * Returns the patients' date of birth. 
     * @return  
     */
    public String getPatientBirthDate( ) 
    {
        return m_patientDOB;
    }

    /**
     * Returns the series description string. 
     * @return  
     */
    public String getSeriesDescription( ) 
    {
        return m_seriesDescription;
    }
    
    /**
     * Returns the series number the dataset was taken from. 
     * @return  
     */
    public String getSeriesNumber( ) 
    {
        return m_seriesNumber;
    }
    
    /**
     * Returns the instance number the dataset was taken from. 
     * @return  
     */
    public String getInstanceNumber( ) 
    {
        return m_instanceNumber;
    }
    
    /**
     * Returns the study date variable of the study. 
     * @return  
     */
    public String getAcquisitionDate( )
    {
        return m_studyDate;
    }

    /**
     * Sets the echo time (TE) attribute.
     * @param TE 
     */
    public void setTE( double TE ) 
    {
        m_TE = (int)TE;
    }
    
    /**
     * Sets the repetition time (TR) attribute. 
     * @return 
     */
    public double getTR( ) 
    {
        return m_TR;
    }

    /**
     * Sets the echo time (TE) attribute. 
     * @return 
     */
    @Override
    public int getTE( ) 
    {
        return m_TE;
    }
    
    /**
     * Sets the repetition time (TR) attribute. 
     * @param TR
     */
    public void setTR( double TR ) 
    {
        m_TR = TR;
    }

    public void doComplexConjugation( ) 
    {
        for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
        {
            m_absorptionChannelTD[ i ] = m_absorptionChannelTD[ i ];
            m_dispersionChannelTD[ i ] = -m_dispersionChannelTD[ i ];
            calculateFDRepresentation( );
        }
    }

    public void rotateInFrequencyDomain( int pointToRotate )
    {
        
        float[] rotatedFDAbs  = new float[ m_absorptionChannelFD.length ]; 
        float[] rotatedFDDisp = new float[ m_absorptionChannelFD.length ]; 
        
        for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
        {
            rotatedFDAbs[ i ] = m_absorptionChannelFD[ ( m_absorptionChannelFD.length + pointToRotate + i ) % m_absorptionChannelFD.length ];
            rotatedFDDisp[ i ] = m_dispersionChannelFD[ ( m_absorptionChannelFD.length + pointToRotate + i ) % m_absorptionChannelFD.length ];
        }
        
        System.arraycopy( rotatedFDAbs, 0, m_absorptionChannelFD, 0, m_absorptionChannelFD.length );
        System.arraycopy( rotatedFDDisp, 0, m_dispersionChannelFD, 0, m_absorptionChannelFD.length );
        
        calculateTDRepresentation( );
    }
    
}
