/*
 * File: $Id: SingleVoxelSpectralDataProcessing.java,v 1.1.1.1 2006/04/19 09:38:51 hans Exp $
 * History
 * Date: 13.05.2003, 14:28  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */

package org.spectrim.tdfdfit;


import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.LinkedList;


//import XYPlot.Curve;

/**
 * This class handles the additional signal processing of a 1D single voxel
 * MR spectrum.
 * @author  hans
 */
public class SingleVoxelSpectralDataProcessing 
{
    
    /**
     * Creates a new instance of SingleVoxelSpectralDataProcessing class.
     * 
     * @param responseList
     */
    public SingleVoxelSpectralDataProcessing( LinkedList responseList )
    {
        // m_specData = new SingleVoxelSpectralData( );
        // m_specData.setSpectralDataPointsTD( specData.getAbsorptionChannelTD( ), specData.getDispersionChannelTD( ) );
        Iterator it = responseList.iterator( );
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
            SingleVoxelMeasuredSpectralData specDataBuffer = new SingleVoxelMeasuredSpectralData( specData );
           
            m_responseList.add( specDataBuffer );
        } 
    }
    
    /** 
     * Apodizes the time domain curve with a Lorentzian and or Gaussian 
     * time constant.
     * 
     * @param lorentzianTC Lorentzian time constant in seconds
     * @param gaussianTC Gaussian time constant in seconds
     * @return a SingleVoxelSpectralData object apodized with the Lorentzian and Gaussian
     * apodization constants
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> apodize( float lorentzianTC, float gaussianTC )
    {
        Iterator it = m_responseList.iterator( );
        
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
        
            float[] absorption = new float[ specData.getAbsorptionChannelTD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelTD( ).length ];
            float time;
        
            for ( int i = 0; i < specData.getAbsorptionChannelTD( ).length; i++ )
            {
                // convert to msec units
                time = i * 1000.0f / specData.getSampleFrequency( ) ;
            
                if ( ( gaussianTC != 0.0f ) && ( lorentzianTC != 0.0f ) )
                {
                    absorption[i] = specData.getAbsorptionChannelTD( )[i] * (float)Math.exp( ( - time / lorentzianTC ) - 
                                           ( time * time / ( gaussianTC * gaussianTC ) ) );
                    dispersion[i] = specData.getDispersionChannelTD( )[i] * (float)Math.exp( ( - time / lorentzianTC ) - 
                                               ( time * time / ( gaussianTC * gaussianTC ) ) );
                }
            
                if ( ( gaussianTC == 0.0f ) && ( lorentzianTC != 0.0f ) )
                {
                    absorption[i] = specData.getAbsorptionChannelTD( )[i] * (float)Math.exp( ( - time / lorentzianTC ) );
                    dispersion[i] = specData.getDispersionChannelTD( )[i] * (float)Math.exp( ( - time / lorentzianTC ) );
                }
            
                if ( ( gaussianTC != 0.0f ) && ( lorentzianTC == 0.0f ) )
                {
                    absorption[i] = specData.getAbsorptionChannelTD( )[i] * (float)Math.exp( - time * time / ( gaussianTC * gaussianTC ) );
                    dispersion[i] = specData.getDispersionChannelTD( )[i] * (float)Math.exp( - time * time / ( gaussianTC * gaussianTC ) );                
                }

                if ( ( gaussianTC == 0.0f ) && ( lorentzianTC == 0.0f ) )
                {
                    absorption[i] =  specData.getAbsorptionChannelTD( )[i];
                    dispersion[i] =  specData.getDispersionChannelTD( )[i];                
                }
            }
        
            specData.setSpectralDataPointsTD( absorption, dispersion );
        }
        
        return m_responseList;
    }
    
    /** 
     * Zero fills the Curve response a factor fillFactor times
     * Fill factor fillFactor = 2^n  - ( 0.125, 0.25, 0.5, 1.0, 2.0, 4.0, 8.0 )
     * @param fillFactor is the multiplication factor of the amount of sample points:
     * new amount of sample points = number of sample points * fillFactor
     * @return
     */
    public LinkedList<SingleVoxelMeasuredSpectralData>  zeroFill( float fillFactor )
    {
        
        Iterator it = m_responseList.iterator( );
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
            
            float[] absorption = new float[ (int)( specData.getAbsorptionChannelTD( ).length * fillFactor ) ];
            float[] dispersion = new float[ (int)( specData.getAbsorptionChannelTD( ).length * fillFactor ) ];
        
            for ( int i = 0; i < absorption.length; i++ )
            {
                if ( i < specData.getAbsorptionChannelTD( ).length )
                {
                    absorption[i] = specData.getAbsorptionChannelTD( )[i];
                    dispersion[i] = specData.getDispersionChannelTD( )[i];
                }
            }
        
            specData.setSpectralDataPointsTD( absorption, dispersion );
        }
        
        return m_responseList; 
    }

    /**
     * Adjusts the zero order phase 
     * @param zeroOrderPhase
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> adjustZeroOrderPhase( int zeroOrderPhase )
    {

        Iterator it = m_responseList.iterator( );
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
        
            float[] absorption = new float[ specData.getAbsorptionChannelTD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelTD( ).length ];
        
            float cosPhi = (float)Math.cos( (double)zeroOrderPhase * Math.PI / 180.0 );
            float sinPhi = (float)Math.sin( (double)zeroOrderPhase * Math.PI / 180.0 );
        
            for ( int i = 0; i < absorption.length; i++ )
            {
                if ( i < specData.getAbsorptionChannelTD( ).length )
                {
                    absorption[i] = specData.getAbsorptionChannelFD( )[i] * cosPhi - specData.getDispersionChannelFD( )[i] * sinPhi;
                    dispersion[i] = specData.getAbsorptionChannelFD( )[i] * sinPhi + specData.getDispersionChannelFD( )[i] * cosPhi;
                }
            }

            specData.setSpectralDataPointsFD( absorption, dispersion );
        }
        
        return m_responseList;
    }
    
    /**
     * Applies a frequency shift over "shift" Hz. 
     * @param shift
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> frequencyShift( float shift )
    {
        
        Iterator it = m_responseList.iterator( );
        
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );        
        
            float[] absorption = new float[ specData.getAbsorptionChannelTD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelTD( ).length ];
        
            for ( int i = 0; i < absorption.length; i++ )
            {
                if ( i < specData.getAbsorptionChannelTD( ).length )
                {
                    absorption[i] = (float)( specData.getAbsorptionChannelTD( )[i] * Math.cos( shift * 2.0 * Math.PI * (float)i / specData.getAbsorptionChannelTD( ).length ) );
                    absorption[i] -= (float)( specData.getDispersionChannelTD( )[i] * Math.sin( shift * 2.0 * Math.PI * (float)i / specData.getAbsorptionChannelTD( ).length ) );
                    dispersion[i] = (float)( specData.getAbsorptionChannelTD( )[i] * Math.sin( shift * 2.0 * Math.PI * (float)i / specData.getAbsorptionChannelTD( ).length ) );
                    dispersion[i] += (float)( specData.getDispersionChannelTD( )[i] * Math.cos( shift * 2.0 * Math.PI * (float)i / specData.getAbsorptionChannelTD( ).length ) );
                }
            }
        
            specData.setSpectralDataPointsTD( absorption, dispersion );
            
        }    
        return m_responseList;
        
    }
    
    
    
    /**
     * Multiplies first TD point with a factor 'factor'.
     * @param factor
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> multiplyFirstTDPoint( float factor )
    {
        
        Iterator it = m_responseList.iterator( );
        
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );        
        
            float[] absorption = new float[ specData.getAbsorptionChannelTD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelTD( ).length ];
        
            absorption = specData.getAbsorptionChannelTD( );
            dispersion = specData.getDispersionChannelTD( );
            
            absorption[0] = absorption[0] * factor;
            dispersion[0] = dispersion[0] * factor;
        
            specData.setSpectralDataPointsTD( absorption, dispersion );
            
        }
        
        return m_responseList;
    }
    
    /**
     * Adjusts the first order phase 
     * @param firstOrderPhase
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> adjustFirstOrderPhase( int firstOrderPhase )
    {

        Iterator it = m_responseList.iterator( );
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );        
        
            float[] absorption = new float[ specData.getAbsorptionChannelTD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelTD( ).length ];
        
            float cosPhi;
            float sinPhi;
            float phaseFactor;
        
            for ( int i = 0; i < absorption.length; i++ )
            {
                phaseFactor = (float)( i - absorption.length / 2 ) / (float)absorption.length;
            
                cosPhi = (float)Math.cos( (double)firstOrderPhase * phaseFactor * Math.PI / 180.0 );
                sinPhi = (float)Math.sin( (double)firstOrderPhase * phaseFactor * Math.PI / 180.0 );
            
                if ( i < specData.getAbsorptionChannelTD( ).length )
                {
                    absorption[i] = specData.getAbsorptionChannelFD( )[i] * cosPhi - specData.getDispersionChannelFD( )[i] * sinPhi;
                    dispersion[i] = specData.getAbsorptionChannelFD( )[i] * sinPhi + specData.getDispersionChannelFD( )[i] * cosPhi;
                }
            }
        
            specData.setSpectralDataPointsFD( absorption, dispersion );
            
        }    
        return m_responseList;
    }
    
    
    /**
     * Performs an order statistics filtering on the dataset
     */
    
    
    /**
     * Apply order statistics filtering.
     * @param domain
     * @param removeOffset
     * @param beginAndEnd
     * @param maximum
     * @param minimum
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> applyOrderStatisticsDenoising( int domain, boolean removeOffset, 
                                                                              boolean beginAndEnd, 
                                                                              int minimum, int maximum )
    {
        
        double[][] signal2Dabs = null;
        double[][] signal2Ddisp = null;

        float[] abs = null;
        float[] disp = null;

        // Collect first the correct data
        Iterator it = m_responseList.iterator( );
        int cntr = 0;
        
        int channels = 0;
        int arraySize = 0;
        
        while ( it.hasNext( ) )
        {
            channels = m_responseList.size( );
            
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
            
            arraySize = specData.getAbsorptionChannelTD().length;
            
            if ( ( signal2Dabs == null ) || ( signal2Ddisp == null ) )
            {
                signal2Dabs = new double[channels][arraySize];
                signal2Ddisp = new double[channels][arraySize];
            }
            
            
            if ( domain == OrderStatisticsFilterDefinitionControl.TIME_DOMAIN ) // time domain osf
            {
                abs = specData.getAbsorptionChannelTD( );
                disp = specData.getDispersionChannelTD( );
            }
            else
            {
                abs = specData.getAbsorptionChannelFD( );
                disp = specData.getDispersionChannelFD( );                
            }
            
            for ( int i = 0; i < arraySize; i++ )
            {
                signal2Dabs[cntr][i] = abs[i];
                signal2Ddisp[cntr][i] = disp[i];
            }
            cntr++;
        }
        
        cntr = 0;
        OrderStatisticsSignalFilter osfAbs = new OrderStatisticsSignalFilter( channels, arraySize, signal2Dabs );
        signal2Dabs = osfAbs.performOSFDenoising( beginAndEnd, minimum, maximum );
        
        OrderStatisticsSignalFilter osfDisp = new OrderStatisticsSignalFilter( channels, arraySize, signal2Ddisp );
        signal2Ddisp = osfDisp.performOSFDenoising( beginAndEnd, minimum, maximum );
        
        // Writeback into data structure to be displayed
        Iterator it2 = m_responseList.iterator( );
        while ( it2.hasNext( ) )
        {
            SingleVoxelSpectralData specData = (SingleVoxelSpectralData)it2.next( );
            
            for ( int i = 0; i < arraySize; i++ )
            {
                abs[i]  = (float)signal2Dabs[cntr][i];
                disp[i] = (float)signal2Ddisp[cntr][i];
            }
            
            if ( domain == OrderStatisticsFilterDefinitionControl.TIME_DOMAIN ) // time domain osf
            {
                specData.setSpectralDataPointsTD( abs, disp );
            }
            else
            {
                specData.setSpectralDataPointsFD( abs, disp );
            }
            
            cntr++;
        }
        
        return m_responseList;
    }

    /**
     * Apply order statistics filtering.
     * @param domain
     * @param removeOffset
     * @param beginAndEnd
     * @param minimum
     * @param maximum
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> applyOrderStatisticsFilter( int domain, boolean removeOffset, 
                                                                                   boolean beginAndEnd, 
                                                                                   int minimum, int maximum )
    {
        
        double[][] signal2Dabs = null;
        double[][] signal2Ddisp = null;

        float[] abs = null;
        float[] disp = null;

        // Collect first the correct data
        Iterator it = m_responseList.iterator( );
        int cntr = 0;
        
        int channels = 0;
        int arraySize = 0;
        
        while ( it.hasNext( ) )
        {
            channels = m_responseList.size( );
            
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );
            
            arraySize = specData.getAbsorptionChannelTD().length;
            
            if ( ( signal2Dabs == null ) || ( signal2Ddisp == null ) )
            {
                signal2Dabs = new double[channels][arraySize];
                signal2Ddisp = new double[channels][arraySize];
            }
            
            
            if ( domain == OrderStatisticsFilterDefinitionControl.TIME_DOMAIN ) // time domain osf
            {
                abs = specData.getAbsorptionChannelTD( );
                disp = specData.getDispersionChannelTD( );
            }
            else
            {
                abs = specData.getAbsorptionChannelFD( );
                disp = specData.getDispersionChannelFD( );                
            }
            
            for ( int i = 0; i < arraySize; i++ )
            {
                signal2Dabs[cntr][i] = abs[i];
                signal2Ddisp[cntr][i] = disp[i];
            }
            cntr++;
        }
        
        cntr = 0;
        OrderStatisticsSignalFilter osfAbs = new OrderStatisticsSignalFilter( channels, arraySize, signal2Dabs );
        signal2Dabs = osfAbs.performFiltering( removeOffset, beginAndEnd, minimum, maximum );
        
        OrderStatisticsSignalFilter osfDisp = new OrderStatisticsSignalFilter( channels, arraySize, signal2Ddisp );
        signal2Ddisp = osfDisp.performFiltering( removeOffset, beginAndEnd, minimum, maximum );
        
        // Writeback into data structure to be displayed
        Iterator it2 = m_responseList.iterator( );
        while ( it2.hasNext( ) )
        {
            SingleVoxelSpectralData specData = (SingleVoxelSpectralData)it2.next( );
            
            for ( int i = 0; i < arraySize; i++ )
            {
                abs[i] = (float)signal2Dabs[cntr][i];
                disp[i] = (float)signal2Ddisp[cntr][i];
            }
            
            if ( domain == OrderStatisticsFilterDefinitionControl.TIME_DOMAIN ) // time domain osf
            {
                specData.setSpectralDataPointsTD( abs, disp );
            }
            else
            {
                specData.setSpectralDataPointsFD( abs, disp );
            }
            
            cntr++;
        }
        
        return m_responseList;
    }
    
    /**
     * Performs a offset correction by estimating the frequency domain offset
     * from a pure noise containing range of the spectrum; This FD range is
     * determined by a mini^mum and maximum value.
     * 
     * @param minimumRangeValue
     * @param maximumRangeValue
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> dc_correct( int minimumRangeValue, int maximumRangeValue )
    {

        Iterator it = m_responseList.iterator( );
        int cntr = 0;
        
        while ( it.hasNext( ) )
        {
            SingleVoxelMeasuredSpectralData specData = (SingleVoxelMeasuredSpectralData)it.next( );        
        
            float[] absorption = new float[ specData.getAbsorptionChannelFD( ).length ];
            float[] dispersion = new float[ specData.getAbsorptionChannelFD( ).length ];
        
            float avAbs = 0f;
            float avDisp = 0f;
            
            for ( int i = 0; i < absorption.length; i++ )
            {
                if ( ( i > minimumRangeValue ) && ( i < maximumRangeValue ) ) 
                {
                    avAbs += specData.getAbsorptionChannelFD( )[i];
                    avDisp += specData.getDispersionChannelFD( )[i];
                }
            }
        
            avAbs = avAbs / ( maximumRangeValue - minimumRangeValue );
            avDisp = avDisp / ( maximumRangeValue - minimumRangeValue );
            
            for ( int i = 0; i < absorption.length; i++ )
            {
                absorption[i] = specData.getAbsorptionChannelFD( )[i] - avAbs;
                dispersion[i] = specData.getDispersionChannelFD( )[i] - avDisp;
            }
            
            specData.setSpectralDataPointsFD( absorption, dispersion );
        }
        
        return m_responseList;        
    }
    
    /**
     * Sets the linked list on which the data processing will be applied.
     * @param responseList
     */
    public void setResponseList( LinkedList<SingleVoxelMeasuredSpectralData> responseList )
    {
        m_responseList = responseList;
    }
    
    
    /**
     * Returns the current m_reponseList reference : needed if the operation 
     * is accepted : in that case old data are overwritten with new ones.
     * @return 
     */
    public LinkedList<SingleVoxelMeasuredSpectralData> getResponseList( )
    {
        return m_responseList;
    }
    
    // Attributes
  
    /**
     * Based on the private attribute  
     *
     *      m_baseLineRangesDataModel
     *
     * this routine computes the baseline spline.
     * @param baseLineRangesDataModel
     * @return 
    */
    public LinkedList<SingleVoxelMeasuredSpectralData> removeBaseLineSpline( OrderedBaseLinePointsDataModel baseLineRangesDataModel )
    {
        
        Iterator it = m_responseList.iterator( );
        
        while ( it.hasNext( ) )
        {
            SingleVoxelSpectralData specData = (SingleVoxelSpectralData)it.next( );        
        
            float[] absCurve = specData.getAbsorptionChannelFD( );
        
            CatmullRomSpline b = new CatmullRomSpline( );
        
            // In order to guarantee continuity we have to preceed the interval 
            // model by two pre-points and two post points.
        
            // Determine the points of the spline.
            for ( int i = 1 ; i < baseLineRangesDataModel.getSize( ); i++ )
            {
                Point2D pnt2 = baseLineRangesDataModel.getPoint( i );
                b.addPoint( (float)pnt2.getX( ), (float)pnt2.getY( ) );
            }
        
            float[] baselineSplineAbs = b.computeSplineCurve( absCurve.length );
            float[] baselineSplineDisp = new float[absCurve.length];
        
            for ( int i = 0; i < absCurve.length;  i++ )
            {
                baselineSplineDisp[i] = 0f;
            }
        
            specData.subtractSingleVoxelSpectralDataObject( new SingleVoxelSpectralData( baselineSplineAbs, baselineSplineDisp,  SingleVoxelSpectralData.FREQ_DOMAIN ) );
            
        }
        
        return m_responseList;
        
    }
    
    // All operations are applied on a LinkedList containing objects 
    // of SingleVoxelSpectralData type 
    private LinkedList<SingleVoxelMeasuredSpectralData> m_responseList = new LinkedList<SingleVoxelMeasuredSpectralData>( );

    
}
