/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package org.spectrim.tdfdfit.mmbaseline;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.spectrim.tdfdfit.MRLineModelParameters;
import org.spectrim.tdfdfit.MRSignalProperties;

/**
 * This class defines the macro molecular baseline.
 * Its function is to make it easy to add the MM-baseline into a tdfdfit
 * in the most easy way.
 * The default values for all components are given in the article
 * Rebecca Birch, Andrew C. Peet, Hamid Dehghani, and Martin Wilson,
 * " "
 * Magn Reson Med. 2017 Jan; 77(1): 34–43.
 * 
 * @author Johannes Slotboom <johannes.slotboom at insel.ch>
 */
public class MacroMolecularBaseline 
{
    
    /**
     * This constructs a macromolecular baseline in a easy way. 
     * N.B. only valid for 1H-spectroscopy of the brain.
     */
    public MacroMolecularBaseline( )
    {
        // Component 1 - MM090 Set initial values
        MRLineModelParameters comp090 = new MRLineModelParameters( );
        
        float freq090   = 0.90f;
        float lwidth090 = 21.20f;
        
        comp090.setName( "MM090" );
        comp090.setArea( 0.72f );   // amplitude according to paper referenced above
        comp090.setOffset( freq090 ); // frequency set in ppm
        comp090.setLorentzianWidth( lwidth090 );
        comp090.setPhase( 0.0f );
        comp090.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp090.setMinimumArea( 0.0f );
        comp090.setMinimumOffset( freq090 - m_deltaFreqTolerance );
        comp090.setMinimumLorentzianWidth( lwidth090 - 5.0f );
        comp090.setMinimumPhase( 0.0f );
        comp090.setMinimumGaussianWidth( 0.0f - 180f );
        
        // Set maximum values
        comp090.setMaximumArea( 0.0f );
        comp090.setMaximumOffset( freq090 + m_deltaFreqTolerance );
        comp090.setMaximumLorentzianWidth( lwidth090 + 5.0f );
        comp090.setMaximumPhase( 0.0f + 180f );
        comp090.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp090.setAreaReference( "Free" );
        comp090.setOffsetReference( "Free" );
        comp090.setLorentzianWidthReference( "Free" );
        comp090.setPhaseReference( "Free" );
        comp090.setGaussianWidthReference( "MM090" );
        
        // Other attributes 
        comp090.setFileReferenceName( "none" );
        comp090.setFitWithArbitraryLineShapeFunction( false );
        comp090.setFitWithReferenceLineShape( false );
        
        // Add to the array list
        m_mmSyncedBaselineList.add( comp090 );
        
        // Component 2 - MM121 Set initial values
        MRLineModelParameters comp121 = new MRLineModelParameters( );
        
        float freq121   = 1.21f;
        float lwidth121 = 19.16f;
        
        comp121.setName( "MM121" );
        comp121.setArea( 0.28f );   // amplitude according to paper referenced above
        comp121.setOffset( freq121 ); // frequency set in ppm
        comp121.setLorentzianWidth( lwidth121 );
        comp121.setPhase( 0.0f );
        comp121.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp121.setMinimumArea( 0.0f );
        comp121.setMinimumOffset( freq121 - m_deltaFreqTolerance );
        comp121.setMinimumLorentzianWidth( lwidth121 - 5.0f );
        comp121.setMinimumPhase( 0.0f - 180f );
        comp121.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp121.setMaximumArea( 100000.0f );
        comp121.setMaximumOffset( freq121 + m_deltaFreqTolerance );
        comp121.setMaximumLorentzianWidth( lwidth121 + 5.0f );
        comp121.setMaximumPhase( 0.0f + 180f );
        comp121.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp121.setAreaReference( "Free" );
        comp121.setOffsetReference( "MM090" );
        comp121.setLorentzianWidthReference( "MM090" );
        comp121.setPhaseReference( "MM090" );
        comp121.setGaussianWidthReference( "MM121" );
        
        // Other attributes 
        comp121.setFileReferenceName( "none" );
        comp121.setFitWithArbitraryLineShapeFunction( false );
        comp121.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp121 );
        
        
        // Component 3 - MM130 Set initial values
        MRLineModelParameters comp130 = new MRLineModelParameters( );
        
        float freq130   = 1.30f;
        float lwidth130 = 21.20f;
        
        comp130.setName( "MM130" );
        comp130.setArea( 0.28f );   // amplitude according to paper referenced above
        comp130.setOffset( freq130 ); // frequency set in ppm
        comp130.setLorentzianWidth( lwidth130 );
        comp130.setPhase( 0.0f );
        comp130.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp130.setMinimumArea( 0.0f );
        comp130.setMinimumOffset( freq130 - m_deltaFreqTolerance );
        comp130.setMinimumLorentzianWidth( lwidth130 - 5.0f );
        comp130.setMinimumPhase( 0.0f - 180f );
        comp130.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp130.setMaximumArea( 100000.0f );
        comp130.setMaximumOffset( freq130 + m_deltaFreqTolerance );
        comp130.setMaximumLorentzianWidth( lwidth130 + 5.0f );
        comp130.setMaximumPhase( 0.0f + 180f );
        comp130.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp130.setAreaReference( "Free" );
        comp130.setOffsetReference( "MM090" );
        comp130.setLorentzianWidthReference( "MM090" );
        comp130.setPhaseReference( "MM090" );
        comp130.setGaussianWidthReference( "MM130" );
        
        // Other attributes 
        comp130.setFileReferenceName( "none" );
        comp130.setFitWithArbitraryLineShapeFunction( false );
        comp130.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp130 );
        
        // Component 4 - MM138 Set initial values
        MRLineModelParameters comp138 = new MRLineModelParameters( );
        
        float freq138   = 1.38f;
        float lwidth138 = 15.90f;
        
        comp138.setName( "MM138" );
        comp138.setArea( 0.38f );   // amplitude according to paper referenced above
        comp138.setOffset( freq138 ); // frequency set in ppm
        comp138.setLorentzianWidth( lwidth138 );
        comp138.setPhase( 0.0f );
        comp138.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp138.setMinimumArea( 0.0f );
        comp138.setMinimumOffset( freq138 - m_deltaFreqTolerance );
        comp138.setMinimumLorentzianWidth( lwidth138 - 5.0f );
        comp138.setMinimumPhase( 0.0f - 180f );
        comp138.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp138.setMaximumArea( 100000.0f );
        comp138.setMaximumOffset( freq138 + m_deltaFreqTolerance );
        comp138.setMaximumLorentzianWidth( lwidth138 + 5.0f );
        comp138.setMaximumPhase( 0.0f + 180f );
        comp138.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp138.setAreaReference( "Free" );
        comp138.setOffsetReference( "MM090" );
        comp138.setLorentzianWidthReference( "MM090" );
        comp138.setPhaseReference( "MM090" );
        comp138.setGaussianWidthReference( "MM138" );
        
        // Other attributes 
        comp138.setFileReferenceName( "none" );
        comp138.setFitWithArbitraryLineShapeFunction( false );
        comp138.setFitWithReferenceLineShape( false );
        
         m_mmSyncedBaselineList.add( comp138 );
        
        // Component 5 - MM163 Set initial values
        MRLineModelParameters comp163 = new MRLineModelParameters( );
        
        float freq163   = 1.63f;  //ppm
        float lwidth163 = 7.50f;  //Hz
        
        comp163.setName( "MM163" );
        comp163.setArea( 0.05f );   // amplitude according to paper referenced above
        comp163.setOffset( freq163 ); // frequency set in ppm
        comp163.setLorentzianWidth( lwidth163 );
        comp163.setPhase( 0.0f );
        comp163.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp163.setMinimumArea( 0.0f );
        comp163.setMinimumOffset( freq163 - m_deltaFreqTolerance );
        comp163.setMinimumLorentzianWidth( lwidth163 - 5.0f );
        comp163.setMinimumPhase( 0.0f - 180f );
        comp163.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp163.setMaximumArea( 100000.0f );
        comp163.setMaximumOffset( freq163 + m_deltaFreqTolerance );
        comp163.setMaximumLorentzianWidth( lwidth163 + 5.0f );
        comp163.setMaximumPhase( 0.0f + 180f );
        comp163.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp163.setAreaReference( "Free" );
        comp163.setOffsetReference( "MM090" );
        comp163.setLorentzianWidthReference( "MM090" );
        comp163.setPhaseReference( "MM090" );
        comp163.setGaussianWidthReference( "MM163" );
        
        // Other attributes 
        comp163.setFileReferenceName( "none" );
        comp163.setFitWithArbitraryLineShapeFunction( false );
        comp163.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp163 );
        
        
        // Component 6 - MM201 Set initial values
        MRLineModelParameters comp201 = new MRLineModelParameters( );
        
        float freq201   = 2.01f;   //ppm
        float lwidth201 = 29.30f;  //Hz
        
        comp201.setName( "MM201" );
        comp201.setArea( 0.45f );   // amplitude according to paper referenced above
        comp201.setOffset( freq201 ); // frequency set in ppm
        comp201.setLorentzianWidth( lwidth201 );
        comp201.setPhase( 0.0f );
        comp201.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp201.setMinimumArea( 0.0f );
        comp201.setMinimumOffset( freq201 - m_deltaFreqTolerance );
        comp201.setMinimumLorentzianWidth( lwidth201 - 5.0f );
        comp201.setMinimumPhase( 0.0f - 180f );
        comp201.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp201.setMaximumArea( 100000.0f );
        comp201.setMaximumOffset( freq201 + m_deltaFreqTolerance );
        comp201.setMaximumLorentzianWidth( lwidth201 + 5.0f );
        comp201.setMaximumPhase( 0.0f + 180f );
        comp201.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp201.setAreaReference( "Free" );
        comp201.setOffsetReference( "MM090" );
        comp201.setLorentzianWidthReference( "MM090" );
        comp201.setPhaseReference( "MM090" );
        comp201.setGaussianWidthReference( "MM201" );
        
        // Other attributes 
        comp201.setFileReferenceName( "none" );
        comp201.setFitWithArbitraryLineShapeFunction( false );
        comp201.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp201 );
          
        // Component 7 - MM209 Set initial values
        MRLineModelParameters comp209 = new MRLineModelParameters( );
        
        float freq209   = 2.09f;   //ppm
        float lwidth209 = 20.53f;  //Hz
        
        comp209.setName( "MM209" );
        comp209.setArea( 0.36f );   // amplitude according to paper referenced above
        comp209.setOffset( freq209 ); // frequency set in ppm
        comp209.setLorentzianWidth( lwidth209 );
        comp209.setPhase( 0.0f );
        comp209.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp209.setMinimumArea( 0.0f );
        comp209.setMinimumOffset( freq209 - m_deltaFreqTolerance );
        comp209.setMinimumLorentzianWidth( lwidth209 - 5.0f );
        comp209.setMinimumPhase( 0.0f - 180f );
        comp209.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp209.setMaximumArea( 100000.0f );
        comp209.setMaximumOffset( freq209 + m_deltaFreqTolerance );
        comp209.setMaximumLorentzianWidth( lwidth209 + 5.0f );
        comp209.setMaximumPhase( 0.0f + 180f );
        comp209.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp209.setAreaReference( "Free" );
        comp209.setOffsetReference( "MM090" );
        comp209.setLorentzianWidthReference( "MM090" );
        comp209.setPhaseReference( "MM090" );
        comp209.setGaussianWidthReference( "MM209" );
        
        // Other attributes 
        comp209.setFileReferenceName( "none" );
        comp209.setFitWithArbitraryLineShapeFunction( false );
        comp209.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp209 );
        
        // Component 8 - MM225 Set initial values
        MRLineModelParameters comp225 = new MRLineModelParameters( );
        
        float freq225   = 2.25f;   //ppm
        float lwidth225 = 17.89f;  //Hz
        
        comp225.setName( "MM225" );
        comp225.setArea( 0.36f );     // amplitude according to paper referenced above
        comp225.setOffset( freq225 ); // frequency set in ppm
        comp225.setLorentzianWidth( lwidth225 );
        comp225.setPhase( 0.0f );
        comp225.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp225.setMinimumArea( 0.0f );
        comp225.setMinimumOffset( freq225 - m_deltaFreqTolerance );
        comp225.setMinimumLorentzianWidth( lwidth225 - 5.0f );
        comp225.setMinimumPhase( 0.0f - 180f );
        comp225.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp225.setMaximumArea( 100000.0f );
        comp225.setMaximumOffset( freq225 + m_deltaFreqTolerance );
        comp225.setMaximumLorentzianWidth( lwidth225 + 5.0f );
        comp225.setMaximumPhase( 0.0f + 180f );
        comp225.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp225.setAreaReference( "Free" );
        comp225.setOffsetReference( "MM090" );
        comp225.setLorentzianWidthReference( "MM090" );
        comp225.setPhaseReference( "MM090" );
        comp225.setGaussianWidthReference( "MM225" );
        
        // Other attributes 
        comp225.setFileReferenceName( "none" );
        comp225.setFitWithArbitraryLineShapeFunction( false );
        comp225.setFitWithReferenceLineShape( false );

        m_mmSyncedBaselineList.add( comp225 );
        
        
        // Component 9 - MM261 Set initial values
        MRLineModelParameters comp261 = new MRLineModelParameters( );
        
        float freq261   = 2.61f;  //ppm
        float lwidth261 = 5.30f;  //Hz
        
        comp261.setName( "MM261" );
        comp261.setArea( 0.04f );     // amplitude according to paper referenced above
        comp261.setOffset( freq261 ); // frequency set in ppm
        comp261.setLorentzianWidth( lwidth261 );
        comp261.setPhase( 0.0f );
        comp261.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp261.setMinimumArea( 0.0f );
        comp261.setMinimumOffset( freq261 - m_deltaFreqTolerance );
        comp261.setMinimumLorentzianWidth( lwidth261 - 5.0f );
        comp261.setMinimumPhase( 0.0f - 180f );
        comp261.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp261.setMaximumArea( 100000.0f );
        comp261.setMaximumOffset( freq261 + m_deltaFreqTolerance );
        comp261.setMaximumLorentzianWidth( lwidth261 + 5.0f );
        comp261.setMaximumPhase( 0.0f + 180f );
        comp261.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp261.setAreaReference( "Free" );
        comp261.setOffsetReference( "MM090" );
        comp261.setLorentzianWidthReference( "MM090" );
        comp261.setPhaseReference( "MM090" );
        comp261.setGaussianWidthReference( "MM261" );
        
        // Other attributes 
        comp261.setFileReferenceName( "none" );
        comp261.setFitWithArbitraryLineShapeFunction( false );
        comp261.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp261 );
        
        // Component 10 - MM296 Set initial values
        MRLineModelParameters comp296 = new MRLineModelParameters( );
        
        float freq296   = 2.96f;  //ppm
        float lwidth296 = 14.02f; //Hz
        
        comp296.setName( "MM296" );
        comp296.setArea( 0.20f );     // amplitude according to paper referenced above
        comp296.setOffset( freq296 ); // frequency set in ppm
        comp296.setLorentzianWidth( lwidth296 );
        comp296.setPhase( 0.0f );
        comp296.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp296.setMinimumArea( 0.0f );
        comp296.setMinimumOffset( freq296 - m_deltaFreqTolerance );
        comp296.setMinimumLorentzianWidth( lwidth296 - 5.0f );
        comp296.setMinimumPhase( 0.0f - 180f );
        comp296.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp296.setMaximumArea( 100000.0f );
        comp296.setMaximumOffset( freq296 + m_deltaFreqTolerance );
        comp296.setMaximumLorentzianWidth( lwidth296 + 5.0f );
        comp296.setMaximumPhase( 0.0f + 180f );
        comp296.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp296.setAreaReference( "Free" );
        comp296.setOffsetReference( "MM090" );
        comp296.setLorentzianWidthReference( "MM090" );
        comp296.setPhaseReference( "MM090" );
        comp296.setGaussianWidthReference( "MM296" );
        
        // Other attributes 
        comp296.setFileReferenceName( "none" );
        comp296.setFitWithArbitraryLineShapeFunction( false );
        comp296.setFitWithReferenceLineShape( false );
        
        
        m_mmSyncedBaselineList.add( comp296 );
        
        // Component 11 - MM311 Set initial values
        MRLineModelParameters comp311 = new MRLineModelParameters( );
        
        float freq311   = 3.11f;   //ppm
        float lwidth311 = 17.89f;  //Hz
        
        comp311.setName( "MM311" );
        comp311.setArea( 0.11f );     // amplitude according to paper referenced above
        comp311.setOffset( freq311 ); // frequency set in ppm
        comp311.setLorentzianWidth( lwidth311 );
        comp311.setPhase( 0.0f );
        comp311.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp311.setMinimumArea( 0.0f );
        comp311.setMinimumOffset( freq311 - m_deltaFreqTolerance );
        comp311.setMinimumLorentzianWidth( lwidth311 - 5.0f );
        comp311.setMinimumPhase( 0.0f - 180f );
        comp311.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp311.setMaximumArea( 100000.0f );
        comp311.setMaximumOffset( freq311 + m_deltaFreqTolerance );
        comp311.setMaximumLorentzianWidth( lwidth311 + 5.0f );
        comp311.setMaximumPhase( 0.0f + 180f );
        comp311.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp311.setAreaReference( "Free" );
        comp311.setOffsetReference( "MM090" );
        comp311.setLorentzianWidthReference( "MM090" );
        comp311.setPhaseReference( "MM090" );
        comp311.setGaussianWidthReference( "MM311" );
        
        // Other attributes 
        comp311.setFileReferenceName( "none" );
        comp311.setFitWithArbitraryLineShapeFunction( false );
        comp311.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp311 );
        
        // Component 12 - MM367 Set initial values
        MRLineModelParameters comp367 = new MRLineModelParameters( );
        
        float freq367   = 3.67f;  //ppm
        float lwidth367 = 33.52f;  //Hz
        
        comp367.setName( "MM367" );
        comp367.setArea( 0.64f );     // amplitude according to paper referenced above
        comp367.setOffset( freq367 ); // frequency set in ppm
        comp367.setLorentzianWidth( lwidth367 );
        comp367.setPhase( 0.0f );
        comp367.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp367.setMinimumArea( 0.0f );
        comp367.setMinimumOffset( freq367 - m_deltaFreqTolerance );
        comp367.setMinimumLorentzianWidth( lwidth367 - 5.0f );
        comp367.setMinimumPhase( 0.0f - 180f );
        comp367.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp367.setMaximumArea( 100000.0f );
        comp367.setMaximumOffset( freq367 + m_deltaFreqTolerance );
        comp367.setMaximumLorentzianWidth( lwidth367 + 5.0f );
        comp367.setMaximumPhase( 0.0f + 180f );
        comp367.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp367.setAreaReference( "Free" );
        comp367.setOffsetReference( "MM090" );
        comp367.setLorentzianWidthReference( "MM090" );
        comp367.setPhaseReference( "MM090" );
        comp367.setGaussianWidthReference( "MM367" );
        
        // Other attributes 
        comp367.setFileReferenceName( "none" );
        comp367.setFitWithArbitraryLineShapeFunction( false );
        comp367.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp367 );
        
        // Component 13 - MM380 Set initial values
        MRLineModelParameters comp380 = new MRLineModelParameters( );
        
        float freq380   = 3.80f;  //ppm
        float lwidth380 = 11.85f;  //Hz
        
        comp380.setName( "MM380" );
        comp380.setArea( 0.07f );     // amplitude according to paper referenced above
        comp380.setOffset( freq380 ); // frequency set in ppm
        comp380.setLorentzianWidth( lwidth380 );
        comp380.setPhase( 0.0f );
        comp380.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp380.setMinimumArea( 0.0f );
        comp380.setMinimumOffset( freq380 - m_deltaFreqTolerance );
        comp380.setMinimumLorentzianWidth( lwidth380 - 5.0f );
        comp380.setMinimumPhase( 0.0f - 180f );
        comp380.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp380.setMaximumArea( 100000.0f );
        comp380.setMaximumOffset( freq380 + m_deltaFreqTolerance );
        comp380.setMaximumLorentzianWidth( lwidth380 + 5.0f );
        comp380.setMaximumPhase( 0.0f + 180f );
        comp380.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp380.setAreaReference( "Free" );
        comp380.setOffsetReference( "MM090" );
        comp380.setLorentzianWidthReference( "MM090" );
        comp380.setPhaseReference( "MM090" );
        comp380.setGaussianWidthReference( "MM380" );
        
        // Other attributes 
        comp380.setFileReferenceName( "none" );
        comp380.setFitWithArbitraryLineShapeFunction( false );
        comp380.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp380 );
        
        // Component 14 - MM396 Set initial values
        MRLineModelParameters comp396 = new MRLineModelParameters( );
        
        float freq396   = 3.96f;   //ppm
        float lwidth396 = 37.48f;  //Hz
        
        comp396.setName( "MM396" );
        comp396.setArea( 1.00f );     // amplitude according to paper referenced above
        comp396.setOffset( freq396 ); // frequency set in ppm
        comp396.setLorentzianWidth( lwidth396 );
        comp396.setPhase( 0.0f );
        comp396.setGaussianWidth( 0.0f );
        
        // Set minimum values
        comp396.setMinimumArea( 0.0f );
        comp396.setMinimumOffset( freq396 - m_deltaFreqTolerance );
        comp396.setMinimumLorentzianWidth( lwidth396 - 5.0f );
        comp396.setMinimumPhase( 0.0f - 180f );
        comp396.setMinimumGaussianWidth( 0.0f );
        
        // Set maximum values
        comp396.setMaximumArea( 100000.0f );
        comp396.setMaximumOffset( freq396 + m_deltaFreqTolerance );
        comp396.setMaximumLorentzianWidth( lwidth396 + 5.0f );
        comp396.setMaximumPhase( 0.0f + 180f );
        comp396.setMaximumGaussianWidth( 0.0f );
        
        // Set prior knowledge
        comp396.setAreaReference( "Free" );
        comp396.setOffsetReference( "MM090" );
        comp396.setLorentzianWidthReference( "MM090" );
        comp396.setPhaseReference( "MM090" );
        comp396.setGaussianWidthReference( "MM396" );
        
        // Other attributes 
        comp396.setFileReferenceName( "none" );
        comp396.setFitWithArbitraryLineShapeFunction( false );
        comp396.setFitWithReferenceLineShape( false );
        
        m_mmSyncedBaselineList.add( comp396 );
        
    }
    
    /**
     * Returns the baseline without the 1.30 ppm component.
     * @param fixedAmplitudes
     * @param withMM130
     * @return 
     */
    public List getBaseLineComponentList( boolean fixedAmplitudes, boolean withMM130 )
    {
        ArrayList<MRLineModelParameters> adHocBaselineList = new ArrayList<>( );       
        List adHocSyncedBaselineList = Collections.synchronizedList( adHocBaselineList );
        for ( int i = 0; i < m_mmSyncedBaselineList.size( ); i++ )
        {
            MRLineModelParameters par = new MRLineModelParameters( (MRLineModelParameters)m_mmSyncedBaselineList.get( i ) );
            
            if ( fixedAmplitudes )
            {
                if ( par.getName( ).compareTo( "MM090" ) != 0  )
                {
                    par.setAreaReference( "MM090" );
                }
            }
            MRSignalProperties props = MRSignalProperties.getInstance( );
            
            // Adapt the internal offset to tdfdfit internal offset representation (Hz)
            float ppmRange =  props.getAcquistitionBandwidth( ) / ( props.getResonanceFrequency( ) / 1000000f );
            float internal_offset = - ( props.getNumberOfSamplePoints( ) / ppmRange ) * par.getOffset( ) + (  props.getNumberOfSamplePoints( ) / ppmRange ) * props.getPPMValueInCenterOfSpectrum( );
            par.setOffset( internal_offset );
            
            // now we can add to the list to be returned to the calling object!
            if ( ( par.getName( ).compareTo( "MM130" ) == 0 ) && 
                 ( withMM130 == false ) )
            {
                // do nothing
            }
            else
            {
                adHocSyncedBaselineList.add( par );
            }
            
        }
        
        return adHocSyncedBaselineList;
    }
    
    private final ArrayList<MRLineModelParameters> m_mmBaselineList = new ArrayList<>( );       
    private final List m_mmSyncedBaselineList = Collections.synchronizedList( m_mmBaselineList );
    
    private final float m_deltaFreqTolerance = 0.05f;
    
}
